Package: SUNGEO
Type: Package
Title: Sub-National Geospatial Data Archive: Geoprocessing Toolkit
Version: 1.4.0
Date: 2026-05-28
Authors@R: c(person(given = c("Yuri", "M."),
                    family = "Zhukov",
                    role = c("aut", "cre"),
                    email = "ymz2@georgetown.edu"),
             person(given = "Jason",
                    family = "Byers",
                    role = "aut"),
             person(given = "Marty",
                    family = "Davidson",
                    role = "aut"))
Description: Integrates spatially-misaligned GIS datasets across incompatible
    geographic units. Part of the Sub-National Geospatial Data Archive System.
    For the underlying methods, see Zhukov, Byers, Davidson, and Kollman (2024)
    "Integrating Data Across Misaligned Spatial Units," Political Analysis,
    Volume 32, Number 1, pp. 17-33 
    <doi:10.1017/pan.2023.5>.
URL: https://www.sungeo.org/, https://github.com/zhukovyuri/SUNGEO
License: GPL-2
Encoding: UTF-8
LazyData: TRUE
Depends: R (>= 3.5.0)
Imports:
        sf,data.table,dplyr,jsonlite,terra,stringr,stats,methods,purrr,measurements,RANN,cartogram,packcircles,rmapshaper,spdep,Rcpp,httr,rlang,curl
Suggests: lwgeom
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-28 17:39:34 UTC; zhukov
Author: Yuri M. Zhukov [aut, cre],
  Jason Byers [aut],
  Marty Davidson [aut]
Maintainer: Yuri M. Zhukov <ymz2@georgetown.edu>
Repository: CRAN
Date/Publication: 2026-06-01 13:50:07 UTC
Built: R 4.5.2; ; 2026-06-01 17:15:42 UTC; unix

Package: SUNGEO
Type: Package
Title: Sub-National Geospatial Data Archive: Geoprocessing Toolkit
Version: 1.4.0
Date: 2026-05-28
Authors@R: c(person(given = c("Yuri", "M."),
                    family = "Zhukov",
                    role = c("aut", "cre"),
                    email = "ymz2@georgetown.edu"),
             person(given = "Jason",
                    family = "Byers",
                    role = "aut"),
             person(given = "Marty",
                    family = "Davidson",
                    role = "aut"))
Description: Integrates spatially-misaligned GIS datasets across incompatible
    geographic units. Part of the Sub-National Geospatial Data Archive System.
    For the underlying methods, see Zhukov, Byers, Davidson, and Kollman (2024)
    "Integrating Data Across Misaligned Spatial Units," Political Analysis,
    Volume 32, Number 1, pp. 17-33 
    <doi:10.1017/pan.2023.5>.
URL: https://www.sungeo.org/, https://github.com/zhukovyuri/SUNGEO
License: GPL-2
Encoding: UTF-8
LazyData: TRUE
Depends: R (>= 3.5.0)
Imports:
        sf,data.table,dplyr,jsonlite,terra,stringr,stats,methods,purrr,measurements,RANN,cartogram,packcircles,rmapshaper,spdep,Rcpp,httr,rlang,curl
Suggests: lwgeom
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-28 17:39:34 UTC; zhukov
Author: Yuri M. Zhukov [aut, cre],
  Jason Byers [aut],
  Marty Davidson [aut]
Maintainer: Yuri M. Zhukov <ymz2@georgetown.edu>
Repository: CRAN
Date/Publication: 2026-06-01 13:50:07 UTC
Built: R 4.5.2; ; 2026-06-01 14:46:44 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,session)
export(df2sf)
export(fix_geom)
export(geocode_osm)
export(geocode_osm_batch)
export(get_data)
export(get_info)
export(hot_spot)
export(line2poly)
export(make_ticker)
export(merge_list)
export(nesting)
export(point2poly_simp)
export(point2poly_tess)
export(poly2poly_ap)
export(sf2raster)
export(smart_round)
export(update_bbox)
export(utm_select)
import(Rcpp)
import(cartogram)
import(packcircles)
import(sf)
importFrom(RANN,nn2)
importFrom(cartogram,cartogram_cont)
importFrom(curl,curl_version)
importFrom(data.table,":=")
importFrom(data.table,as.data.table)
importFrom(data.table,copy)
importFrom(data.table,data.table)
importFrom(data.table,fread)
importFrom(data.table,rbindlist)
importFrom(data.table,setnames)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,everything)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,last)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_all)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,starts_with)
importFrom(dplyr,summarize)
importFrom(httr,GET)
importFrom(httr,add_headers)
importFrom(httr,config)
importFrom(httr,content)
importFrom(httr,handle)
importFrom(httr,headers)
importFrom(httr,status_code)
importFrom(httr,warn_for_status)
importFrom(jsonlite,fromJSON)
importFrom(measurements,conv_unit)
importFrom(methods,as)
importFrom(purrr,reduce)
importFrom(rlang,abort)
importFrom(rmapshaper,ms_dissolve)
importFrom(sf,sf_use_s2)
importFrom(sf,st_area)
importFrom(sf,st_as_sf)
importFrom(sf,st_as_sfc)
importFrom(sf,st_bbox)
importFrom(sf,st_buffer)
importFrom(sf,st_cast)
importFrom(sf,st_centroid)
importFrom(sf,st_collection_extract)
importFrom(sf,st_combine)
importFrom(sf,st_convex_hull)
importFrom(sf,st_coordinates)
importFrom(sf,st_crs)
importFrom(sf,st_difference)
importFrom(sf,st_geometry)
importFrom(sf,st_intersection)
importFrom(sf,st_is)
importFrom(sf,st_is_valid)
importFrom(sf,st_make_valid)
importFrom(sf,st_transform)
importFrom(sf,st_union)
importFrom(spdep,localG)
importFrom(spdep,localmoran)
importFrom(spdep,nb2listw)
importFrom(spdep,poly2nb)
importFrom(stats,as.dist)
importFrom(stats,median)
importFrom(stats,weighted.mean)
importFrom(stringr,str_extract_all)
importFrom(stringr,str_split)
importFrom(terra,as.data.frame)
importFrom(terra,crds)
importFrom(terra,ext)
importFrom(terra,extract)
importFrom(terra,mask)
importFrom(terra,ncell)
importFrom(terra,plot)
importFrom(terra,project)
importFrom(terra,rasterize)
importFrom(terra,res)
importFrom(terra,values)
importFrom(terra,vect)
importFrom(utils,URLencode)
importFrom(utils,data)

# `SUNGEO` / Sub-National Geospatial Data Archive: Geoprocessing Toolkit
R package for integrating spatially-misaligned GIS datasets.

Version 1.4.0 (May 28, 2026)

Jason Byers, Marty Davidson, Yuri M. Zhukov

Feedback, bug reports welcome: ymz2-at-georgetown-dot-edu

* `df2sf` / Convert data.frame object into simple features object
* `fix_geom` / Fix polygon geometries
* `geocode_osm` / Geocode addresses with OpenStreetMap
* `geocode_osm_batch` / Batch geocode addresses with OpenStreetMap
* `get_data` / Download data from SUNGEO server
* `get_info` / Information on available SUNGEO data files
* `hot_spot` / Automatically calculate Local G hot spot intensity
* `line2poly` / Line-in-polygon analysis
* `make_ticker` / Make date ticker
* `merge_list` / Merge list of tables on common variable(s)
* `nesting` / Relative scale and nesting coefficients
* `point2poly_simp` / Point-to-polygon interpolation, simple overlay method
* `point2poly_tess` / Point-to-polygon interpolation, tessellation method
* `poly2poly_ap` / Area and population weighted polygon-to-polygon interpolation
* `sf2raster` / Convert simple features object into regularly spaced raster
* `smart_round` / Smart numerical rounding function
* `utm_select` / Automatically convert geographic (degree) to planar coordinates (meters)
* `update_bbox` / Update bounding box of sf object


To install in R:

```
library(devtools)
devtools::install_github("zhukovyuri/SUNGEO", dependencies = TRUE)
```

Load package:

```
library(SUNGEO)
```

Read help files:

```
?poly2poly_ap
?utm_select
?get_data
```

Example: geocode an address with OpenStreetMap

```
# Get geographic coordinates for the Big House (top match only)
geocode_osm("Michigan Stadium")

# Return detailed results for top match
geocode_osm("Michigan Stadium", details=TRUE)

# Find all the Buddy's Pizza locations (the original is on Conant St.)
geocode_osm("Buddy's Pizza", details=TRUE, return_all = TRUE)

# Reverse geocode a coordinate pair
geocode_osm(reverse = TRUE, lon = -83.74868, lat = 42.26587, zoom = 18)

```

Example: geocode multiple addresses

```
# Geocode multiple addresses (top matches only)
geocode_osm_batch(c("Ann Arbor","East Lansing","Columbus"))

# ... with progress reports
geocode_osm_batch(c("Ann Arbor","East Lansing","Columbus"), 
                  verbose = TRUE)

# Return detailed results for all matches
geocode_osm_batch(c("Ann Arbor","East Lansing","Columbus"),
                  details = TRUE, return_all = TRUE)

```

Example: scale and nesting metrics for two polygons

```
# Load source polygons (legislative districts)
data(clea_deu2009)

# Load destination polygons (grid cells)
data(hex_05_deu)

# Preview
plot(clea_deu2009["geometry"],reset=FALSE)
plot(hex_05_deu["geometry"],add=TRUE,border="grey")

# Calculate all scale and nesting metrics at once
nest_1 <- nesting(
              poly_from = clea_deu2009,
              poly_to = hex_05_deu
              )
nest_1

# Calculate just Relative Nesting, in the opposite direction
nest_2 <- nesting(
              poly_from = hex_05_deu,
              poly_to = clea_deu2009,
              metrix = "rn"
              )
nest_2
```

Example: area-weighted polygon-to-polygon interpolation

```
# Load legislative election results (from CLEA)
data(clea_deu2009)

# Visualize voter turnout at constituency level
plot(clea_deu2009["to1"])

# Load 0.5 degree hexagonal grid
data(hex_05_deu)

# Interpolate
out_1 <- poly2poly_ap(poly_from = clea_deu2009,
                      poly_to = hex_05_deu,
                      poly_to_id = "HEX_ID",
                      varz = "to1"
                      )

# Visualize voter turnout at grid cell level
plot(out_1["to1_aw"])
```

Example: population-weighted polygon-to-polygon interpolation

```
# Load population raster (from GPW v4)
data(gpw4_deu2010)

# Unwrap PackedSpatRaster
gpw4_deu2010 <- terra::rast(gpw4_deu2010) 

# Interpolate
out_2 <- poly2poly_ap(poly_from = clea_deu2009,
                      poly_to = hex_05_deu,
                      poly_to_id = "HEX_ID",
                      varz = "to1",
                      methodz = "pw",
                      pop_raster = gpw4_deu2010)

# Visualize voter turnout at grid cell level
plot(out_2["to1_pw"])
```

Example: point-to-polygon interpolation using tessellation method and area weights

```
# Load point-level election results
data(clea_deu2009_pt)

# Interpolate
out_4 <- point2poly_tess(pointz = clea_deu2009_pt,
                         polyz = hex_05_deu,
                         poly_id = "HEX_ID",
                         varz = "to1")

# Visualize voter turnout at grid cell level 
plot(out_4["to1_aw"])
```

Example: line-in-polygon analysis

```
# Load road data (from Digital Chart of the World) and extract highways
data(highways_deu1992)

# Basic map overlay
plot(hex_05_deu["geometry"], reset=TRUE)
plot(highways_deu1992$geometry, add=TRUE, col = "blue", lwd=2)

# Calculate road lengths, densities and distances from each polygon to nearest highway
out_8 <- line2poly(linez = highways_deu1992,
                   polyz = hex_05_deu,
                   poly_id = "HEX_ID")
                   
# Visualize results
plot(out_8["line_length"])
plot(out_8["line_density"])
plot(out_8["line_distance"])

# Replace missing road lengths and densities with 0's, rename variables
out_9 <- line2poly(linez = highways_deu1992,
                   polyz = hex_05_deu,
                   poly_id = "HEX_ID",
                   outvar_name = "road",
                   na_val = 0)

# Visualize results
plot(out_9["road_length"])
plot(out_9["road_density"])
plot(out_9["road_distance"])
dev.off()
```

Example: Automatically find a planar CRS for a GIS dataset

```
# Visualize original geometries (WGS1984, degrees)
plot(clea_deu2009["geometry"], axes=TRUE)

# Find a suitable CRS and re-project
out_10 <- utm_select(clea_deu2009)

# Visualize transformed geometries (UTM 32N, meters)
plot(out_10["geometry"], axes=TRUE)

# proj4string of transformed data
utm_select(clea_deu2009, return_list=TRUE)$proj_out
```

Example: Rasterization of polygons

```
# Transform sf polygon layer into 1km-by-1km RasterLayer (requires planar CRS)
out_11 <- sf2raster(polyz_from = utm_select(clea_deu2009),
                   input_variable = "to1")
# Visualize raster
terra::plot(out_11)

# 25km-by-25km RasterLayer (requires planar CRS)
out_12 <- sf2raster(polyz_from = utm_select(clea_deu2009),
                   input_variable = "to1",
                   grid_dim = c(100, 100))
# Visualize raster
terra::plot(out_12)
```

Example: Create cartogram

```
# Cartogram of turnout scaled by number of valid votes
out_13 <- sf2raster(polyz_from = utm_select(clea_deu2009),
                  input_variable = "to1",
                  cartogram = TRUE,
                  carto_var = "vv1")
terra::plot(out_13)
```

Example: reverse rasterization

```
# Polygonization of cartogram raster
out_14a <- sf2raster(polyz_from = utm_select(clea_deu2009),
                    input_variable = "to1",
                    cartogram = TRUE,
                    carto_var = "vv1",
                    return_list = TRUE)
out_14 <- sf2raster(reverse = TRUE,
                   poly_to = out_14a$poly_to,
                   return_output = out_14a$return_output,
                   return_field = out_14a$return_field)
plot(out_14["to1"])

```

Example: download data through SUNGEO API

```
# Single country, single topic
out_15 <- get_data(country_name="Afghanistan",topics="Demographics:Population:GHS")
out_15

# Multiple countries, multiple topics
out_16 <- get_data(
	country_name=c("Albania","Moldova"),
	topics=c("Demographics:Ethnicity:EPR","Demographics:Population:GHS"))
out_16

# Other boundary sets, spatial and time units
out_17 <- get_data(
	country_name="Albania",
	topics="Weather:AirTemperatureAndPrecipitation:NOAA",
	geoset="GAUL",geoset_yr=1990,space_unit="adm2",time_unit="month",
	year_min=1990,year_max=1991)
out_17
```

Example: get list of available data through SUNGEO API

```
# Get list of all available data
out_18 <- get_info()
out_18["summary"]
out_18["topics"]
out_18["geosets"]

# Get list of available data for a single country
out_19 <- get_info(country_names="Afghanistan")
out_19["summary"]
out_19["topics"]
out_19["geosets"]

# Get list of available data for a single topic
out_20 <- get_info(topics="Elections:LowerHouse:CLEA")
out_20["summary"]
out_20["topics"]

# Get list of available data for a multiple countries and topics
out_21 <- get_info(
           country_names=c("Afghanistan","Zambia"),
           topics=c("Elections:LowerHouse:CLEA","Events:PoliticalViolence:GED"))
out_21["summary"]
```

Additional examples in help files of individual functions.

#' Automatically convert geographic (degree) to planar coordinates (meters)
#'
#' Function to automatically convert simple feature, spatial and raster objects with geographic coordinates (longitude, latitude / WGS 1984, EPSG:4326) to planar UTM coordinates. If the study region spans multiple UTM zones, defaults to Albers Equal Area.
#'
#' @param x Layer to be reprojected. \code{sf}, \code{sp}, \code{SpatRaster} or \code{RasterLayer} object.
#' @param max_zones Maximum number of UTM zones for single layer. Default is 5. Numeric.
#' @param return_list Return list object instead of reprojected layer (see Details). Default is \code{FALSE}. Logical.
#' @return Re-projected layer. \code{sf} or \code{RasterLayer} object, depending on input.
#'
#' If \code{return_list=TRUE}, returns a list object containing
#' \itemize{
##'  \item "x_out". The re-projected layer. \code{sf} or \code{RasterLayer} object, depending on input.
##'  \item "proj4_best".proj4string of the projection. Character string.
##'  }
#' @details Optimal map projection for the object \code{x} is defined by matching its horizontal extent with that of the 60 UTM zones. If object spans multiple UTM zones, uses either the median zone (if number of zones is equal to or less than \code{max_zones}) or Albers Equal Area projection with median longitude as projection center (if number of zones is greater than \code{max_zones}).
#' @importFrom sf st_bbox st_transform st_coordinates
#' @importFrom terra ext project crds
#' @importFrom stats median
#' @examples
#' # Find a planar projection for an unprojected (WSG 1984) hexagonal grid of Germany
#' data(hex_05_deu)
#' sf::st_crs(hex_05_deu)
#' out_1 <- utm_select(hex_05_deu)
#' sf::st_crs(out_1)
#'
#' # Find a planar projection for a raster
#' data(gpw4_deu2010)
#' gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
#' sf::st_crs(gpw4_deu2010)
#' out_2 <- utm_select(gpw4_deu2010)
#' sf::st_crs(out_2)
#'
#' @export

utm_select <- function(x, max_zones=5, return_list=FALSE){

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  # Fix sf geometry type
  if(any(grepl("sf", class(x)))){
    if(grepl("GEOMETRY",sf::st_geometry_type(x,by_geometry = FALSE))&any(grepl("POLYGON",sf::st_geometry_type(x)))){
      x <- sf::st_cast(x,"MULTIPOLYGON")
    }
    if(grepl("GEOMETRY",sf::st_geometry_type(x,by_geometry = FALSE))&any(grepl("LINE",sf::st_geometry_type(x)))){
      x <- sf::st_cast(x,"MULTILINESTRING")
    }
    if(grepl("GEOMETRY",sf::st_geometry_type(x,by_geometry = FALSE))&any(grepl("POINT",sf::st_geometry_type(x)))){
      x <- sf::st_cast(x,"POINT")
    }
  }

  # Extract bounding box & median x coordinate
  if(any(grepl("sp", attr(class(x), 'package')))){
    bb <- sf::st_bbox(x)
    median_x <- median(sf::st_coordinates(sf::st_as_sf(x))[,1],na.rm=T)
  }
  if(any(grepl("sf", class(x)))){
    bb <- sf::st_bbox(x)
    median_x <- median(sf::st_coordinates(x)[,"X"],na.rm=TRUE)
  }
  if(any(grepl("raster", attr(class(x), 'package')))) {
    x <- as(x,"SpatRaster")
  }
  if(any(grepl("terra", attr(class(x), 'package')))) {
    bb <- as.vector(terra::ext(x))[c("xmin","ymin","xmax","ymax")]
    median_x <- median(terra::crds(x)[,"x"],na.rm=TRUE)
  }

  # Conversion table
  UTM_Converter_Table <- data.frame(
    Begin = as.numeric(seq(-180, 174, by = 6)),
    End = as.numeric(seq(-174, 180, by = 6) - .000000001),
    UTM = as.numeric(seq(1, 60, by = 1)),
    CRS = paste0("+proj=utm +zone=", seq(1, 60, by = 1), " +datum=WGS84"),
    stringsAsFactors = FALSE)
  # Find UTM zone
  Location_UTM_Begin <- which(bb[3] >= UTM_Converter_Table$Begin)
  Location_UTM_End <- which(bb[1] <= UTM_Converter_Table$End)
  Location_UTM <- base::intersect(Location_UTM_Begin, Location_UTM_End)
  proj_out <- ifelse(length(Location_UTM) <= max_zones, as.character(UTM_Converter_Table$CRS[median(Location_UTM)]),paste0("+proj=aea +lon_0=", median_x," +lat_1=", bb[2]," +lat_2=",bb[4]))
  # Re-project
  if(any(grepl("sf", class(x)))){
    x_out <- sf::st_transform(x, proj_out)
  }
  if(any(grepl("terra", attr(class(x), 'package')))){
    x_out <- terra::project(x, y = proj_out)
  }
  if(any(grepl("sp", attr(class(x), 'package')))){
    suppressMessages({
      suppressWarnings({
        x_out <- as(sf::st_transform(sf::st_as_sf(x), proj_out),"Spatial")
      })
    })
  }
  # Output
  if (return_list == TRUE) {
    return(list(x_out = x_out, proj_out = proj_out))
  }
  if (return_list == FALSE) {
    return(x_out)
  }
}

#' Automatically calculate Local G hot spot intensity
#'
#' Function automatically calculates the Local G hot spot intensity measure for spatial points or spatial polygons. Uses RANN for efficient nearest neighbor calculation (spatial points only); users can specify the number of neighbors (k). Users can specify the neighborhood style (see spdep::nb2listw) with default being standardized weight matrix (W).
#'
#' @param insert Spatial point or spatial polygon object. Acceptable formats include \code{sf}, \code{SpatialPolygonsDataFrame}, \code{SpatialPointsDataFrame}.
#' @param variable Column name or numeric vector containing the variable from which the local G statistic will be calculated. Must possess a natural scale that orders small and large observations (i.e. number, percentage, ratio and not model residuals).
#' @param style Style can take values \code{'W'}, \code{'B'}, \code{'C'}, \code{'U'}, \code{'mimax'}, \code{'S'} (see  \code{\link[spdep]{nb2listw}}). Character string.
#' @param k Number of neighbors. Default is 9. Numeric.
#' @param remove_missing Whether to calculate statistic without missing values. If \code{FALSE}, substitute value must be supplied to \code{NA_Value}.
#' @param NA_Value Substitute for missing values. Default value is 0. Numeric.
#' @param include_Moran Calculate local Moran's I statistics. Default is \code{FALSE}. Logical.
#' @return If \code{input} is \code{sf}, \code{SpatialPolygonsDataFrame} or \code{SpatialPointsDataFrame} object, returns \code{sf} object with same geometries and columns as \code{input}, appended with additional column containing Local G estimates (\code{LocalG}). If \code{input} is \code{RasterLayer} object, returns \code{RasterBrick} object containing original values (\code{Original}) and Local G estimates (\code{LocalG}).
#' @importFrom sf st_as_sf st_is st_coordinates
#' @importFrom terra values as.data.frame plot
#' @importFrom spdep poly2nb nb2listw localG localmoran
#' @importFrom RANN nn2
#' @examples
#' # Calculate Local G for sf point layer
#' data(clea_deu2009_pt)
#' out_1 <- hot_spot(insert=clea_deu2009_pt, variable = clea_deu2009_pt$to1)
#' class(out_1)
#' plot(out_1["LocalG"])
#'
#' # Calculate Local G for sf polygon layer (variable as numeric vector)
#' data(clea_deu2009)
#' out_2 <- hot_spot(insert=clea_deu2009, variable = clea_deu2009$to1)
#' summary(out_2$LocalG)
#' plot(out_2["LocalG"])
#'
#' \donttest{
#' # Calculate Local G for sf polygon layer (variable as column name)
#' out_3 <- hot_spot(insert=clea_deu2009, variable = "to1")
#' summary(out_3$LocalG)
#' plot(out_3["LocalG"])
#' }
#'
#' \donttest{
#' # Calculate Local G for a SpatialPolygonsDataFrame object (variable as column name)
#' out_4 <- hot_spot(insert=as(clea_deu2009,"Spatial"), variable = "to1")
#' summary(out_4$LocalG)
#' plot(out_4["LocalG"])
#' }
#'
#' @export
hot_spot <- function(insert,
                      variable = NULL,
                      style = 'W',
                      k = 9,
                      remove_missing = TRUE,
                      NA_Value = 0,
                      include_Moran = FALSE){

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  #Section 1 -
  if(any(class(insert)%in%c("SpatialPolygonsDataFrame", "SpatialPointsDataFrame"))){
    insert <- sf::st_as_sf(insert)
  }

  #Section 2 -
  if(is.character(variable) && 'sf'%in%class(insert)){
    insert_df <- terra::as.data.frame(insert)

    variable <- insert_df[,names(insert_df)%in%variable]
    rm(insert_df)

    variable <- as.vector(variable)
  }

  #Section 3 -
  if('sf'%in%class(insert)){
    if(all(sf::st_is(insert, 'POLYGON')) | all(sf::st_is(insert, 'MULTIPOLYGON'))){
      #Part A -
      suppressMessages({suppressWarnings({
        NNobj <- spdep::poly2nb(insert)
      })})

      #Part B -
      suppressMessages({suppressWarnings({
        NNobj <- spdep::nb2listw(NNobj, style = style)
      })})

      #Part C -
      suppressMessages({suppressWarnings({
        LocalG_Output <- spdep::localG(variable, NNobj)
      })})

      if(include_Moran%in%TRUE){
        suppressMessages({suppressWarnings({
          LocalM_Output <- spdep::localmoran(variable, NNobj)[,1]
        })})
      }

      #Part D -
      insert$LocalG <- LocalG_Output

      if(include_Moran%in%TRUE){
        insert$LocalM <- LocalM_Output
      }

      #Part E -
      RANN_Act <- 0

      #RETURN
      return(insert)
    }
  }

  #Section 4 -
  if('RasterLayer'%in%class(insert)){
    #Part A -
    variable <- terra::values(insert)

    #Part B -
    IDxy_Matrix <- terra::as.data.frame(insert, xy = TRUE)

    #Part C -
    IDxy_Matrix <- IDxy_Matrix[,c(1:2)]

    #Part D -
    names(IDxy_Matrix) <- c('x', 'y')

    #Part E -
    Location_Missing <- which(is.na(variable))

    #Part F -
    if(TRUE%in%remove_missing && length(Location_Missing) > 0){
      IDxy_Matrix <- IDxy_Matrix[-Location_Missing,]

      variable <- variable[-Location_Missing]
    }

    #Part G -
    if(FALSE%in%remove_missing && length(Location_Missing) > 0){
      variable[Location_Missing] <- NA_Value
    }

    #Part H -
    RANN_Act <- 1

  }

  #Section 5 -
  if('sf'%in%class(insert)){
    if(all(sf::st_is(insert, 'POINT'))){
      #Part A -
      IDxy_Matrix <- terra::as.data.frame(sf::st_coordinates(insert), xy = TRUE)

      #Part B -
      Location_Missing <- which(is.na(variable))

      #Part C -
      if(TRUE%in%remove_missing && length(Location_Missing) > 0){
        IDxy_Matrix <- IDxy_Matrix[-Location_Missing,]
      }

      #Part D -
      if(FALSE%in%remove_missing && length(Location_Missing) > 0){
        variable[Location_Missing] <- NA_Value
      }

      #Part E -
      RANN_Act <- 1

    }
  }

  #Section 6 -
  if(RANN_Act == 1){
    #Part A -
    KNN <- RANN::nn2(IDxy_Matrix[,1:2], IDxy_Matrix[,1:2], k = k)

    #Part B -
    KNN <- lapply(1:nrow(KNN$nn.idx[,1:k]), function(x){KNN$nn.idx[x,1:k]})

    #Part C -
    class(KNN) <- 'nb'

    #Part D -
    suppressMessages({suppressWarnings({
      NNobj <- spdep::nb2listw(KNN, style = style)
    })})

    #Part E -
    suppressMessages({suppressWarnings({
      LocalG_Output <- spdep::localG(variable, NNobj)
    })})

    if(include_Moran%in%TRUE){
      suppressMessages({suppressWarnings({
        LocalM_Output <- spdep::localmoran(variable, NNobj)[,1]
      })})
    }
  }

  #Section 7 -
  if('sf'%in%class(insert)){
    if(all(sf::st_is(insert, 'POINT'))){
      #Part A -
      insert$LocalG <- NA

      #Part B -
      if(length(Location_Missing) > 0){
        insert$LocalG[-Location_Missing] <- as.vector(LocalG_Output)
      } else {
        insert$LocalG <- as.vector(LocalG_Output)
      }


      #RETURN
      return(insert)
    }
  }

  #Section 8 -
  if('RasterLayer'%in%class(insert)){
    #Part A -
    IDxy_Matrix <- terra::as.data.frame(insert, xy = TRUE)

    ##########################
    #Part B -
    ##########################
    IDxy_Matrix$LocalG <- NA

    #Part C -
    if(length(Location_Missing) > 0){
      IDxy_Matrix$LocalG[-Location_Missing] <- as.vector(LocalG_Output)
    } else {
      IDxy_Matrix$LocalG <- as.vector(LocalG_Output)
    }

    #Part D -
    TemporaryRaster <- insert

    #Part E -
    terra::values(TemporaryRaster) <- IDxy_Matrix$LocalG

    ##########################
    #Part C -
    ##########################
    if(include_Moran%in%TRUE){

      IDxy_Matrix$LocalM <- NA

      #Part C -
      if(length(Location_Missing) > 0){
        IDxy_Matrix$LocalM[-Location_Missing] <- as.vector(LocalM_Output)
      } else {
        IDxy_Matrix$LocalM <- as.vector(LocalM_Output)
      }

      #Part D -
      TemporaryRaster_Moran <- insert

      #Part E -
      terra::values(TemporaryRaster_Moran) <- IDxy_Matrix$LocalM


    }

    if(include_Moran%in%FALSE){
      #Part F -
      returnBrick <- c(insert, TemporaryRaster)

      #Part G -
      names(returnBrick) <- c('Original', 'LocalG')
    } else {
      #Part F -
      returnBrick <- c(insert, TemporaryRaster, TemporaryRaster_Moran)

      #Part G -
      names(returnBrick) <- c('Original', 'LocalG', 'Moran')
    }


    #RETURN
    return(returnBrick)
  }

  gc()
}

#' Merge list of tables on common variable(s)
#'
#' Function that finds a set of common columns in a list of tables, and merges the tables on these columns.
#'
#' @param lst List of tables to be merged. List object.
#' @return \code{data.table} object
#' @importFrom data.table data.table as.data.table
#' @examples
#' # Merge list of three tables with different common variables
#' A <- data.table::data.table(month=month.name,year=rep(1991:1992,each=12),A=rnorm(24))
#' B <- data.table::data.table(year=c(1991,1992),B=rbeta(2,1,1))
#' C <- data.table::data.table(month=month.name,C=runif(12))
#'
#' out_1 <- merge_list(list(A,B,C))
#' out_1
#'
#' @export

merge_list <- function(lst){
  while(length(lst) > 1) {
    # Create index of list objects
    idxlst <- seq(from=1, to=length(lst), by=2)
    # Iterate over pairs of list objects
    lst <- lapply(idxlst, function(i){
      if(i==length(lst)){ return(lst[[i]]) }
      # Identify common variables
      cmnvar <- intersect(names(lst[[i]]),names(lst[[i+1]]))
      # Merge tables
      out <- data.table::as.data.table(merge(lst[[i]], lst[[i+1]],by=cmnvar,all=TRUE,allow.cartesian=TRUE))
      # Drop duplicate columns
      out <- out[,.SD,.SDcols=unique(names(out))]
      return(out)
    })
  }
  return(lst[[1]])
}

#' Fix polygon geometries
#'
#' Function to check validity and fix broken geometries in simple features polygon objects
#'
#' @param x Polygon layer to be checked and fixed. \code{sf} object.
#' @param n_it Number of iterations. Default is 10. Numeric..
#' @return Returns a \code{sf} polygon object, with self-intersections and other geometry problems fixed.
#' @importFrom sf st_make_valid st_is_valid st_buffer st_cast
#' @examples
#' # Fix geometries for a single dataset
#' data(clea_deu2009)
#' out_1 <- fix_geom(clea_deu2009)
#' all(sf::st_is_valid(out_1))
#' @export

fix_geom <- function(x, n_it = 10){
  #Part 1 -
  ValidityVector <- sf::st_is_valid(x)

  #Part 2 -
  countIter <- 0

  while(all(ValidityVector) == FALSE || countIter < n_it){
    suppressMessages({
      suppressWarnings({
        x <- sf::st_buffer(x,dist=0)
      })
    })

    ValidityVector <- sf::st_is_valid(x)

    countIter <- 1 + countIter

    if(all(ValidityVector)){
      break
    }

    if(all(ValidityVector) == FALSE && countIter == n_it){
      stop('Non-Convergence for Polygon; Please increase number of iterations or check to see whether polygon is corrupted')
    }

  }

  message(paste0('Completed ', countIter, ' Round of Buffering'))

  # Simplify geometry type
  x <- sf::st_cast(x)

  return(x)
}

#' Point-to-polygon interpolation, tessellation method
#'
#' Function for interpolating values from a source point layer to a destination polygon layer, using Voronoi tessellation and area/population weights.
#'
#' @param pointz Source points layer. \code{sf} object.
#' @param polyz Destination polygon layer. Must have identical CRS to \code{pointz}. \code{sf} object.
#' @param poly_id Name of unique ID column for destination polygon layer. Character string.
#' @param methodz Interpolation method(s) for numeric covariates. Could be either of "aw" (areal weighting, default) and/or "pw" (population weighting). See "details". Character string or vector of character strings.
#' @param char_methodz Interpolation method(s) for character strings. Could be either of "aw" (areal weighting, default) or "pw" (population weighting). See "details". Character string.
#' @param pop_raster Population raster to be used for population weighting, Must be supplied if \code{methodz="pw"}. Must have identical CRS to \code{poly_from}. \code{raster} or \code{SpatRaster} object.
#' @param varz Names of numeric variable(s) to be interpolated from source polygon layer to destination polygons. Character string or list of character strings.
#' @param funz Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a numeric vector \code{x} and vector of weights \code{w}. Function or list of functions.
#' @param pycno_varz Names of spatially extensive numeric variables for which the pycnophylactic (mass-preserving) property should be preserved. Character string or vector of character strings.
#' @param char_varz  Names of character string variables to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.
#' @param char_assign Assignment rule to be used for variables specified in \code{char_varz}. Could be either "biggest_overlap" (default) or "all_overlap". See "details". Character string or vector of character strings.
#' @param return_tess Return Voronoi polygons, in addition to destinaton polygon layer? Default is \code{FALSE}. Logical.
#' @param seed Seed for generation of random numbers. Default is 1. Numeric.
#' @return If \code{return_tess=FALSE}, returns a \code{sf} polygon object, with variables from \code{pointz} interpolated to the geometries of \code{polyz}.
#'
#' If \code{return_tess=TRUE}, returns a list, containing
#' \itemize{
##'  \item "result". The destination polygon layer. \code{sf} object.
##'  \item "tess". The (intermediate) Voronoi tessellation polygon layer. \code{sf} object.
##'  }
#' @details
#' This function interpolates point data to polygons with a two-step process. In the first step (tessellation), each point is assigned a Voronoi cell, drawn such that (a) the distance from its borders to the focal point is less than or equal to the distance to any other point, and (b) no gaps between cells remain. The second step (interpolation) performs a polygon-in-polygon interpolation, using the Voronoi cells as source polygons.
#'
#' Currently supported integration methods in the second step (\code{methodz}) include:
#' \itemize{
##'  \item Areal weighting ("aw"). Values from \code{poly_from} weighted in proportion to relative area of spatial overlap between source features and geometries of \code{poly_to}.
##'  \item Population weighting ("pw"). Values from \code{poly_from} weighted in proportion to relative population sizes in areas of spatial overlap between source features and geometries of \code{poly_to}. This routine uses a third layer (supplied in \code{pop_raster}) to calculate the weights.
##' }
#' When a list of variables are supplied and one methods argument specified, then the chosen method will be applied to all variables.
#'
#' When a list of of variables are supplied and multiple methods arguments specified, then weighting methods will be applied in a pairwise order. For example, specifying \code{varz = list(c("to1","pvs1_margin"), c("vv1"))} and \code{methodz = c('aw', 'pw')} will apply areal weighting to the the first set of variables (to1 and pvs1_margin) and population weighing to the second set (vv1).
#'
#' Interpolation procedures are handled somewhat differently for numeric and character string variables. For numeric variables supplied in \code{varz}, "aw" and/or "pw" weights are passed to the function specified in \code{funz}. If different sets of numeric variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).
#'
#' For character string (and any other) variables supplied in \code{char_varz}, "aw" and/or "pw" weights are passed to the assignment rule(s) specified in \code{char_assign}. Note that the \code{char_varz} argument may include numerical variables, but \code{varz} cannot include character string variables.
#'
#' Currently supported assignment rules for character strings (\code{char_assign}) include:
#' \itemize{
##'  \item "biggest_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned a single value from overlapping \code{poly_from} features, corresponding to the intersection with largest area and/or population weight.
##'  \item "all_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned all values from overlapping \code{poly_from} features, ranked by area and/or population weights (largest-to-smallest) of intersections.
##' }
#' It is possible to pass multiple arguments to \code{char_assign} (e.g. \code{char_assign=c("biggest_overlap","all_overlap")}), in which case the function will calculate both, and append the resulting columns to the output.
#' @import sf
#' @importFrom stats as.dist weighted.mean
#' @importFrom data.table data.table rbindlist as.data.table setnames
#' @importFrom terra extract
#' @importFrom methods as
#' @importFrom rmapshaper ms_dissolve
#' @importFrom dplyr select bind_cols left_join
#' @importFrom purrr reduce
#' @examples
#' # Interpolation of a single variable, with area weights
#' data(hex_05_deu)
#' data(clea_deu2009_pt)
#' out_1 <- point2poly_tess(pointz = clea_deu2009_pt,
#'                              polyz = hex_05_deu,
#'                              poly_id = "HEX_ID",
#'                              varz = "to1")
#' plot(out_1["to1_aw"])
#'
#' # Extract and inspect tessellation polygons
#' out_2 <- point2poly_tess(pointz = clea_deu2009_pt,
#'                              polyz = hex_05_deu,
#'                              poly_id = "HEX_ID",
#'                              varz = "to1",
#'                              return_tess = TRUE)
#' plot(out_2$tess["to1"])
#' plot(out_2$result["to1_aw"])
#'
#' \donttest{
#' # Interpolation of multiple variables, with area and population weights
#' data(gpw4_deu2010)
#' gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
#' out_3 <- point2poly_tess(pointz = clea_deu2009_pt,
#'                          polyz = hex_05_deu,
#'                          poly_id = "HEX_ID",
#'                          methodz = c("aw","pw"),
#'                          varz = list(
#'                            c("to1","pvs1_margin"),
#'                            c("vv1")
#'                          ),
#'                          pycno_varz = "vv1",
#'                          funz = list(
#'                            function(x,w){stats::weighted.mean(x,w)},
#'                            function(x,w){sum(x*w)}
#'                            ),
#'                          char_varz = c("incumb_pty_n","win1_pty_n"),
#'                          pop_raster = gpw4_deu2010)
#' plot(out_3["vv1_pw"])
#' }
#' @export


point2poly_tess <- function(
  pointz,
  polyz,
  poly_id,
  char_methodz = "aw",
  methodz="aw",
  pop_raster=NULL,
  varz=NULL,
  pycno_varz=NULL,
  char_varz=NULL,
  char_assign="biggest_overlap",
  funz=function(x,w){stats::weighted.mean(x,w,na.rm=TRUE)},
  return_tess=FALSE,
  seed = 1){

  set.seed(seed)

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  ###########################################################
  #Section A - Preparing the Gemetries for Tessalation Process
  ###########################################################


  # Put variables and functions into list
  if(inherits(varz,"character")){varz <- list(varz)}
  if(inherits(funz,"function")){funz <- list(funz)}

  # Stop if no population raster
  if("pw"%in%methodz & length(pop_raster)==0){stop("No population raster provided.")}

  ###########################################################
  #Section B - Preparing the Gemetries for Tessalation Process
  ###########################################################
  ############################
  #Part 1 - Create union layer
  ###########################
  suppressWarnings({
    suppressMessages({
      # polyz_u <- dplyr::select(fix_geom(rmapshaper::ms_dissolve(polyz)),-1)
      polyz_u <- fix_geom(sf::st_union(polyz))
    })
  })

  #####################################
  #Part 2 -  Jitter and crop by polygon
  #####################################
  #Part i -
  suppressWarnings({
    suppressMessages(
      pointz_crop <-  sf::st_crop(sf::st_jitter(pointz),sf::st_bbox(polyz_u))
    )
  })

  #Part ii -
  pointz_crop$Unique_ID <- 1:nrow(pointz_crop)

  #
  #

  ###############################
  #Part 3 - Convert to multipoint
  ###############################
  suppressWarnings({
    suppressMessages(
      pointz_geom <- sf::st_union(sf::st_geometry(pointz_crop))
    )
  })

  #
  #

  #################################
  #Part 4 - Create Voronoi polygons
  #################################
  suppressWarnings({
    geo_vor <- suppressMessages(
      sf::st_as_sf(data.table::setnames(data.table::as.data.table(sf::st_intersection(sf::st_as_sf(sf::st_cast(sf::st_voronoi(pointz_geom))),polyz_u)), old="x",new="geometry"))
    )
  })

  #
  #

  ###############################################
  #Part 5 - Exception for single-point geometries
  ###############################################
  if(nrow(pointz_crop)==1 | length(unique(st_geometry(pointz_crop)))==1){
    suppressWarnings({
      suppressMessages({
        geo_vor <- sf::st_as_sf(data.frame(x=NA,geometry=sf::st_geometry(polyz_u)))
      })
    })
  }

  #
  #

  ###############################################
  #Part 6 - Combine with point feature attributes
  ###############################################
  #Part i -
  suppressWarnings({
    suppressMessages(
      int <- sf::st_intersects(geo_vor,pointz_crop)
    )
  })
  Errors <- sapply(int, function(x) length(x))

  #Part ii -
  geo_vor$Unique_ID <- NA
  geo_vor$Unique_ID[Errors == 1] <- pointz_crop$Unique_ID[unlist(int[Errors == 1])]

  #
  #

  #############################
  #Part 7 - Crop the Geometries
  #############################
  suppressWarnings({
    suppressMessages({
      geo_vor <-  fix_geom(sf::st_intersection(geo_vor,polyz_u))
    })
  })

  #
  #

  ################################
  #Part 8 - Merge with Point Layer
  ################################
  pointz_crop_dt <- pointz_crop

  sf::st_geometry(pointz_crop_dt) <- NULL

  suppressWarnings({
    suppressMessages({
      geo_vor <- dplyr::left_join(geo_vor, pointz_crop_dt, by = 'Unique_ID')
    })
  })

  #
  #
  #
  #

  ###########################################################
  #Section C - Preparing the Gemetries for Tessalation Prcess
  ###########################################################
  # Poly-in-poly
  polyz_ <- poly2poly_ap(
    poly_from = geo_vor,
    poly_to = polyz,
    poly_to_id = poly_id,
    geo_vor = geo_vor,
    methodz = methodz,
    pop_raster = pop_raster,
    varz = varz,
    char_methodz = char_methodz,
    pycno_varz = pycno_varz,
    char_varz = char_varz,
    char_assign = char_assign,
    funz = funz,
    seed = seed
  )


  # Output
  polyz_$Unique_ID <- NULL; geo_vor$Unique_ID <- NULL
  if(!return_tess){return(polyz_)}
  if(return_tess){return(list(result=polyz_,tess=geo_vor))}

  #Equal Area Projection as a Default

}

#' Make date ticker
#'
#' Function to create a table of consecutive dates, in SUNGEO-compliant format.
#'
#' @param date_min Start date, in YYYYMMDD format. Default is \code{19000101}. Integer.
#' @param date_max End date, in YYYYMMDD format. Default is today. Integer.
#' @return \code{data.table} object, with seven columns:
#' \itemize{
##'  \item \code{DATE}. Date in YYYYMMDD format. Integer.
##'  \item \code{DATE_ALT}. Date in \code{Date} (YYYY-MM-DD) format. Date.
##'  \item \code{TID}. Date ID, in consecutive integer format. Integer.
##'  \item \code{YRWK}. Week in YYYYWW format. Integer.
##'  \item \code{WID}. Weed ID, in consecutive integer format. Integer.
##'  \item \code{YRMO}. Month in YYYYMM format. Integer.
##'  \item \code{MID}. Month ID, in consecutive integer format. Integer.
##'  \item \code{YEAR}. Year in YYYY format. Integer.
##' }
#' @importFrom data.table data.table as.data.table
#' @examples
#' # All dates from January 1, 1900 to today
#' out_1 <- make_ticker()
#' out_1
#'
#' # All dates from January 1, 1200 to today
#' out_2 <- make_ticker(date_min=12000101)
#' out_2
#'
#' # All dates from January 1, 1500 to December 31, 1899
#' out_3 <- make_ticker(date_min=15000101, date_max=18991231)
#' out_3
#'
#' @export


make_ticker <- function(
  date_min=19000101,
  date_max=as.integer(gsub("-","",as.Date(Sys.Date())))
  ){
  # Convert to "Date" format
  date_min_ <- paste0(substr(date_min,1,4),"-",substr(date_min,5,6),"-",substr(date_min,7,8))
  date_max_ <- paste0(substr(date_max,1,4),"-",substr(date_max,5,6),"-",substr(date_max,7,8))
  # Create time table
  if(date_max>=19000101){
    ticker <- data.table::data.table(DATE_ALT =  seq(as.Date("1900-01-01"), as.Date(date_max_), by="days"))[,DATE:=as.integer(gsub("-","",DATE_ALT))][,TID:=1:.N][,WID:=rep(1:.N,each=7)[1:.N]][,YRMO:=as.integer(substr(gsub("-","",DATE),1,6))][,MID:=as.integer(as.factor(YRMO))][,YEAR:=as.integer(substr(gsub("-","",DATE),1,4))]
  }
  # Combine with pre-1900 dates (if applicable)
  if(date_min<19000101&date_max>=19000101){
    ticker2 <- data.table(DATE_ALT =  seq(as.Date(date_min_), as.Date("1899-12-31"), by="days"))[,DATE:=as.integer(gsub("-","",DATE_ALT))][,TID:=1-(.N:1)][,WID:=as.integer(1-rev(rep(1:.N,each=7)[1:.N]))][,YRMO:=as.integer(substr(gsub("-","",DATE),1,6))][,MID:=as.integer(1-as.integer(factor(YRMO,levels=rev(unique(YRMO)))))][,YEAR:=as.integer(substr(gsub("-","",DATE),1,4))]
    ticker <- data.table::as.data.table(rbind(ticker2,ticker))[DATE>=date_min&DATE<=date_max]
    rm(ticker2)
  }
  if(date_min<19000101&date_max<19000101){
    ticker <- data.table(DATE_ALT =  seq(as.Date(date_min_), as.Date("1899-12-31"), by="days"))[,DATE:=as.integer(gsub("-","",DATE_ALT))][,TID:=1-(.N:1)][,WID:=as.integer(1-rev(rep(1:.N,each=7)[1:.N]))][,YRMO:=as.integer(substr(gsub("-","",DATE),1,6))][,MID:=as.integer(1-as.integer(factor(YRMO,levels=rev(unique(YRMO)))))][,YEAR:=as.integer(substr(gsub("-","",DATE),1,4))][DATE>=date_min&DATE<=date_max]
  }
  # Add week codes, re-order columns
  ticker <- data.table::setkey(merge(ticker,data.table::setnames(ticker[,unique(YEAR),by=WID],"V1","YEAR")[!duplicated(WID,fromLast=TRUE)][, YRWK := match(WID, unique(WID)) + (YEAR*1e3), by=YEAR][,YEAR := NULL] ,by="WID")[DATE>=date_min&DATE<=date_max][,.(DATE,DATE_ALT,TID,YRWK,WID,YRMO,MID,YEAR)],TID)
  return(ticker)
}

#' @importFrom httr handle config status_code headers GET warn_for_status
#' @importFrom rlang abort

## html_sessionSunGeo Function
## Function that replaces rvest::html_session()

request_GET <- function(x, url, ...) {
  x$response <- httr::GET(url, x$config, ..., handle = x$handle)
  x$html <- new.env(parent = emptyenv(), hash = FALSE)
  x$url <- x$response$url
  httr::warn_for_status(x$response)
  x
}

html_sessionSunGeo <- function(url, ...) {
  session <- structure(
    list(
      handle   = httr::handle(url),
      config   = c(..., httr::config(autoreferer = 1L)),
      url      = NULL,
      back     = character(),
      forward  = character(),
      response = NULL,
      html     = new.env(parent = emptyenv(), hash = FALSE)
    ),
    class = "session"
  )
  request_GET(session, url)
}


#' @exportS3Method print session
print.session <- function(x, ...) {
  cat("<session> ", x$url, "\n", sep = "")
  cat("  Status: ", httr::status_code(x$response), "\n", sep = "")
  cat("  Type:   ", httr::headers(x)$`Content-Type`, "\n", sep = "")
  cat("  Size:   ", length(x$response$content), "\n", sep = "")
}




# ## udconvertSunGeo Function
# ## Function that replaces udunits2::udconvert()
# udconvertSunGeo <-
#   function(x, u1, u2) {
#     if (! ud.are.convertible(u1, u2)) {
#       stop(paste("Units", u1, "and", u2, "are not convertible"))
#     }
#     ## Filter out NA's before passing them to the C function
#     ## since it can't handle them
#     rv <- rep(NA, length(x))
#     i <- which(! is.na(x))
#
#     len <- length(i)
#     c.rv <- .C('R_ut_convert',
#                as.double(x)[i],
#                as.integer(len),
#                as.character(u1),
#                as.character(u2),
#                converted=double(len)
#     )
#     rv[i] <- c.rv$converted
#     ## If it's a matrix/vector or anything else, convert it back to it's original type
#     attributes(rv) <- attributes(x)
#     return(rv)
#   }
#
# .onLoad <- function(libname, pkgname) {
#   ## By default, configure udunits with path set (presumably) by the
#   ## user through the UDUNITS2_XML_PATH environment variable
#   .C('R_ut_init', as.integer(0))
#   if (!ud.have.unit.system()) {
#     ## Failing that, override it with the in-package XML file
#     p0 <- system.file("share/udunits2.xml", package="udunits2")
#     Sys.setenv(UDUNITS2_XML_PATH=p0)
#     .C('R_ut_init', as.integer(1))
#     ## If *that* fails, give the user some instructions for how to remedy
#     ## the problem
#     if (!ud.have.unit.system()) {
#       warning(
#         "Failed to read udunits system database: udunits2 will not work properly.\n Please set the UDUNITS2_XML_PATH environment variable before attempting to load the package")
#     }
#   }
# }
#
# .onAttach <- function(libname, pkgname) {
#   msg <- "udunits system database read"
#   p0 <- Sys.getenv("UDUNITS2_XML_PATH")
#   if (p0 != "") {
#     msg <- paste(msg, "from", p0)
#   }
#   packageStartupMessage(msg)
# }
#
# ud.are.convertible <-
#   function(u1, u2) {
#     if (! (ud.is.parseable(u1) && ud.is.parseable(u2))) {
#       return(FALSE)
#     }
#     rv <- .C('R_ut_are_convertible',
#              as.character(u1),
#              as.character(u2),
#              convertible=logical(1))
#     return(rv$convertible)
#   }
#
# ud.get.name <-
#   function(unit.string) {
#     stopifnot(ud.is.parseable(unit.string))
#     rv <- .C('R_ut_get_name',
#              as.character(unit.string),
#              ud.name=character(length=1))
#     return(rv$ud.name)
#   }
#
# ud.get.symbol <-
#   function(unit.string) {
#     stopifnot(ud.is.parseable(unit.string))
#     rv <- .C('R_ut_get_symbol',
#              as.character(unit.string),
#              ud.symbol=character(length=1))
#     return(rv$ud.symbol)
#   }
#
# ud.is.parseable <-
#   function(unit.string) {
#     rv <- .C('R_ut_is_parseable',
#              as.character(unit.string),
#              parseable=logical(1))
#     return(rv$parseable)
#   }
#
# ud.set.encoding <-
#   function(enc.string) {
#     .C('R_ut_set_encoding',
#        as.character(enc.string))
#     return()
#   }
#
# ud.have.unit.system <-
#   function() {
#     rv <- .C('R_ut_has_system',
#              exists=logical(1))
#     return(rv$exists)
#   }



## all_ofSunGeo Function
## Function that replaces tidyselect::all_of()
## We utilize the function from tidyselec::all_of(), which calls vctrs::vec_as_location(), and ellipsis::check_dots_empty()
all_ofSunGeo <- function (x, ...)
{
  if (is.function(x)) {
    if (!missing(...))
      dots <- dots(parent.frame())
    if (length(dots) == 0) {
      return()
    }
    action_dots(action = abort, message = "`...` is not empty.",
                dot_names = names(dots), note = "These dots only exist to allow future extensions and should be empty.",
                .subclass = "rlib_error_dots_nonempty")
    .Call(vctrs_as_location, i = x, n = 0L, names = NULL, loc_negative = "invert",
          loc_oob = "error", loc_zero = "remove", missing =  c("propagate", "error"),
          arg = NULL)
    rlang::abort("Internal error: `all_of()` should have failed sooner")
  }
  x
}



# #
# ## fromJSONSunGeo Function
# ## Function that replaces jsonlite::fromJSON()

# fromJSONSunGeo <- function(txt, simplifyVector = TRUE, simplifyDataFrame = simplifyVector,
#                            simplifyMatrix = simplifyVector, flatten = FALSE, ...) {

#   # check type
#   if (!is.character(txt) && !inherits(txt, "connection")) {
#     stop("Argument 'txt' must be a JSON string, URL or file.")
#   }

#   # overload for URL or path
#   if (is.character(txt) && length(txt) == 1 && nchar(txt, type="bytes") < 2084 && !validate(txt)) {
#     if (grepl("^https?://", txt, useBytes=TRUE)) {
#       loadpkg("curl")
#       h <- curl::new_handle(useragent = paste("jsonlite /", R.version.string))
#       curl::handle_setheaders(h, Accept = "application/json, text/*, */*")
#       txt <- curl::curl(txt, handle = h)
#     } else if (file.exists(txt)) {
#       # With files we can never know for sure the encoding. Lets try UTF8 first.
#       # txt <- raw_to_json(readBin(txt, raw(), file.info(txt)$size));
#       txt <- file(txt)
#     }
#   }

#   # call the actual function (with deprecated arguments)
#   parse_and_simplify(txt = txt, simplifyVector = simplifyVector, simplifyDataFrame = simplifyDataFrame,
#                      simplifyMatrix = simplifyMatrix, flatten = flatten, ...)
# }

# parse_and_simplify <- function(txt, simplifyVector = TRUE, simplifyDataFrame = simplifyVector,
#                                simplifyMatrix = simplifyVector, flatten = FALSE, unicode = TRUE, validate = TRUE, bigint_as_char = FALSE, ...){

#   if(!missing(unicode)){
#     message("Argument unicode has been deprecated. YAJL always parses unicode.")
#   }

#   if(!missing(validate)){
#     message("Argument validate has been deprecated. YAJL automatically validates json while parsing.")
#   }

#   # parse
#   obj <- parseJSON(txt, bigint_as_char)

#   # post processing
#   if (any(isTRUE(simplifyVector), isTRUE(simplifyDataFrame), isTRUE(simplifyMatrix))) {
#     return(simplify(obj, simplifyVector = simplifyVector, simplifyDataFrame = simplifyDataFrame,
#                     simplifyMatrix = simplifyMatrix, flatten = flatten, ...))
#   } else {
#     return(obj)
#   }
# }


# # Backward compatiblity
# fromJSON_string <- parse_and_simplify





# parseJSON <- function(txt, bigint_as_char = FALSE) {
#   if(inherits(txt, "connection")){
#     parse_con(txt, bigint_as_char)
#   } else {
#     parse_string(txt, bigint_as_char)
#   }
# }

# #' @useDynLib SUNGEO R_parse
# parse_string <- function(txt, bigint_as_char){
#   if (length(txt) > 1) {
#     txt <- paste(txt, collapse = "\n")
#   }
#   .Call("R_parse", txt, bigint_as_char)
# }


# #simplify (from JSON)
# simplify <- function(x, simplifyVector = TRUE, simplifyDataFrame = TRUE, simplifyMatrix = TRUE,
#                      simplifyDate = simplifyVector, homoList = TRUE, flatten = FALSE, columnmajor = FALSE,
#                      simplifySubMatrix = simplifyMatrix) {

#   #This includes '[]' and '{}')
#   if (!is.list(x) || !length(x)) {
#     return(x)
#   }

#   # list can be a dataframe recordlist
#   if (isTRUE(simplifyDataFrame) && is.recordlist(x)) {
#     mydf <- simplifyDataFrame(x, flatten = flatten, simplifyMatrix = simplifySubMatrix)
#     if(isTRUE(simplifyDate) && is.data.frame(mydf) && is.datelist(mydf)){
#       return(parse_date(mydf[["$date"]]))
#     }
#     return(mydf)
#   }

#   # or a scalar list (atomic vector)
#   if (isTRUE(simplifyVector) && is.null(names(x)) && is.scalarlist(x)) {
#     return(list_to_vec(x))
#   }

#   # apply recursively
#   out <- lapply(x, simplify, simplifyVector = simplifyVector, simplifyDataFrame = simplifyDataFrame,
#                 simplifyMatrix = simplifySubMatrix, columnmajor = columnmajor, flatten = flatten)

#   # fix for mongo style dates turning into scalars *after* simplifying
#   # only happens when simplifyDataframe=FALSE
#   if(isTRUE(simplifyVector) && is.scalarlist(out) && all(vapply(out, inherits, logical(1), "POSIXt"))){
#     return(structure(list_to_vec(out), class=c("POSIXct", "POSIXt")))
#   }

#   # test for matrix. Note that we have to take another look at x (before
#   # list_to_vec on its elements) to differentiate between matrix and vector.
#   if (isTRUE(simplifyMatrix) && isTRUE(simplifyVector) && is.matrixlist(out) && all(unlist(vapply(x, is.scalarlist, logical(1))))) {
#     if(isTRUE(columnmajor)){
#       return(do.call(cbind, out))
#     } else {
#       #this is currently the default
#       return(do.call(rbind, out))
#     }
#   }

#   # Simplify higher arrays
#   if (isTRUE(simplifyMatrix) && is.arraylist(out)){
#     if(isTRUE(columnmajor)){
#       return(array(
#         data = do.call(cbind, out),
#         dim = c(dim(out[[1]]), length(out))
#       ));
#     } else {
#       #this is currently the default
#       return(array(
#         data = do.call(rbind, lapply(out, as.vector)),
#         dim = c(length(out), dim(out[[1]]))
#       ));
#     }
#   }

#   # try to enfoce homoList on unnamed lists
#   if (isTRUE(homoList) && is.null(names(out))) {
#     # coerse empty lists, caused by the ambiguous fromJSON('[]')
#     isemptylist <- vapply(out, identical, logical(1), list())
#     if (any(isemptylist) & !all(isemptylist)) {
#       # if all the others look like data frames, coerse to data frames!
#       if (all(vapply(out[!isemptylist], is.data.frame, logical(1)))) {
#         for (i in which(isemptylist)) {
#           out[[i]] <- data.frame()
#         }
#         return(out)
#       }

#       # if all others look like atomic vectors, unlist all
#       if (all(vapply(out[!isemptylist], function(z) {
#         isTRUE(is.vector(z) && is.atomic(z))
#       }, logical(1)))) {
#         for (i in which(isemptylist)) {
#           out[[i]] <- vector(mode = typeof(out[[which(!isemptylist)[1]]]))
#         }
#         return(out)
#       }
#     }
#   }

#   # convert date object
#   if( isTRUE(simplifyDate) && is.datelist(out) ){
#     return(parse_date(out[["$date"]]))
#   }

#   # return object
#   return(out)
# }

# is.matrixlist <- function(x) {
#   isTRUE(is.list(x)
#          && length(x)
#          && is.null(names(x))
#          && all(vapply(x, is.atomic, logical(1)))
#          && all.identical(vapply(x, length, integer(1)))
#          #&& all.identical(vapply(x, mode, character(1))) #this fails for: [ [ 1, 2 ], [ "NA", "NA" ] ]
#   );
# }

# is.arraylist <- function(x) {
#   isTRUE(is.list(x)
#          && length(x)
#          && is.null(names(x))
#          && all(vapply(x, is.array, logical(1)))
#          && all.identical(vapply(x, function(y){paste(dim(y), collapse="-")}, character(1)))
#   );
# }

# is.datelist <- function(x){
#   isTRUE(is.list(x)
#          && identical(names(x), "$date")
#          && (is.numeric(x[["$date"]]) || is.character(x[["$date"]]))
#   );
# }

# parse_date <- function(x){
#   if(is.numeric(x)){
#     return(structure(x/1000, class=c("POSIXct", "POSIXt")))
#   } else if(is.character(x)) {
#     #tz is not vectorized, so assume all() are the same
#     is_utc <- ifelse(all(grepl("Z$", x)), "UTC", "")
#     return(as.POSIXct(strptime(x, format = '%Y-%m-%dT%H:%M:%OS', tz = is_utc)))
#   } else {
#     return(x)
#   }
# }

# all.identical <- function(x){
#   if(!length(x)) return(FALSE)
#   for(i in x){
#     if(x[1] != i) return(FALSE)
#   }
#   return(TRUE)
# }


# #is.recordlist (from JSON)
# is_recordlist_c <- function(x){
#   .Call("C_is_recordlist", x)
# }

# is_recordlist_r <- function(x) {
#   if (!(is.unnamedlist(x) && length(x))) {
#     return(FALSE)
#   }
#   at_least_one_object = FALSE
#   for(i in x){
#     if(!(is.namedlist(i) || is.null(i))) return(FALSE)
#     if(!at_least_one_object && is.namedlist(i)) at_least_one_object <- TRUE
#   }
#   return(at_least_one_object)
# }

# is.recordlist <- is_recordlist_c;

# is.namedlist <- function(x) {
#   isTRUE(is.list(x) && !is.null(names(x)))
# }

# is.unnamedlist <- function(x) {
#   isTRUE(is.list(x) && is.null(names(x)))
# }


# #simplifyDataFrame (from JSON)
# simplifyDataFrame <- function(recordlist, columns, flatten, simplifyMatrix) {

#   # no records at all
#   if (!length(recordlist)) {
#     if (!missing(columns)) {
#       return(as.data.frame(matrix(ncol = length(columns), nrow = 0, dimnames = list(NULL,
#                                                                                     columns))))
#     } else {
#       return(data.frame())
#     }
#   }

#   # only empty records and unknown columns
#   if (!any(vapply(recordlist, length, integer(1), USE.NAMES = FALSE)) && missing(columns)) {
#     return(data.frame(matrix(nrow = length(recordlist), ncol = 0)))
#   }

#   # find columns if not specified
#   if (missing(columns)) {
#     columns <- unique(unlist(lapply(recordlist, names), recursive = FALSE, use.names = FALSE))
#   }

#   # Convert row lists to column lists. This is the heavy lifting
#   # columnlist <- lapply(columns, function(x) lapply(recordlist, "[[", x))
#   # Now slighlty optimized
#   columnlist <- transpose_list(recordlist, columns)

#   # simplify vectors and nested data frames
#   columnlist <- lapply(columnlist, simplify, simplifyVector = TRUE, simplifyDataFrame = TRUE,
#                        simplifyMatrix = FALSE, simplifySubMatrix = simplifyMatrix, flatten = flatten)

#   # check that all elements have equal length
#   columnlengths <- unlist(vapply(columnlist, function(z) {
#     ifelse(length(dim(z)) > 1, nrow(z), length(z))
#   }, integer(1)))
#   n <- unique(columnlengths)
#   if (length(n) > 1) {
#     stop("Elements not of equal length: ", paste(columnlengths, collapse = " "))
#   }

#   # add the column names before flattening
#   names(columnlist) <- columns

#   # flatten nested data frames
#   if(isTRUE(flatten)) {
#     dfcolumns <- vapply(columnlist, is.data.frame, logical(1))
#     if(any(dfcolumns)){
#       columnlist <- c(columnlist[!dfcolumns], do.call(c, columnlist[dfcolumns]))
#     }
#   }

#   # make into data frame
#   class(columnlist) <- "data.frame"

#   # set row names
#   if("_row" %in% names(columnlist)) {
#     rn <- columnlist[["_row"]];
#     columnlist["_row"] <- NULL;

#     # row.names() casts double to character which is undesired.
#     if(is.double(rn)) {
#       rn <- as.integer(rn);
#     }

#     # Replace missing values with numbers
#     rn_na <- is.na(rn)
#     if(sum(rn_na) > 0){
#       rn[rn_na] <- paste0("NA_", seq_len(sum(rn_na)))
#     }

#     # data frames MUST have row names
#     if(any(duplicated(rn))){
#       warning('Duplicate names in "_row" field. Data frames must have unique row names.', call. = FALSE)
#       if(is.character(rn)) {
#         row.names(columnlist)  <- make.unique(rn)
#       } else {
#         row.names(columnlist) <- seq_len(n)
#       }
#     } else {
#       row.names(columnlist) <- rn;
#     }
#   } else {
#     row.names(columnlist) <- seq_len(n)
#   }

#   return(columnlist)
# }


# #is.scalarlist (from JSON)
# is_scalarlist_r <- function(x) {
#   if(!is.list(x)) return(FALSE)
#   for(i in x){
#     if(!is.atomic(i) || length(i) > 1) return(FALSE)
#   }
#   return(TRUE)
# }
# is.scalarlist <- is_scalarlist_r;


# #' #' @useDynLib jsonlite C_is_scalarlist
# #' is_scalarlist_c <- function(x){
# #'   .Call("C_is_scalarlist", x)
# #' }
# #' is.scalarlist <- is_scalarlist_c;


# #transpose_list (from JSON)
# transpose_list <- function(x, names) {
#   .Call("C_transpose_list", x, names)
# }


# #list_to_vec (from JSON)
# list_to_vec <- function(x) {
#   isdates <- is_datelist(x)
#   out <- unlist(null_to_na(x), recursive = FALSE, use.names = FALSE)
#   if(isdates && is.numeric(out)){
#     structure(out, class = c("POSIXct", "POSIXt"))
#   } else{
#     out
#   }
# }


# #is_datelist (from JSON)
# is_datelist <- function(x){
#   .Call("C_is_datelist", x)
# }


# #null_to_na (from JSON)
# null_to_na <- function(x) {
#   .Call("C_null_to_na", x)
# }

#' Geocode addresses and coordinates with OpenStreetMap
#'
#' Finds geographic coordinates of addresses and place names (forward geocoding),
#' or converts longitude/latitude coordinates to place names and administrative
#' units (reverse geocoding), using OpenStreetMap's Nominatim API.
#'
#' @param query Address or place name to be geocoded. Character string. Ignored
#'   if \code{reverse=TRUE}.
#' @param match_num If query matches multiple locations, which match to return?
#'   Default is 1 (highest-ranking match, by relevance). Numeric.
#' @param return_all Should all matches be returned? Overrides \code{match_num}
#'   if \code{TRUE}. Default is \code{FALSE}. Logical.
#' @param details Should detailed results be returned? Default is \code{FALSE}.
#'   Logical. Ignored if \code{reverse=TRUE}.
#' @param reverse Should reverse geocoding be performed (coordinates to address)?
#'   Default is \code{FALSE}. Logical.
#' @param lon Longitude of the point to reverse geocode. Required if
#'   \code{reverse=TRUE}. Numeric.
#' @param lat Latitude of the point to reverse geocode. Required if
#'   \code{reverse=TRUE}. Numeric.
#' @param zoom Zoom level for reverse geocoding, controlling the level of detail
#'   returned (0 = country, 10 = city, 18 = building). Default is 10. Numeric.
#' @param user_agent Valid User-Agent identifying the application for
#'   OSM Nominatim. If none supplied, function will attempt to auto-detect.
#'   Character string.
#' @return A \code{data.frame} object.
#'
#'   For forward geocoding (\code{reverse=FALSE}):
#'   \itemize{
#'   \item \code{query}. User-supplied address query. Character string.
#'   \item \code{osm_id}. OpenStreetMap ID. Character string.
#'   \item \code{address}. OpenStreetMap address. Character string.
#'   \item \code{longitude}. Horizontal coordinate. Numeric.
#'   \item \code{latitude}. Vertical coordinate. Numeric.
#'   }
#'   If \code{details=TRUE}, also contains:
#'   \itemize{
#'   \item \code{osm_type}. OpenStreetMap feature type. Character string.
#'   \item \code{importance}. Relevance of Nominatim match to query, from 0
#'     (worst) to 1 (best). Numeric.
#'   \item \code{bbox_ymin}. Minimum vertical coordinate of bounding box. Numeric.
#'   \item \code{bbox_ymax}. Maximum vertical coordinate of bounding box. Numeric.
#'   \item \code{bbox_xmin}. Minimum horizontal coordinate of bounding box. Numeric.
#'   \item \code{bbox_xmax}. Maximum horizontal coordinate of bounding box. Numeric.
#'   }
#'
#'   For reverse geocoding (\code{reverse=TRUE}):
#'   \itemize{
#'   \item \code{osm_name}. Full display name from OpenStreetMap. Character string.
#'   \item \code{osm_adm0_code}. ISO country code. Character string.
#'   \item \code{osm_adm0}. Country name. Character string.
#'   \item \code{osm_adm1}. State or first-level administrative division. Character string.
#'   \item \code{osm_adm2}. District or county. Character string.
#'   \item \code{osm_adm3}. Municipality. Character string.
#'   \item \code{osm_adm4}. Borough, village or street. Character string.
#'   }
#' @details Note that the Nominatim Usage Policy stipulates an absolute maximum
#'   of 1 request per second
#'   (\url{https://operations.osmfoundation.org/policies/nominatim/}).
#'   For batch geocoding of multiple addresses, please use
#'   \code{\link[SUNGEO]{geocode_osm_batch}}.
#' @importFrom curl curl_version
#' @importFrom data.table data.table :=
#' @importFrom dplyr mutate_all slice mutate select starts_with
#' @importFrom jsonlite fromJSON
#' @importFrom utils URLencode
#' @importFrom httr handle config status_code headers GET warn_for_status add_headers content
#' @importFrom rlang abort
#' @examples
#' # Geocode an address (top match only)
#' geocode_osm("Michigan Stadium")
#'
#' # Return detailed results for top match
#' geocode_osm("Michigan Stadium", details = TRUE)
#'
#' # Return detailed results for all matches
#' geocode_osm("Michigan Stadium", details = TRUE, return_all = TRUE)
#'
#' # Reverse geocode a coordinate pair
#' geocode_osm(reverse = TRUE, lon = -83.74868, lat = 42.26587, zoom = 18)
#'
#' @export

geocode_osm <- function(
    query,
    match_num   = 1,
    return_all  = FALSE,
    details     = FALSE,
    reverse     = FALSE,
    lon         = NULL,
    lat         = NULL,
    zoom        = 10,
    user_agent  = NULL
) {

  # Internal helpers

  request_GET <- function(x, url, ...) {
    x$response <- httr::GET(url, x$config, ..., handle = x$handle)
    x$html     <- new.env(parent = emptyenv(), hash = FALSE)
    x$url      <- x$response$url
    httr::warn_for_status(x$response)
    x
  }

  html_sessionSunGeo <- function(url, ...) {
    session <- structure(
      list(
        handle   = httr::handle(url),
        config   = c(..., httr::config(autoreferer = 1L)),
        url      = NULL,
        back     = character(),
        forward  = character(),
        response = NULL,
        html     = new.env(parent = emptyenv(), hash = FALSE)
      ),
      class = "session"
    )
    request_GET(session, url)
  }

  #' @exportS3Method print session
  print.session <- function(x, ...) {
    cat("<session> ", x$url, "\n", sep = "")
    cat("  Status: ", httr::status_code(x$response), "\n", sep = "")
    cat("  Type:   ", httr::headers(x)$`Content-Type`, "\n", sep = "")
    cat("  Size:   ", length(x$response$content), "\n", sep = "")
  }

  # Reverse geocoder

  if (isTRUE(reverse)) {
    if (is.null(lon) || is.null(lat)) {
      rlang::abort("When reverse=TRUE, both `lon` and `lat` must be supplied.")
    }

    reverse_osm <- function(lon, lat, zoom = 10, user_agent = NULL) {
      url <- sprintf(
        "https://nominatim.openstreetmap.org/reverse?lat=%f&lon=%f&format=json&zoom=%d&addressdetails=1",
        lat, lon, zoom
      )
      res <- tryCatch(
        httr::GET(url, httr::add_headers("User-Agent" = user_agent %||% "r-geocoder")),
        error = function(e) NULL
      )
      if (is.null(res)) return(NULL)

      doc <- jsonlite::fromJSON(httr::content(res, as = "text", encoding = "UTF-8"))
      if (is.null(doc$address)) return(NULL)

      out <- data.table::data.table(
        osm_name      = doc$display_name,
        osm_adm0_code = NA_character_,
        osm_adm0      = NA_character_,
        osm_adm1      = NA_character_,
        osm_adm2      = NA_character_,
        osm_adm3      = NA_character_,
        osm_adm4      = NA_character_
      )
      if ("country_code" %in% names(doc$address)) {
        out <- out[, osm_adm0_code := doc$address$country_code]
      }
      if ("country" %in% names(doc$address)) {
        out <- out[, osm_adm0 := doc$address$country]
      }
      if ("state" %in% names(doc$address)) {
        out <- out[, osm_adm1 := doc$address$state]
      }
      if ("district" %in% names(doc$address)) {
        out <- out[, osm_adm2 := doc$address$district]
      }
      if ("county" %in% names(doc$address) & !"district" %in% names(doc$address)) {
        out <- out[, osm_adm2 := doc$address$county]
      }
      if ("municipality" %in% names(doc$address)) {
        out <- out[, osm_adm3 := doc$address$municipality]
      }
      if ("city" %in% names(doc$address)) {
        out <- out[, osm_adm3 := doc$address$city]
      }
      if ("borough" %in% names(doc$address)) {
        out <- out[, osm_adm4 := doc$address$borough]
      }
      if ("village" %in% names(doc$address)) {
        out <- out[, osm_adm4 := doc$address$village]
      }
      if ("road" %in% names(doc$address) & !"village" %in% names(doc$address)) {
        out <- out[, osm_adm4 := doc$address$road]
      }
      rm(lon, lat, user_agent, doc, res, url, zoom)
      return(out)
    }

    # User-Agent auto-detect
    if (length(user_agent) == 0) {
      user_agent <- tryCatch({
        jsonlite::fromJSON(httr::content(httr::GET("https://httpbin.org/user-agent"), as = "text", encoding = "UTF-8"))$`user-agent`
      }, error = function(e) "r-geocoder")
    }

    result <- reverse_osm(lon = lon, lat = lat, zoom = zoom, user_agent = user_agent)
    if (is.null(result)) {
      message("Cannot access OSM server or no result returned. Please check your internet connection and coordinates.")
      return(invisible(NULL))
    }
    return(as.data.frame(result))
  }

  # Forward geocoder

  # Batch geocoding warning
  if (length(query) > 1) {
    query <- query[1]
    warning("Returning first result only. Please use geocode_osm_batch() to geocode multiple addresses.")
  }

  # Error handling
  downloadFail <- FALSE
  tryCatch({

    # Create empty objects
    osm_id     <- NA_character_
    osm_type   <- NA_character_
    address    <- NA_character_
    longitude  <- NA_real_
    latitude   <- NA_real_
    importance <- NA_real_
    bbx <- data.frame(
      bbox_ymin = NA_real_, bbox_ymax = NA_real_,
      bbox_xmin = NA_real_, bbox_xmax = NA_real_,
      stringsAsFactors = FALSE
    )

    # User-Agent
    if (length(user_agent) == 0) {
      user_agent <- tryCatch({
        jsonlite::fromJSON(httr::content(httr::GET("https://httpbin.org/user-agent"), as = "text", encoding = "UTF-8"))$`user-agent`
      }, error = function(e) "r-geocoder")
    }


    # Send query to OSM Nominatim API
    root_url <- "https://nominatim.openstreetmap.org/search?q="
    sufx_url <- "&format=json&polygon=1&addressdetails=1"
    response <- httr::GET(
      utils::URLencode(paste0(root_url, query, sufx_url), repeated = TRUE),
      httr::add_headers("User-Agent" = user_agent)
    )
    doc <- httr::content(response, as = "text", encoding = "UTF-8")

  }, warning = function(w) {
    downloadFail <<- TRUE
  }, error = function(e) {
    message("Cannot access OSM server. Please check your internet connection and try again.")
    downloadFail <<- TRUE
  }, finally = {
  })

  if (downloadFail) {
    cat("Cannot access OSM server. Please check your internet connection and try again.")
    return()
  } else {

    # Parse results
    if (nchar(doc) > 2) {
      dat <- jsonlite::fromJSON(doc)
      if (return_all) { match_num <- 1:nrow(dat) }
      if (nrow(dat) > 0) {
        osm_id     <- as.character(dat$osm_id[match_num])
        osm_type   <- as.character(dat$osm_type[match_num])
        address    <- as.character(dat$display_name[match_num])
        longitude  <- as.numeric(as.character(dat$lon[match_num]))
        latitude   <- as.numeric(as.character(dat$lat[match_num]))
        importance <- as.numeric(as.character(dat$importance[match_num]))
        bbx <- dplyr::slice(
          dplyr::mutate_all(
            as.data.frame(do.call(rbind, dat$boundingbox), stringsAsFactors = FALSE),
            function(.) { as.numeric(.) }
          ),
          match_num
        )
        colnames(bbx) <- c("bbox_ymin", "bbox_ymax", "bbox_xmin", "bbox_xmax")
      }
    }

    # Output
    out <- cbind(
      data.frame(
        query      = query,
        osm_id     = osm_id,
        osm_type   = osm_type,
        importance = importance,
        address    = address,
        longitude  = longitude,
        latitude   = latitude,
        stringsAsFactors = FALSE
      ),
      bbx
    )
    if (!details) {
      out <- dplyr::select(
        dplyr::mutate(out, osm_type = NULL, importance = NULL),
        -dplyr::starts_with("bbox")
      )
    }
    return(as.data.frame(out))
  }
}

#' \code{SUNGEO}
#'
#' Sub-National Geospatial Data Archive System: Geoprocessing Toolkit
#'
#' See the README on
#' [GitHub](https://github.com/zhukovyuri/SUNGEO#readme)
#'
#' @docType package
#' @name SUNGEO
#'
## quiets concerns of R CMD check re: the .'s that appear in pipelines
if(getRversion() >= "2.15.1"){
  utils::globalVariables(c("%>%",".",":=","AREA_W","POP_INT","POP_TOTAL","POP_W","TERM","alternatenames","asciiname","feature_code","geometry","geonameid","gn_type","latitude","longitude","match_type","name","sgeoz_length","strdist","area_code","col.id","overlap_1","overlap_2","action_dots","abort","vctrs_as_location","ST_MULTIPOINT","Field","Var","var_","Return_ID",".N",".SD","DATE","DATE_ALT","MID","POP_W_INV","TID","Varz","WID","X","Y","YEAR","YRMO","YRWK","osm_adm0_code","osm_adm0","osm_adm1","osm_adm2","osm_adm3","osm_adm4"))
  }

#' Convert simple features object into regularly spaced raster
#'
#' This function takes in an \code{sf} spatial object (polygon or point) and returns a regularly spaced RasterLayer. Reverse translation option allows users to create an \code{sf} polygon object from the regularly spaced RasterLayer. This function can also conver the \code{sf} object into a cartogram with a user-specified variable name.
#'
#' @param polyz_from Source polygon layer. \code{sf} object.
#' @param pointz_from Source point layer. \code{sf} object.
#' @param input_variable Name of input variable from source layer. Character string.
#' @param aggregate_function Aggregation function to be applied to variables specified in \code{input_variable}. Must take as an input a numeric vector \code{x}. Function or list of functions. Default is \code{mean}.
#' @param reverse Reverse translation from raster layer to \code{sf} polygon object (polygon features only). Default is \code{FALSE}.
#' @param poly_to Destination polygon layer for reverse conversion. Must be specified if \code{reverse=TRUE}. \code{sf} object.
#' @param return_output Return output for reverse conversion. Must be specified if \code{reverse=TRUE}.
#' @param return_field Return field for reverse conversion. Must be specified if \code{reverse=TRUE}.
#' @param reverse_function  Aggregation function for reverse conversion. Must be specified if \code{reverse=TRUE}. Function or list of functions. Default is \code{mean}.
#' @param grid_dim Dimensions of raster grid. Numerical vector of length 2 (number of rows, number of columns). Default is \code{c(1000,1000)}.
#' @param cartogram Cartogram transformation. Logical. Default is \code{FALSE}.
#' @param carto_var Input variable for cartogram transformation. Must be specified if \code{cartogram=TRUE}. Character string.
#' @param message_out Print informational messages. Logical. Default is \code{TRUE}.
#' @param return_list Return full set of results, including input polygons, centroids and field raster. Default is \code{FALSE}. Logical.
#' @return If \code{return_list=FALSE} (default) and \code{reverse=FALSE} (default), returns \code{RasterLayer} object, with cell values corresponding to \code{input_variable}.
#'
#' If \code{return_list=TRUE} and input layer is polygon, returns a list containing
#' \itemize{
##'  \item "return_output". Output raster, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
##'  \item "return_centroid". Raster of centroids, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
##'  \item "poly_to". Source polygons, with columns corresponding to \code{input_variable} and auto-generated numerical ID \code{Field}. \code{sf} object.
##'  \item "return_field". Output raster, with values corresponding to auto-generated numerical ID \code{Field}. \code{RasterLayer} object.}
##'
##' If \code{return_list=TRUE} and  input layer is points, returns a list containing
#' \itemize{
##'  \item "return_output". Output raster, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
##'  \item "return_point". Source points, with column corresponding to \code{input_variable}.
##'  }
##' If \code{reverse=TRUE}, returns an \code{sf} polygon layer, with columns corresponding to \code{input_variable} and auto-generated numerical ID \code{Field}.
#' @import packcircles cartogram Rcpp
#' @importFrom terra rasterize ncell res ext vect as.data.frame res extract values mask plot
#' @importFrom cartogram cartogram_cont
#' @importFrom sf st_collection_extract st_crs st_geometry st_centroid st_buffer st_transform st_convex_hull st_union
#' @importFrom dplyr group_by summarize left_join
#' @importFrom purrr reduce
#' @examples
#' # Rasterization of polygon layer.
#' data(clea_deu2009)
#' out_1 <- sf2raster(polyz_from = utm_select(clea_deu2009),
#'                    input_variable = "to1")
#' terra::plot(out_1)
#'
#' \donttest{
#' # Rasterization of point layer
#' data(clea_deu2009_pt)
#' out_2 <- sf2raster(pointz_from = utm_select(clea_deu2009_pt),
#'                    input_variable = "to1",
#'                    grid_dim = c(25,25))
#' terra::plot(out_2)
#' }
#' \donttest{
#' # Cartogram (vote turnout scaled by number of valid votes)
#' out_3 <- sf2raster(polyz_from = utm_select(clea_deu2009),
#'                    input_variable = "to1",
#'                    cartogram = TRUE,
#'                    carto_var = "vv1")
#' terra::plot(out_3)
#' }
#' \donttest{
#' # Polygonization of cartogram raster
#' out_4a <- sf2raster(polyz_from = utm_select(clea_deu2009),
#'                     input_variable = "to1",
#'                     cartogram = TRUE,
#'                     carto_var = "vv1",
#'                     return_list = TRUE)
#' out_4 <- sf2raster(reverse = TRUE,
#'                    poly_to = out_4a$poly_to,
#'                    return_output = out_4a$return_output,
#'                    return_field = out_4a$return_field)
#' terra::plot(out_4)
#' }
#' @export


sf2raster <- function(polyz_from = NULL,
                       pointz_from = NULL,
                       input_variable = NULL,
                       reverse = FALSE,
                       poly_to = NULL,
                       return_output = NULL,
                       return_field = NULL,
                       aggregate_function = list(function(x) mean(x, na.rm = TRUE)),
                       reverse_function = list(function(x) mean(x, na.rm = TRUE)),
                       grid_dim = c(1000, 1000),
                       cartogram = FALSE,
                       carto_var = NULL,
                       message_out = TRUE,
                       return_list = FALSE){

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })


  if (!inherits(aggregate_function, "list")) {
    aggregate_function <- list(aggregate_function)
  }
  if (!inherits(reverse_function, "list")) {
    reverse_function <- list(reverse_function)
  }

  if (!is.null(polyz_from) & message_out && !reverse && !any(grepl("meter|metre",sf::st_crs(polyz_from),ignore.case = TRUE))) {
    message("Note: Please project the SF object into a meter-based projection coordinate system prior to using this function. Default grid resolution is in meters.")
  }
  if (!is.null(pointz_from) & message_out && !reverse && !any(grepl("meter|metre",sf::st_crs(pointz_from),ignore.case = TRUE))) {
    message("Note: Please project the SF object into a meter-based projection coordinate system prior to using this function. Default grid resolution is in meters.")
  }
  if (!is.null(polyz_from) && !is.null(pointz_from) && !reverse) {
    stop("Error: Please enter only one type of SF object {polygon or point}.")
  }
  if (!reverse) {
    CheckObj <- if (!is.null(polyz_from)) {
      polyz_from
    }
    else {
      pointz_from
    }
  } else {
    CheckObj <- poly_to
  }
  if (all(!("sf" %in% class(CheckObj)))) {
    stop("Error: Only SF formatted objects are accepted. Please convert to an sf object and then re-run function")
  }
  if (reverse && (is.null(poly_to) || is.null(return_output) ||
                  is.null(return_field))) {
    stop("Error: If translating back into SF Polygon object, then ensure poly_to, return_output, and return_field are specified")
  }

  if (!reverse) {
    if (message_out) {
      message(paste("Converting SF object into a regularly spaced raster grid of dimensions:",
                    grid_dim[1], "by", grid_dim[2]))
    }
    if (cartogram && !is.null(polyz_from)) {
      if (is.null(carto_var)) {
        stop("Please supply a input_variable for cartogram transformation")
      }
      polyz_from <- cartogram::cartogram_cont(polyz_from,
                                              carto_var)
      polyz_from <- sf::st_collection_extract(polyz_from,
                                              "POLYGON")
    }
    if (cartogram && is.null(polyz_from)) {
      stop("Cartogram translation is only available for spatial polygon objects. Please insert the appropriate spatial feature.")
    }
    if (!is.null(polyz_from)) {
      polyz_from$Field <- 1:nrow(polyz_from)
      sel.col <- which(names(polyz_from) %in% c("Field",
                                                input_variable))
      polyz_from <- polyz_from[, sel.col]
      polyz_from_SP_keep <- polyz_from
    }
    if (!is.null(pointz_from)) {
      sel.col <- which(names(pointz_from) %in% c(input_variable))
      pointz_from <- pointz_from[, sel.col]
      names(pointz_from)[names(pointz_from) %in% input_variable] <- "Var"
      pointz_from_SP_keep <- pointz_from
      names(pointz_from_SP_keep)[names(pointz_from_SP_keep) %in%
                                   "Var"] <- input_variable
    }
    ProjectionObj <- if (!is.null(polyz_from)) {
      sf::st_crs(polyz_from)
    } else {
      sf::st_crs(pointz_from)
    }
    if (grepl("^EPSG", ProjectionObj$input) & !grepl("^\\+",
                                                     ProjectionObj$input)) {
      ProjectionObj$input <- tolower(paste0("+init=", ProjectionObj$input))
    }
    if (!is.null(polyz_from)){
      Template_Raster <- terra::rast(ncol = grid_dim[2], nrow = grid_dim[1], crs = ProjectionObj$input, extent=terra::ext(polyz_from))
    } else {
      Template_Raster <- terra::rast(ncol = grid_dim[2], nrow = grid_dim[1], crs = ProjectionObj$input, extent=terra::ext(pointz_from))
    }
    suppressWarnings({
      terra::values(Template_Raster) <- 1:terra::ncell(Template_Raster)
    })
    names(Template_Raster) <- "ID"
    impute_layer <- Template_Raster
    if (!is.null(pointz_from)) {
      pointz_from$Field <- unlist(terra::extract(x=Template_Raster,y=terra::vect(pointz_from),ID=FALSE))
      pointz_from_xy <- pointz_from
      sf::st_geometry(pointz_from_xy) <- NULL
      suppressMessages({
        pointz_from_xy <- dplyr::group_by(pointz_from_xy,
                                          Field)
        pointz_from_xy <- dplyr::summarize(pointz_from_xy,
                                           Var = aggregate_function[[1]](Var))
        pointz_from_xy <- as.data.frame(pointz_from_xy)
      })
      Template_Raster_xy <- terra::as.data.frame(Template_Raster,na.rm=FALSE)
      # head(Template_Raster_xy)
      Template_Raster_xy <- dplyr::left_join(Template_Raster_xy,
                                             pointz_from_xy, by = c(ID = "Field"))
      output_data <- Template_Raster
      terra::values(output_data) <- Template_Raster_xy$Var
      names(output_data) <- input_variable
      ReturnList <- list(return_output = output_data, return_point = pointz_from_SP_keep)
    }
    if (!is.null(polyz_from)) {
      polyz_from_layer <- terra::rasterize(polyz_from,
                                           Template_Raster, field = "Field")
      # terra::res(polyz_from_layer) <- c(grid_res[1], grid_res[2])
      # terra::plot(polyz_from_layer)
      field_layer <- polyz_from_layer
      names(field_layer) <- "Field"
      polyz_from_xy <- terra::as.data.frame(polyz_from_layer,na.rm=FALSE)
      names(polyz_from_xy)[1] <- "Field"
      polyz_from_df <- polyz_from
      sf::st_geometry(polyz_from_df) <- NULL
      dim(polyz_from_xy)
      dim(polyz_from_df)
      terra::ncell(polyz_from_layer)
      terra::ncell(Template_Raster)
      polyz_from_xy <- dplyr::left_join(polyz_from_xy,
                                        polyz_from_df, by = "Field")
      # polyz_from_xy <- base::merge(polyz_from_xy,polyz_from_df, by = "Field",all.x=TRUE,all.y=FALSE)

      dim(polyz_from_xy)
      names(polyz_from_xy)[names(polyz_from_xy) %in% c(input_variable)] <- "Var"
      terra::values(polyz_from_layer) <- polyz_from_xy$Var
      output_data <- polyz_from_layer
      names(output_data) <- input_variable
      suppressWarnings({
        centroid_polygon <- sf::st_centroid(polyz_from_SP_keep)
        centroid_polygon <- sf::st_buffer(centroid_polygon,
                                          dist = terra::res(Template_Raster)[1])
      })
      centroid_layer <- terra::mask(output_data, centroid_polygon)
      # terra::plot(centroid_layer)
      ReturnList <- list(return_output = output_data, return_centroid = centroid_layer,
                         poly_to = polyz_from_SP_keep, return_field = field_layer)
    }
    if (return_list == FALSE) {
      ReturnList <- ReturnList["return_output"][[1]]
    }
    return(ReturnList)
  }
  if (reverse) {
    if (message_out) {
      message("Converting regularly spaced raster grid cell into SF polygon object")
    }
    namesOutput <- names(return_output)
    Polygon_Frame <- c(return_field, return_output)
    names(Polygon_Frame)[1] <- "Field"
    names(Polygon_Frame)[-1] <- namesOutput
    Polygon_Frame <- terra::as.data.frame(Polygon_Frame)
    Polygon_Frame <- lapply(2:ncol(Polygon_Frame), function(col) {
      Selected_Matrix <- Polygon_Frame[, c(1, col)]
      names(Selected_Matrix)[2] <- "Var"
      suppressMessages({
        Selected_Matrix <- dplyr::group_by(Selected_Matrix,
                                           Field)
        Selected_Matrix <- dplyr::summarize(Selected_Matrix,
                                            Var = reverse_function[[1]](Var))
      })
      names(Selected_Matrix)[2] <- names(Polygon_Frame)[col]
      return(Selected_Matrix)
    })
    if (length(Polygon_Frame) > 1) {
      Polygon_Frame <- purrr::reduce(.x = Polygon_Frame,
                                     .f = full_join, by = "Field")
    }
    else {
      Polygon_Frame <- Polygon_Frame[[1]]
    }
    poly_to <- poly_to[, which(names(poly_to) %in% "Field")]
    poly_to <- dplyr::left_join(poly_to, Polygon_Frame, by = "Field")
    return(poly_to)
  }
}

#' Information on available SUNGEO data files
#'
#' This function reports the availability of data files on the SUNGEO server, searchable by country and topic.
#'
#' @param country_names Country name(s). Character string (single country) or vector of character strings (multiple countries).
#' @param country_iso3s Country code (ISO 3166-1 alpha-3). Character string (single country) or vector of character strings (multiple countries).
#' @param topics Data topics. See \code{get_info()} for full list. Character string (single topic) or vector of character strings (multiple topics).
#' @return list object, with three slots: 'summary', 'topics', and 'geoset'.
#' @importFrom data.table data.table copy
#' @importFrom stringr str_split
#' @importFrom utils data
#' @examples
#' # Get list of all available data
#' out_1 <- get_info()
#' out_1["summary"]
#' out_1["topics"]
#' out_1["geosets"]
#'
#' # Get list of available data for a single country
#' out_2 <- get_info(country_names="Afghanistan")
#' out_2
#'
#' # Get list of available data for a single topic
#' out_3 <- get_info(topics="Elections:LowerHouse:CLEA")
#' out_3
#'
#' # Get list of available data for a multiple countries and topics
#' out_4 <- get_info(
#'                  country_names=c("Afghanistan","Zambia"),
#'                  topics=c("Elections:LowerHouse:CLEA","Events:PoliticalViolence:GED"))
#' out_4
#'
#' @seealso \code{\link{get_data}}
#' @export
get_info <- function(
    country_names=NULL,
    country_iso3s=NULL,
    topics=NULL
){

  # Get object
  #utils::data("available_data")
  country_iso3 <- desc <- geoset_years <- geosets <- space_units <- year_range <- country_name_alt <- NULL

  # Country codes
  if(length(country_iso3s) == 0 & length(country_names) > 0){
    country_iso3s <- unique(SUNGEO::cc_dict[match(country_names,SUNGEO::cc_dict[,country_name_alt]),country_iso3])
  }

  # Default behavior
  if(is.null(country_iso3s)&is.null(topics)){
    avail_subset <- data.table::copy(SUNGEO::available_data)
    # Summarize
    country_by_topic <- sapply(sapply(avail_subset,function(x){x[,country_iso3]}),function(x){paste0(x,collapse=", ")})
    units_by_topic <- sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){paste0(x,collapse=", ")})
    years_by_topic <- sapply(sapply(sapply(avail_subset[!grepl("^Geoset",names(avail_subset))],function(x){x[,year_range]}),unique),function(x){paste0(x,collapse=", ")})
    years_by_geoset <- sapply(sapply(avail_subset[grepl("^Geoset",names(avail_subset))],function(x){x[,geoset_years]}),unique)
    years <- unlist(stringr::str_split(years_by_topic,", |-"))
    years <- paste0(range(years[is.finite(as.numeric(years))]),collapse="-")
    n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
    n_units <- length(unique(unlist(sapply(stringr::str_split(units_by_topic,", "),length))))
    n_geosets <- length(grep("^Geoset",names(avail_subset)))
    avail_geosets <- names(avail_subset)[grep("^Geoset",names(avail_subset))]
    avail_geosets <- data.table::data.table(geosets=avail_geosets)[,country_iso3:=country_by_topic[match(geosets,names(country_by_topic))]][,desc:=paste0(geosets," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(geosets," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    n_topics <- sum(!grepl("^Geoset",names(avail_subset)))
    avail_topics <- names(avail_subset)[!grepl("^Geoset",names(avail_subset))]
    avail_topics <- data.table::data.table(topics=avail_topics)[,country_iso3:=country_by_topic[match(topics,names(country_by_topic))]][,desc:=paste0(topics," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(topics," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    # Output
    summary_message <- c("-----------------------------------------","Summary of data available through SUNGEO","-----------------------------------------","",paste0("Found data for ",n_countries," countries, ",n_topics," topics"),paste0(n_units," spatial units, ",n_geosets," geographic boundary sets"),"","-------","Topics","-------",avail_topics[,(desc)],"","-------","Geosets","-------",avail_geosets[,(desc)],"","-------","See 'topics' and 'geosets' slots for details on spatial units and years.","Please narrow your search criteria for more informative results.","-------"
    )
    print(summary_message)
    out <- list(summary=summary_message,topics=avail_subset[!grepl("^Geoset",names(avail_subset))],geosets=avail_subset[grepl("^Geoset",names(avail_subset))])
  }

  # By topic
  if(is.null(country_iso3s)&!is.null(topics)){
    # Take subset
    avail_subset <- SUNGEO::available_data[topics]
    # Summarize
    if(length(topics)==1){
      country_by_topic <- paste0(sapply(avail_subset,function(x){x[,country_iso3]}),collapse=", ")
      names(country_by_topic) <- topics
      units_by_topic <- paste0(sort(unique(unlist(sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){stringr::str_split(x,", ")})))),collapse=", ")
      names(units_by_topic) <- topics
      years_by_topic <- paste0(range(unlist(sapply(sapply(avail_subset,function(x){x[,year_range]}),function(x){stringr::str_split(x,", |-")}))),collapse="-")
      names(years_by_topic) <- topics
      years <- paste0(range(unlist(stringr::str_split(years_by_topic,", |-"))),collapse="-")
      n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
      n_units <- length(unique(unlist(unique(stringr::str_split(units_by_topic,", ")))))
      n_geosets <- length(unique(unlist(sapply(sapply(sapply(avail_subset,function(x){x[,geosets]}),function(x){stringr::str_split(as.character(x),"-|, ")}),unlist))))
    }
    if(length(topics)>1){
      country_by_topic <- sapply(sapply(avail_subset,function(x){x[,country_iso3]}),function(x){paste0(x,collapse=", ")})
      names(country_by_topic) <- topics
      units_by_topic <- sapply(sapply(sapply(sapply(sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){stringr::str_split(x,", ")}),unlist),unique),sort),function(x){paste0(x,collapse=", ")})
      years_by_topic <- apply(sapply(sapply(sapply(avail_subset,function(x){x[,year_range]}),function(x){stringr::str_split(as.character(x),"-|, ")}),range),2,function(x){paste0(unique(x),collapse="-")})
      years <- paste0(range(unlist(stringr::str_split(years_by_topic,", |-"))),collapse="-")
      n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
      n_units <- length(unique(unlist(unique(stringr::str_split(units_by_topic,", ")))))
      n_geosets <- length(unique(unlist(sapply(sapply(sapply(avail_subset,function(x){x[,geosets]}),function(x){stringr::str_split(as.character(x),"-|, ")}),unlist))))
    }
    avail_countries <- unique(unlist(stringr::str_split(country_by_topic,", ")))
    avail_geosets <- paste0("Geoset:",unique(unlist(stringr::str_split(unlist(sapply(avail_subset,function(x){x[,geosets]})),", "))))
    avail_countries <- sapply(SUNGEO::available_data[avail_geosets],function(x){avail_countries[avail_countries%in%x[,country_iso3]]})
    suppressWarnings({
      avail_geosets <- data.table::data.table(geosets=avail_geosets)[,country_iso3:=avail_countries[match(geosets,names(avail_countries))]][,desc:=paste0(geosets," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(geosets," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    })
    n_topics <- sum(!grepl("^Geoset",names(avail_subset)))
    avail_topics <- names(avail_subset)[!grepl("^Geoset",names(avail_subset))]
    avail_topics <- data.table::data.table(topics=avail_topics)[,country_iso3:=country_by_topic[match(topics,names(country_by_topic))]][,desc:=paste0(topics," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(topics," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    # Output
    summary_message <- c("-----------------------------------------","Summary of data available through SUNGEO","-----------------------------------------","",paste0("Found data for ",n_countries," countries, ",n_topics," topics"),paste0(n_units," spatial units, ",n_geosets," geographic boundary sets"),"","-------","Topics","-------",avail_topics[,(desc)],"","-------","Geosets","-------",avail_geosets[,(desc)],"","-------","See 'topics' and 'geosets' slots for details on spatial units and years.","-------" )
    print(summary_message)
    out <- list(summary=summary_message,topics=avail_subset[!grepl("^Geoset",names(avail_subset))],geosets=lapply(SUNGEO::available_data[avail_geosets[,geosets]],function(x){x[country_iso3%in%unique(unlist(avail_countries)),]}))
  }

  # By country
  if(!is.null(country_iso3s)&is.null(topics)){
    # Take subset
    avail_subset <- sapply(SUNGEO::available_data,function(x){x[country_iso3%in%c(country_iso3s)]})
    avail_subset <- avail_subset[sapply(avail_subset,nrow)>0]
    # Summarize
    country_by_topic <- sapply(sapply(avail_subset,function(x){x[,country_iso3]}),function(x){paste0(x,collapse=", ")})
    units_by_topic <- sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){paste0(x,collapse=", ")})
    years_by_topic <- sapply(sapply(sapply(avail_subset[!grepl("^Geoset",names(avail_subset))],function(x){x[,year_range]}),unique),function(x){paste0(x,collapse=", ")})
    years_by_geoset <- sapply(sapply(avail_subset[grepl("^Geoset",names(avail_subset))],function(x){x[,geoset_years]}),unique)
    years <- paste0(range(unlist(stringr::str_split(years_by_topic,", |-"))),collapse="-")
    n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
    n_units <- length(unique(unlist(sapply(stringr::str_split(units_by_topic,", "),length))))
    n_geosets <- length(grep("^Geoset",names(avail_subset)))
    avail_geosets <- names(avail_subset)[grep("^Geoset",names(avail_subset))]
    avail_geosets <- data.table::data.table(geosets=avail_geosets)[,country_iso3:=country_by_topic[match(geosets,names(country_by_topic))]][,desc:=paste0(geosets," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(geosets," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    n_topics <- sum(!grepl("^Geoset",names(avail_subset)))
    avail_topics <- names(avail_subset)[!grepl("^Geoset",names(avail_subset))]
    avail_topics <- data.table::data.table(topics=avail_topics)[,country_iso3:=country_by_topic[match(topics,names(country_by_topic))]][,desc:=paste0(topics," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(topics," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    # Output
    summary_message <- c("-----------------------------------------","Summary of data available through SUNGEO","-----------------------------------------","",paste0("Found data for ",n_countries," countries, ",n_topics," topics"),paste0(n_units," spatial units, ",n_geosets," geographic boundary sets"),"","-------","Topics","-------",avail_topics[,(desc)],"","-------","Geosets","-------",avail_geosets[,(desc)],"","-------","See 'topics' and 'geosets' slots for details on spatial units and years.","-------" )
    print(summary_message)
    out <- list(summary=summary_message,topics=avail_subset[!grepl("^Geoset",names(avail_subset))],geosets=avail_subset[grepl("^Geoset",names(avail_subset))])
  }

  # By topic and country
  if(!is.null(country_iso3s)&!is.null(topics)){
    # Take subset
    avail_subset <- sapply(SUNGEO::available_data,function(x){x[country_iso3%in%c(country_iso3s)]})
    avail_subset <- avail_subset[topics]
    avail_subset <- avail_subset[sapply(avail_subset,nrow)>0]
    # Summarize
    if(length(topics)==1){
      country_by_topic <- paste0(sapply(avail_subset,function(x){x[,country_iso3]}),collapse=", ")
      names(country_by_topic) <- topics
      units_by_topic <- paste0(sort(unique(unlist(sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){stringr::str_split(x,", ")})))),collapse=", ")
      names(units_by_topic) <- topics
      years_by_topic <- paste0(range(unlist(sapply(sapply(avail_subset,function(x){x[,year_range]}),function(x){stringr::str_split(x,", |-")}))),collapse="-")
      names(years_by_topic) <- topics
      years <- paste0(range(unlist(stringr::str_split(years_by_topic,", |-"))),collapse="-")
      n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
      n_units <- length(unique(unlist(unique(stringr::str_split(units_by_topic,", ")))))
      n_geosets <- length(unique(unlist(sapply(sapply(sapply(avail_subset,function(x){x[,geosets]}),function(x){stringr::str_split(as.character(x),"-|, ")}),unlist))))
    }
    if(length(topics)>1){
      country_by_topic <- sapply(sapply(avail_subset,function(x){x[,country_iso3]}),function(x){paste0(x,collapse=", ")})
      names(country_by_topic) <- topics
      units_by_topic <- sapply(sapply(sapply(sapply(sapply(sapply(avail_subset,function(x){x[,space_units]}),function(x){stringr::str_split(x,", ")}),unlist),unique),sort),function(x){paste0(x,collapse=", ")})
      years_by_topic <- apply(sapply(sapply(sapply(avail_subset,function(x){x[,year_range]}),function(x){stringr::str_split(as.character(x),"-|, ")}),range),2,function(x){paste0(unique(x),collapse="-")})
      years <- paste0(range(unlist(stringr::str_split(years_by_topic,", |-"))),collapse="-")
      n_countries <- max(sapply(stringr::str_split(country_by_topic,", "),length))
      n_units <- length(unique(unlist(unique(stringr::str_split(units_by_topic,", ")))))
      n_geosets <- length(unique(unlist(sapply(sapply(sapply(avail_subset,function(x){x[,geosets]}),function(x){stringr::str_split(as.character(x),"-|, ")}),unlist))))
    }
    avail_countries <- unique(unlist(stringr::str_split(country_by_topic,", ")))
    avail_geosets <- paste0("Geoset:",unique(unlist(stringr::str_split(unlist(sapply(avail_subset,function(x){x[,geosets]})),", "))))
    avail_countries <- sapply(sapply(SUNGEO::available_data[avail_geosets],function(x){avail_countries[avail_countries%in%x[,country_iso3]]}),function(x){paste0(x,collapse=", ")})
    suppressWarnings({
      avail_geosets <- data.table::data.table(geosets=avail_geosets)[,country_iso3:=avail_countries[match(geosets,names(avail_countries))]][,desc:=paste0(geosets," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(geosets," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    })
    n_topics <- sum(!grepl("^Geoset",names(avail_subset)))
    avail_topics <- names(avail_subset)[!grepl("^Geoset",names(avail_subset))]
    avail_topics <- data.table::data.table(topics=avail_topics)[,country_iso3:=country_by_topic[match(topics,names(country_by_topic))]][,desc:=paste0(topics," (",country_iso3,")")][sapply(stringr::str_split(country_iso3,","),length)>3,desc:=paste0(topics," (",sapply(stringr::str_split(country_iso3,","),length)," countries)")]
    # Output
    summary_message <- c("-----------------------------------------","Summary of data available through SUNGEO","-----------------------------------------","",paste0("Found data for ",n_countries," countries, ",n_topics," topics"),paste0(n_units," spatial units, ",n_geosets," geographic boundary sets"),"","-------","Topics","-------",avail_topics[,(desc)],"","-------","Geosets","-------",avail_geosets[,(desc)],"","-------","See 'topics' and 'geosets' slots for details on spatial units and years.","-------" )
    print(summary_message)
    out <- list(summary=summary_message,topics=avail_subset[!grepl("^Geoset",names(avail_subset))],geosets=lapply(SUNGEO::available_data[avail_geosets[,geosets]],function(x){x[country_iso3%in%unique(unlist(avail_countries)),]}))
  }



  return(out)
}

#' Download data from SUNGEO server
#'
#' Function to download data files through the SUNGEO API. Function produces a data.table object, corresponding to the user's choice of countries, topics, sources, and spatial and temporal units.
#'
#' @param country_names Country name(s). Character string (single country) or vector of character strings (multiple countries).
#' @param country_iso3 Country code (ISO 3166-1 alpha-3). Character string (single country) or vector of character strings (multiple countries).
#' @param geoset Name of geographic boundary set. Can be one of \code{"GADM"} (Database of Global Administrative Areas), \code{"GAUL"} (Global Administrative Unit Layers), \code{"geoBoundaries"}, \code{"GRED"} (GeoReferenced Electoral Districts Datasets), \code{"HEXGRID"} (SUNGEO Hexagonal Grid), \code{"MPIDR"} (Max Planck Institute for Demographic Research Population History GIS Collection), \code{"NHGIS"} (National Historical Geographic Information System), \code{"PRIOGRID"} (PRIO-GRID 2.0), \code{"SHGIS"} (SUNGEO Historical GIS). Default is \code{"GADM"}. Character string.
#' @param geoset_yr Year of geographic boundaries. See \code{get_info()['geosets']} for availability. Default is \code{2018}. Integer.
#' @param space_unit Geographic level of analysis. Can be one of \code{"adm0"} (country), \code{"adm1"} (province), \code{"adm2"} (district), \code{"cst"} (GRED electoral constituency), \code{"hex05"} (SUNGEO Hexagonal Grid cell), \code{"prio"} (PRIO-GRID cell). See \code{get_info()['geosets']} for availability by geoset, country and topic. Default is \code{"adm1"}. Character string.
#' @param time_unit Temporal level of analysis. Can be one of \code{"year"}, \code{"month"}, \code{"week"}. See \code{get_info()['topics']} for availability by topic. Default is \code{"year"}. Character string.
#' @param topics Data topics. See \code{get_info()['summary']} for full list. Character string (single topic) or vector of character strings (multiple topics).
#' @param year_min Time range of requested data: start year. See \code{get_info()['topics']} for availability by topic. Default is \code{1990}. Integer.
#' @param year_max Time range of requested data: end year. See \code{get_info()['topics']} for availability by topic. Default is \code{2017}. Integer.
#' @param print_url Print url string of requested data to console? Default is \code{TRUE}. Logical.
#' @param print_time Print processing time for API query to console? Default is \code{TRUE}. Logical.
#' @param error_stop Error handling. If \code{TRUE}, function terminates request if an error is encountered. If \code{FALSE}, error is skipped and error message is recorded in a new \code{message} column. Default is \code{FALSE}. Logical.
#' @param by_topic Break query down by topic and country? If \code{TRUE}, a separate request is sent to the API for each country and topic, and the results are combined on the client side. This ensures that data that are available for some, but not all countries are returned, rather than resulting in a failed request. If \code{FALSE}, a single request is sent to the API for all countries and topics, and the results are combined on the server side. Only data that are available for all countries are returned. Default is \code{TRUE}. Logical.
#' @param skip_missing Skip missing data topics? If \code{TRUE}, missing data topics are skipped, columns are populated with NAs, and corresponding error message is recorded in a new \code{message} column. If \code{FALSE}, returns NULL results for missing topics. Default is \code{TRUE}. Logical.
#' @param cache_param Store cached query on server? This can speed up processing for repeated queries. Default is \code{FALSE}. Logical.
#' @param short_message Shorten error messages? If \code{TRUE}, a short, informative error message is recorded in the \code{message} column. If \code{FALSE}, full error message is recorded. Default is \code{TRUE}. Logical.
#' @return data.table object, with requested data from SUNGEO API.
#' @importFrom data.table data.table as.data.table setnames rbindlist copy
#' @importFrom jsonlite fromJSON
#' @importFrom httr GET content
#' @importFrom stringr str_split str_extract_all
#' @importFrom dplyr last
#' @seealso \code{\link{get_info}}
#' @examples
#' \donttest{
#' # Single country, single topic
#' out_1 <- get_data(country_name="Afghanistan",topics="Demographics:Population:GHS")
#' out_1
#' }
#'
#' \donttest{
#  # Multiple countries, multiple topics
#' out_2 <- get_data(
#' 	country_name=c("Afghanistan","Moldova"),
#' 	topics=c("Demographics:Ethnicity:EPR","Demographics:Population:GHS"))
#' out_2
#' }
#'
#' \donttest{
#' # Other boundary sets, spatial and time units
#' out_3 <- get_data(
#' 	country_name="Albania",
#' 	topics="Weather:AirTemperatureAndPrecipitation:NOAA",
#' 	geoset="GAUL",geoset_yr=1990,space_unit="adm2",time_unit="month",
#' 	year_min=1990,year_max=1991)
#' out_3
#' }
#' @export
get_data <- function(
    country_names=NULL,
    country_iso3=NULL,
    geoset="GADM",
    geoset_yr=2018,
    space_unit="adm1",
    time_unit="year",
    topics=NULL,
    year_min=1990,
    year_max=2017,
    print_url=TRUE,
    print_time=TRUE,
    error_stop=FALSE,
    by_topic=TRUE,
    skip_missing =TRUE,
    cache_param=FALSE,
    short_message=TRUE){

  ####
  # Error messages
  ####

  if(is.null(country_names)&&is.null(country_iso3)){
    stop("ERROR: Please provide at least one country_names or country_iso3")
  }

  if(is.null(topics)){
    stop("ERROR: Please specify at least one topic")
  }


  ####
  # Define internal functions
  ####

  # Bindings
  .SD <- .N <- DATE <- MID <-TID <- WID <- YEAR <- YRMO <- YRWK <- error <- topic <- country_name_alt <- NULL


  ####
  # Download and integrate data
  ####

  # Country codes
  if (length(country_iso3) == 0 & length(country_names) > 0) {
    country_iso3 <- unique(SUNGEO::cc_dict[match(country_names,SUNGEO::cc_dict[,country_name_alt]),country_iso3])
    country_iso3 <- country_iso3[!is.na(country_iso3)]
  }

  if(by_topic==FALSE){
    # Create query
    url_string <- paste0("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/",paste0(country_iso3,collapse=","),"/",geoset,"/",geoset_yr,"/",toupper(space_unit),"/",toupper(time_unit),"?include=",paste0(topics,collapse=","),"&date=",year_min,"-",year_max,"&cacheEnabled=",cache_param)
    if(print_url==TRUE){
      print(paste0("Fetching: ",gsub("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/","",url_string,fixed=TRUE)))
    }
    # Download, parse JSON, convert to dt
    t1 <- Sys.time()
    json_return <- jsonlite::fromJSON(gsub("\\}\\{","},{", httr::content(httr::GET(url_string), as = "text", encoding = "UTF-8")))
    if("error"%in%names(json_return)&error_stop==TRUE){
      stop(paste0("ERROR ",json_return$error$code,". ",json_return$error$message," (",gsub("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/","",url_string,fixed=TRUE),")"))
    } else if("error"%in%names(json_return)&error_stop==FALSE){
      print(paste0("ERROR ",json_return$error$code," (",gsub("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/","",url_string,fixed=TRUE),")"))
      sungeo.data <- data.table::data.table(error=json_return$error$code,message=json_return$error$message)
    } else if(!"error"%in%names(json_return)){
      sungeo.data <- data.table::as.data.table(data.table::rbindlist(json_return[["data"]],fill=TRUE))
    }

    t2 <- Sys.time(); if(print_time==TRUE){print(t2-t1)}
  } else {
    # Time precision table
    time_invariant <- c("Demographics:Ethnicity:GREG","Infrastructure:Roads:gRoads","Terrain:Elevation:ETOPO1","Terrain:LandCover:GLCC")
    topic_time <- data.table::data.table(topic=c("Demographics:Ethnicity:EPR","Demographics:Ethnicity:GREG","Demographics:Population:GHS","Elections:LowerHouse:CLEA","Events:PoliticalViolence:ABADarfur","Events:PoliticalViolence:ACLED","Events:PoliticalViolence:BeissingerProtest","Events:PoliticalViolence:BeissingerRiot","Events:PoliticalViolence:BeissingerUkraine","Events:PoliticalViolence:COCACW","Events:PoliticalViolence:ESOCAfghanistanWITS","Events:PoliticalViolence:ESOCIraqSIGACT","Events:PoliticalViolence:ESOCIraqWITS","Events:PoliticalViolence:ESOCMexicoDrugRelatedMurders","Events:PoliticalViolence:ESOCMexicoHomicide","Events:PoliticalViolence:ESOCPakistanBFRS","Events:PoliticalViolence:ESOCPakistanWITS","Events:PoliticalViolence:GED","Events:PoliticalViolence:Lankina","Events:PoliticalViolence:NIRI","Events:PoliticalViolence:NVMS","Events:PoliticalViolence:PITF","Events:PoliticalViolence:SCAD","Events:PoliticalViolence:yzCaucasus2000","Events:PoliticalViolence:yzChechnya","Events:PoliticalViolence:yzLibya","Events:PoliticalViolence:yzUkraine2014","Infrastructure:Roads:gRoads","Infrastructure:NightLights:DMSP","PublicHealth:Covid19:JHUCSSEC19","Terrain:Elevation:ETOPO1","Terrain:LandCover:GLCC","Weather:AirTemperatureAndPrecipitation:NOAA"),
                                         time_unit=c("YEAR","YEAR","YEAR","YEAR|MONTH","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR","YEAR","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR|MONTH|WEEK","YEAR","YEAR","YEAR|MONTH|WEEK","YEAR","YEAR","YEAR|MONTH")
    )
    t1 <- Sys.time()
    # tp0 <- topics[1]
    # is0 <- country_iso3[1]
    data_list <- lapply(country_iso3,function(is0){
      topic_list <- lapply(topics,function(tp0){
        # Adjust time precision
        if(!grepl(topic_time[topic%in%tp0,time_unit],time_unit,ignore.case=TRUE)){
          time_unit_ <- topic_time[topic%in%tp0,sapply(stringr::str_split(time_unit,"\\|"),dplyr::last)]
        } else {time_unit_ <- time_unit}
        if(tp0%in%time_invariant){
          url_string <- paste0("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/",is0,"/",geoset,"/",geoset_yr,"/",toupper(space_unit),"/",toupper(time_unit_),"?include=",tp0,"&cacheEnabled=",cache_param)
        } else {
          url_string <- paste0("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/",is0,"/",geoset,"/",geoset_yr,"/",toupper(space_unit),"/",toupper(time_unit_),"?include=",tp0,"&date=",year_min,"-",year_max,"&cacheEnabled=",cache_param)
        }
        if(print_url==TRUE){
          print(paste0("Fetching: ",gsub("https://api-sungeo-org-sungeo-api.apps.gnosis.lsa.umich.edu/data/","",url_string,fixed=TRUE)))
        }
        # Download, parse JSON, convert to dt
        tryCatch({
          json_return_ <- httr::content(httr::GET(url_string), as = "text", encoding = "UTF-8")
          if(grepl("Fatal error",json_return_)){
            json_return <- list()
            json_return$error <- data.table::data.table(code=500,message=gsub("<.*?>|\\n", "", json_return_))
          }else{
            json_return <-  jsonlite::fromJSON(gsub("\\}\\{","},{",json_return_))
          }
        },error=function(e){
          print(paste0("ERROR: ",url_string," ",e))
          json_return <<- list(); json_return$error <<- data.table::data.table(code=500,message=paste0(e))
        })

        # }
        if("error"%in%names(json_return)&error_stop==TRUE){
          if(short_message == TRUE & grepl("not all of the provided topics",json_return$error$message)){
            errm <- unlist(stringr::str_extract_all(json_return$error$message,"\\[.+?\\]"))
            errm <- paste0(grep("\\:",errm,value=TRUE)," missing for ",errm[!grepl("\\:",errm)])
          } else {
            errm <- json_return$error$message
          }
          stop(paste0("ERROR ",json_return$error$code,". ",errm))
        } else if("error"%in%names(json_return)&error_stop==FALSE){
          if(short_message == TRUE & grepl("not all of the provided topics",json_return$error$message)){
            errm <- unlist(stringr::str_extract_all(json_return$error$message,"\\[.+?\\]"))
            errm <- paste0(grep("\\:",errm,value=TRUE)," missing for ",errm[!grepl("\\:",errm)])
          } else {
            errm <- json_return$error$message
          }
          print(paste0("ERROR ",json_return$error$code,", ",errm))
          sungeo.data <- data.table::data.table(error=json_return$error$code,message=errm)
        } else if(!"error"%in%names(json_return)){
          # Parse JSON
          sungeo.data <- data.table::as.data.table(data.table::rbindlist(json_return,fill=TRUE))
          # Add date fields if applicable
          if(any(names(sungeo.data)%in%c("WID","DATE","TID","YRMO","MID","YEAR","YRWK"))){
            ticker <- make_ticker(date_min=paste0(year_min,"0101"),date_max=min(paste0(year_max,"1231"),as.numeric(gsub("-","",as.Date(Sys.Date())))))[,c("YEAR","YRMO","YRWK"):=lapply(.SD,as.character),.SDcols=c("YEAR","YRMO","YRWK")]
            common_var <- intersect(names(ticker),names(sungeo.data))
            if(length(common_var)==1){
              if(grepl("YEAR",time_unit,ignore.case=TRUE)){
                ticker <- ticker[!duplicated(YEAR)&get(common_var)%in%sungeo.data[,get(common_var)]]
              }
              if(grepl("MONTH",time_unit,ignore.case=TRUE)){
                ticker <- ticker[!duplicated(YRMO)&get(common_var)%in%sungeo.data[,get(common_var)]]
              }
              if(grepl("WEEK",time_unit,ignore.case=TRUE)){
                ticker <- ticker[!duplicated(YRWK)&get(common_var)%in%sungeo.data[,get(common_var)]]
              }
              sungeo.data <- merge(data.table::copy(ticker)[,eval(common_var):=as.character(get(common_var))],data.table::copy(sungeo.data)[,eval(common_var):=as.character(get(common_var))],by=common_var,all=TRUE,allow.cartesian=TRUE)
            }
            if(length(common_var)>1){
              if(grepl("YEAR",time_unit,ignore.case=TRUE)){
                common_var <- intersect(c("YEAR"),common_var)
                ticker <- ticker[!duplicated(YEAR)&YEAR%in%sungeo.data[,YEAR]][,.SD,.SD=common_var]
              }
              if(grepl("MONTH",time_unit,ignore.case=TRUE)){
                common_var <- intersect(c("YEAR","YRMO","MID"),common_var)
                ticker <- ticker[!duplicated(YRMO)&YRMO%in%sungeo.data[,YRMO]][,.SD,.SD=common_var]
              }
              if(grepl("WEEK",time_unit,ignore.case=TRUE)){
                common_var <- intersect(c("YEAR","YRMO","MID","YRWK","WID"),common_var)
                ticker <- ticker[!duplicated(YRWK)&WID%in%sungeo.data[,WID]][,.SD,.SD=common_var]
              }
              sungeo.data <- merge(data.table::copy(ticker)[,eval(common_var):=lapply(.SD,as.character),.SDcols=common_var],data.table::copy(sungeo.data)[,eval(common_var):=lapply(.SD,as.character),.SDcols=common_var],by=common_var,all=TRUE,allow.cartesian=TRUE)
            }
          }
        }

        sungeo.data
      })
      # Skip missing topics?
      good_results <- which(sapply(topic_list,function(x){nrow(x)>0&!"error"%in%names(x)}))
      errorz <- which(sapply(topic_list,function(x){"error"%in%names(x)}))
      zeroz <- which(sapply(topic_list,function(x){nrow(x)==0}))
      if(skip_missing == TRUE){
        if(length(good_results)>0){
          if(length(errorz)>0&length(zeroz)>0){
            error_msgz <- unlist(sapply(topic_list[errorz],function(x){x[,substring(sapply(stringr::str_split(paste0(error,". ",message)," in \\'on clause"),"[",1),1,200)]}))
            topic_out <- merge_list(topic_list[good_results])[,message:=paste0("ERRORS: ",paste0(paste0(topics[errorz],". Error code ",error_msgz),collapse=", "),".   ZERO ROWS: ",paste0(topics[zeroz],collapse=", "),".")]
          }
          if(length(errorz)>0&length(zeroz)==0){
            error_msgz <- unlist(sapply(topic_list[errorz],function(x){x[,substring(sapply(stringr::str_split(paste0(error,". ",message)," in \\'on clause"),"[",1),1,200)]}))
            topic_out <- merge_list(topic_list[good_results])[,message:=paste0("ERRORS: ",paste0(paste0(topics[errorz],". Error code ",error_msgz),collapse=", "),".")]
          }
          if(length(errorz)==0&length(zeroz)>0){
            topic_out <- merge_list(topic_list[good_results])[,message:=paste0("ZERO ROWS: ",paste0(topics[zeroz],collapse=", "),".")]
          }
          if(length(errorz)==0&length(zeroz)==0){
            topic_out <- merge_list(topic_list[good_results])
          }
        } else {
          topic_out <- NULL
        }
      }
      topic_out
    })
    if(print_url==TRUE){
      print(paste0("Combining..."))
    }
    sungeo.data <- data.table::rbindlist(data_list,fill=TRUE)
    # Remove redundancies
    suppressWarnings({
      if(grepl("YEAR",time_unit,ignore.case=TRUE)){
        sungeo.data[,c("YRMO","MID","YRWK","WID","DATE","DATE_ALT","TID") := NULL]
      }
      if(grepl("MONTH",time_unit,ignore.case=TRUE)){
        sungeo.data[,c("YRWK","WID","DATE","DATE_ALT","TID") := NULL]
      }
      if(grepl("WEEK",time_unit,ignore.case=TRUE)){
        sungeo.data[,c("DATE","DATE_ALT","TID") := NULL]
      }
    })
    # Remove duplicates
    sungeo.data <- unique(sungeo.data)
    t2 <- Sys.time(); if(print_time==TRUE){print(t2-t1)}
  }

  # Return dt
  return(sungeo.data)
}

#' Line-in-polygon analysis
#'
#' Function for basic geometry calculations on polyline features, within an overlapping destination polygon layer.
#'
#' @param linez Source polyline layer. \code{sf} object.
#' @param polyz Destination polygon layer. Must have identical CRS to \code{linez}. \code{sf} object.
#' @param poly_id Name of unique ID column for destination polygon layer. Character string.
#' @param measurez Desired measurements. Could be any of "length" (sum of line lengths by polygon), "density" (sum of line lengths divided by area of polygon) and/or "distance" (distance from each polygon to nearest line feature). Default is to report all three. Character string or vector of character strings.
#' @param outvar_name Name (root) to be given to output variable. Default is \code{"line"}. Character string.
#' @param unitz Units of measurement (linear). Defaul is \code{"km"}. Character string.
#' @param reproject Temporarily reproject layers to planar projection for geometric operations? Defaul is \code{TRUE}. Logical.
#' @param na_val Value to be assigned to missing values (line lengths and densities only). Defaul is \code{NA}. Logical or list.
#' @param verbose Print status messages and progress? Default is \code{TRUE}. Logical.
#' @return An \code{sf} polygon object, with summary statisics of \code{linez} features aggregated to the geometries of \code{polyz}.
#'
#' If \code{measurez = "lengths"}, contains fields with suffixes
#' \itemize{
##'  \item "\code{_length}". Sum of line lengths within each polygon, in km or other units supplied in \code{unitz}.
##'  }
#' If \code{measurez = "density"}, contains fields with suffixes
#' \itemize{
##'  \item "\code{_length}". Sum of line lengths within each polygon, in km or other units supplied in \code{unitz}.
##'  \item "\code{_area}". Area of each polygon, in km^2 or the square of linear units supplied in \code{unitz}.
##'  \item "\code{_density}". Sum of line lengths divided by area of each polygon, in km/km^2 or other units supplied in \code{unitz}.
##'  }
#' If \code{measurez = "distance"}, contains fields with suffixes
#' \itemize{
##'  \item "\code{_distance}". Distance from each polygon to nearest line feature, in km or other units supplied in \code{unitz}.
##'  }
#' If \code{measurez = c("length","density","distance")} (default), contains all of the above.
#' @import sf
#' @importFrom stats as.dist
#' @importFrom measurements conv_unit
#' @importFrom dplyr select everything
#' @examples
#' # Road lengths, densities and distance from polygon to nearest highway
#' data(hex_05_deu)
#' data(highways_deu1992)
#' out_1 <- line2poly(linez = highways_deu1992,
#'                    polyz = hex_05_deu,
#'                    poly_id = "HEX_ID")
#' plot(out_1["line_length"])
#' plot(out_1["line_density"])
#' plot(out_1["line_distance"])
#'
#' \donttest{
#' # Replace missing road lengths and densities with 0's, rename variables
#' out_2 <- line2poly(linez = highways_deu1992,
#'                    polyz = hex_05_deu,
#'                    poly_id = "HEX_ID",
#'                    outvar_name = "road",
#'                    na_val = 0)
#' plot(out_2["road_length"])
#' plot(out_2["road_density"])
#' plot(out_2["road_distance"])
#' }
#' @export

line2poly <- function(linez,
                      polyz,
                      poly_id,
                      measurez=c("length","density","distance"),
                      outvar_name="line",
                      unitz="km",
                      reproject=TRUE,
                      na_val=NA,
                      verbose=TRUE){


  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  # Find optimal planar projection for map
  # if(reproject){
  #   if(verbose){print("Finding optimal planar projection...")}
  #   suppressWarnings({polyz_tr <- crs_select(polyz = fix_geom(polyz))})
  #   polyz_tr_ <- polyz_tr[["sf"]]
  #   epsg <- polyz_tr[["epsg_best"]]
  #   linez_tr_ <- sf::st_transform(linez,sf::st_crs(paste0("EPSG:",epsg)))
  # }
  if(reproject){
    polyz_tr_ <- utm_select(polyz)
    linez_tr_ <- utm_select(linez)
  }
  if(!reproject){
    polyz_tr_ <- polyz
    linez_tr_ <- linez
  }

  # Set units
  unitz_sq <- paste0(unitz,"2")

  # Overlay
  if(verbose){print("Conducting overlay operations...")}
  lp_o <- suppressWarnings({sf::st_intersection(linez_tr_,polyz_tr_)})
  lp_o$sgeoz_length <- as.numeric(measurements::conv_unit(sf::st_length(lp_o),"m", unitz))
  lp_o <- base::as.data.frame(lp_o)
  lp_o <- stats::aggregate(lp_o$sgeoz_length, by=list(lp_o[,which(colnames(lp_o)==poly_id)]), FUN=sum)
  colnames(lp_o) <- c(poly_id,"sgeoz_length")

  # Distance
  if("distance"%in%measurez){
    polyz_tr_$sgeoz_distance <- as.numeric(measurements::conv_unit(sf::st_distance(polyz_tr_,sf::st_union(linez_tr_)),"m", unitz))
  }

  # Length
  if("length"%in%measurez|"density"%in%measurez){
    polyz_tr_ <- merge(polyz_tr_,dplyr::select(lp_o,all_ofSunGeo(poly_id),"sgeoz_length"),by=poly_id,all.x=TRUE,all.y=FALSE)
    polyz_tr_$sgeoz_length <- replace(polyz_tr_$sgeoz_length,is.na(polyz_tr_$sgeoz_length), na_val)
  }

  # Density
  if("density"%in%measurez){
    polyz_tr_$sgeoz_area <- as.numeric(measurements::conv_unit(sf::st_area(polyz_tr_),"m2", unitz_sq))
    polyz_tr_$sgeoz_density <- replace(polyz_tr_$sgeoz_length/polyz_tr_$sgeoz_area,is.na(polyz_tr_$sgeoz_length/polyz_tr_$sgeoz_area), na_val)
  }

  # Rename variables
  polyz_tr_2 <- dplyr::select(polyz_tr_, all_ofSunGeo(poly_id) ,grep("^sgeoz",names(polyz_tr_),value=TRUE),dplyr::everything())
  colnames(polyz_tr_) <- gsub("^sgeoz",outvar_name,colnames(polyz_tr_))

  # Reproject
  if(reproject){
    if(verbose){print("Restoring original projection...")}
    polyz_ <- sf::st_transform(polyz_tr_,sf::st_crs(polyz))
  }
  if(!reproject){
    polyz_ <- polyz_tr_
  }

  # Output
  return(polyz_)
}

#' Update bounding box of sf object
#'
#' Function to update the coordinates of the bounding box of sf vector data objects (e.g. after cropping or subsetting).
#'
#' @param sfobj Layer to be updated. \code{sf} object.
#' @return \code{sf} object, with corrected bounds.
#' @importFrom sf st_coordinates st_geometry
#' @importFrom data.table data.table as.data.table
#' @examples
#' # Update bbox for subset of sf object
#' data(clea_deu2009)
#' out_1 <- update_bbox(clea_deu2009[clea_deu2009$cst_n%in%c("Berlin"),])
#' out_1
#'
#' # Bounding box of full dataset
#' data.table::as.data.table(clea_deu2009)[,sf::st_bbox(geometry)]
#'
#' # Bounding box of subset (incorrect)
#' data.table::as.data.table(clea_deu2009)[cst_n%in%c("Berlin"),sf::st_bbox(geometry)]
#'
#' # Corrected bounding box
#' data.table::as.data.table(out_1)[,sf::st_bbox(geometry)]
#'
#' @export

update_bbox <- function(
  sfobj
  ){
  # Manually calculate bounds from coordinates
  new_bb <- data.table::as.data.table(sf::st_coordinates(sfobj))[,c(min(X),min(Y),max(X),max(Y))]
  # Rename columns
  names(new_bb) <- c("xmin", "ymin", "xmax", "ymax")
  # Change object class
  attr(new_bb, "class") <- "bbox"
  # Assign to bbox slot of sfobj
  attr(sf::st_geometry(sfobj), "bbox") <- new_bb

  return(sfobj)
}

#' Relative scale and nesting coefficients
#'
#' Function to calculate relative scale and nesting metrics for changes of support from a source polygon layer to an overlapping (but spatially misaligned) destination polygon layer.
#'
#' @param poly_from Source polygon layer. \code{sf} object (polygon or multipolygon).
#' @param poly_to Destination polygon layer. Must have identical CRS to \code{poly_from}. \code{sf} object (polygon or multipolygon).
#' @param metrix Requested scaling and nesting metrics. See "details". Default is "all". Character string or vector of character strings.
#' @param tol_ Minimum area of polygon intersection, in square meters. Default is 0.001. Numeric.
#' @param by_unit Include a by-unit decomposition of requested nesting metrics (if available)? Default is FALSE. Logical.
#' @return Named list, with numeric values for each requested metric in \code{metrix}. If \code{by_unit==TRUE}, last element of list is a data.table, with nesting metrics disaggregated by source unit, where the first column is a row index for the source polygon layer.
#' @details Currently supported metrics (\code{metrix}) include:
#' \itemize{
##'  \item Relative scale ("rs"). Measures whether a change-of-support (CoS) task is one of aggregation or disaggregation, by calculating the share of source units that are smaller than destination units. Its range is from 0 to 1, where values of 1 indicate pure aggregation (all source units are smaller than destination units) and values of 0 indicate no aggregation (all source units are at least as large as destination units). Values between 0 and 1 indicate a hybrid (i.e. some source units are smaller, others are larger than target units).
##'  \item Relative nesting ("rn"). Measures how closely source and destination boundaries align, by calculating the share of source units that cannot be split across multiple destination units. Its range is from 0 to 1, where values of 0 indicate no nesting (every source unit can be split across multiple destination units) and values of 1 indicate full nesting (no source unit can be split across multiple destination units).
##'  \item Relative scale, symmetric ("rs_sym"). Alternative measure of "rs", which ranges from -1 to 1. It calculates a difference between two proportions: the share of source units that is smaller than destination units (i.e. "rs" from standpoint of source units), and the share that is larger (i.e. "rs" from standpoint of destination units). Values of -1 indicate pure disaggregation (all source units are larger than destination units), 1 indicates pure aggregation (all source units are smaller than destination units). Values of 0 indicate that all source units are the same size as target units.
##'  \item Relative nesting, symmetric ("rn_sym"). Alternative measure of "rn", which ranges from -1 to 1. It calculates a difference between two components: the nesting of source units within destination units (i.e. "rn" from standpoint of source units), and the nesting of destination units within source units (i.e. "rn" from standpoint of destination units. Values of 1 indicate that source units are perfectly nested within destination units; -1 indicates that destination units are perfectly nested within source units.
##'  \item Relative scale, alternative ("rs_alt"). Alternative measure of "rs", rescaled as a proportion of destination unit area. This measure can take any value on the real line, with positive values indicating aggregation and negative values indicating disaggregation.
##'  \item Relative nesting, alternative ("rn_alt"). Alternative measure of "rn", which places more weight on areas of maximum overlap. The main difference between this measure and "rn" is its use of the maximum intersection area for each source polygon instead of averaging over the quadratic term. Two sets of polygons are considered nested if one set is completely contained within another, with as few splits as possible. If none or only a sliver of a source polygon area falls outside a single destination polygon, those polygons are "more nested" than a case where half of a source polygon falls in destination polygon A and half falls into another polygon B.
##'  \item Relative scale, conditional ("rs_nn"). Alternative measure of "rs", calculated for the subset of source units that are not fully nested within destination units.
##'  \item Relative nesting, conditional ("rn_nn"). Alternative measure of "rn", calculated for the subset of source units that are not fully nested within destination units.
##'  \item Proportion intact ("p_intact"). A nesting metric that requires no area calculations at all. This measure ranges from 0 to 1, where 1 indicates full nesting (i.e. every source unit is intact/no splits), and 0 indicates no nesting (i.e. no source unit is intact/all are split).
##'  \item Proportion fully nested ("full_nest"). A stricter version of "p_intact". This measure ranges from 0 to 1, where 1 indicates full nesting (i.e. every source unit is intact/no splits AND falls completely inside the destination layer), and 0 indicates no nesting (i.e. no source unit is both intact and falls inside destination layer).
##'  \item Relative overlap ("ro"). Assesses extent of spatial overlap between source and destination polygons. This measure is scaled between -1 and 1. Values of 0 indicate perfect overlap (there is no part of source units that fall outside of destination units, and vice versa). Values between 0 and 1 indicate a "source underlap" (some parts of source polygons fall outside of destination polygons; more precisely, a larger part of source polygon area falls outside destination polygons than the other way around). Values between -1 and 0 indicate a "destination underlap" (some parts of destination polygons fall outside of source polygons; a larger part of destination polygon area falls outside source polygons than the other way around). Values of -1 and 1 indicate no overlap (all source units fall outside destination units, and vice versa). This is a theoretical limit only; the function returns an error if there is no overlap.
##'  \item Gibbs-Martin index of diversification ("gmi"). Inverse of "rn", where values of 1 indicate that every source unit is evenly split across multiple destination units, and 0 indicates that no source unit is split across any destination units.
#' }
#' It is possible to pass multiple arguments to \code{metrix} (e.g. \code{metrix=c("rn","rs")}). The default (\code{metrix="all"}) returns all of the above metrics.
#'
#' The function automatically reprojects source and destination geometries to Lambert Equal Area prior to calculation, with map units in meters.
#'
#' Values of \code{tol_} can be adjusted to increase or decrease the sensitivity of these metrics to small border misalignments. The default value discards polygon intersections smaller than 0.001 square meters in area.
#' @importFrom sf sf_use_s2 st_geometry st_crs st_transform st_coordinates st_as_sfc st_intersection st_bbox st_area st_buffer st_difference st_union st_combine
#' @importFrom data.table data.table as.data.table
#' @importFrom stats median
#' @examples
#' # Calculate all scale and nesting metrics for two sets of polygons
#' data(clea_deu2009)
#' data(hex_05_deu)
#' nest_1 <- nesting(
#'               poly_from = clea_deu2009,
#'               poly_to = hex_05_deu
#'               )
#' nest_1
#'
#' # Calculate just Relative Nesting, in the opposite direction
#' nest_2 <- nesting(
#'               poly_from = hex_05_deu,
#'               poly_to = clea_deu2009,
#'               metrix = "rn"
#'               )
#' nest_2
#' @export

nesting <- function(
  poly_from = NULL,
  poly_to = NULL,
  metrix = "all",
  tol_ = 0.001,
  by_unit=FALSE){

  # Null out undefined variables
  .N <- ID_1_ <- area_in_ <- V1 <- index <- area_1_ <- area_ix_ <-NULL

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  # Check input geometries
  if(is.null(poly_from)|is.null(poly_to)){
    base::stop("poly_from and poly_to geographies must both be sfc_MULTIPOLYGON or sfc_POLYGON.")
  }
  if(!("sfc_MULTIPOLYGON"%in%base::class(sf::st_geometry(poly_from))|"sfc_POLYGON"%in%base::class(sf::st_geometry(poly_from)))|!("sfc_MULTIPOLYGON"%in%base::class(sf::st_geometry(poly_to))|"sfc_POLYGON"%in%base::class(sf::st_geometry(poly_to)))){
    base::stop("poly_from and poly_to geographies must both be sfc_MULTIPOLYGON or sfc_POLYGON.")
  }

  # Select scale and nesting metrics
  if("all"%in%metrix){
    metrix <- c("rs","rn","rs_sym","rn_sym","rs_alt","rn_alt","rs_nn","rn_nn","p_intact","full_nest","ro","gmi")
  }
  metrix <- metrix[metrix%in%c("rs","rn","rs_sym","rn_sym","rs_alt","rn_alt","rs_nn","rn_nn","p_intact","full_nest","ro","gmi")]

  # Ensure layers have same CRS
  if(sf::st_crs(poly_from)!=sf::st_crs(poly_to)){
    poly_from <- sf::st_transform(poly_from,crs=sf::st_crs(poly_to))
  }

  # Reproject to equal area
  if(!grepl("4326",sf::st_crs(poly_from)["input"]$input)){
    poly_from <- sf::st_transform(poly_from,crs=4326)
    poly_to <- sf::st_transform(poly_to,crs=4326)
  }
  xy0 <- data.table::as.data.table(sf::st_coordinates(poly_to))[,lapply(.(get("X"),get("Y")),stats::median)]
  xy0 <- data.table::setnames(xy0, c("X","Y"))
  poly_from <- sf::st_transform(poly_from, crs = paste0("+proj=laea +lon_0=",xy0[,get("X")]," +lat_0=",xy0[,get("Y")]))
  poly_to <- sf::st_transform(poly_to, crs = paste0("+proj=laea +lon_0=",xy0[,get("X")]," +lat_0=",xy0[,get("Y")]))

  # Stop if no overlap
  if(length(sf::st_intersection(sf::st_as_sfc(sf::st_bbox(poly_from)),sf::st_as_sfc(sf::st_bbox(poly_to))))==0){
    base::stop("Zero overlap between polygons.")
  } else {

    # Assign temporary IDs
    poly_from$ID_1_ <- 1:nrow(poly_from)
    poly_to$ID_2_ <- 1:nrow(poly_to)

    # Calculate area of original units
    poly_from$area_1_ <- sf::st_area(poly_from)
    poly_to$area_2_ <- sf::st_area(poly_to)

    # Intersect
    suppressWarnings({
      o_sf <- sf::st_intersection(poly_from,poly_to)[,c("ID_1_","ID_2_","area_1_","area_2_")]
    })
    o_sf$area_ix_ <- sf::st_area(o_sf)
    o_sf$area_in_ <- o_sf$area_ix_/o_sf$area_1_

    # Create unit matrix
    if(by_unit==TRUE){
      nesting_dt <- data.table::data.table(index = 1:nrow(poly_from))
    }

    # Relative scale
    if("rs"%in%metrix){
      rs <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::mean(1*(get("area_1_")<get("area_2_")),na.rm=TRUE)]
      if(by_unit==TRUE){
        rs_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::mean(1*(get("area_1_")<get("area_2_")),na.rm=TRUE),by="ID_1_"]
        nesting_dt[,rs := rs_byun[match(nesting_dt[,index],rs_byun[,ID_1_]),V1]]
      }
    }

    # Relative nesting
    if("rn"%in%metrix|"rn_sym"%in%metrix){
      rn <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"][,base::mean(get("V1"),na.rm=TRUE)]
      if(by_unit==TRUE){
        rn_byun <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"]
        nesting_dt[,rn := rn_byun[match(nesting_dt[,index],rn_byun[,ID_1_]),V1]]
      }
    }

    # Relative scale, symmetric
    if("rs_sym"%in%metrix){
      rs_sym <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::mean(1*(get("area_1_")<get("area_2_"))-1*(get("area_1_")>get("area_2_")),na.rm=TRUE)]
    }

    # Relative nesting, symmetric
    if("rn_sym"%in%metrix){
      suppressWarnings({
        o_sf_2 <- sf::st_intersection(poly_to,poly_from)[,c("ID_1_","ID_2_","area_1_","area_2_")]
      })
      o_sf_2$area_ix_ <- sf::st_area(o_sf_2)
      o_sf_2$area_in_ <- o_sf_2$area_ix_/o_sf_2$area_2_
      rn_sym <- data.table::as.data.table(o_sf_2)[as.numeric(get("area_in_"))>tol_,rn - base::sum(as.numeric(get("area_ix_")/get("area_2_"))^2),by="ID_2_"][,base::mean(get("V1"))]
    }

    # Relative scale, alternative
    if("rs_alt"%in%metrix){
      rs_alt <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,as.numeric(base::mean((get("area_2_")-get("area_1_"))/base::mean(get("area_2_")),na.rm=TRUE))]
      if(by_unit==TRUE){
        rs_alt_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,as.numeric(base::mean((get("area_2_")-get("area_1_"))/base::mean(get("area_2_")),na.rm=TRUE)),by="ID_1_"]
        nesting_dt[,rs_alt := rs_alt_byun[match(nesting_dt[,index],rs_alt_byun[,ID_1_]),V1]]
      }
    }

    # Relative nesting, alternative
    if("rn_alt"%in%metrix){
      rn_alt <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,as.numeric(base::max(area_in_)),by="ID_1_"][,mean(V1)]
      if(by_unit==TRUE){
        rn_alt_byun <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,as.numeric(base::max(area_in_)),by="ID_1_"]
        nesting_dt[,rn_alt := rn_alt_byun[match(nesting_dt[,index],rn_alt_byun[,ID_1_]),V1]]
      }
    }

    # Relative scale, conditional
    if("rs_nn"%in%metrix){
      full_nest_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1*(as.numeric(sum(area_ix_)/mean(area_1_))==1&.N==1),by=ID_1_]
      nn <- full_nest_byun[data.table::data.table(o_sf)[,match(ID_1_,full_nest_byun[,ID_1_])],V1]!=1
      rs_nn <- data.table::data.table(o_sf)[nn&as.numeric(get("area_in_"))>tol_,base::mean(1*(get("area_1_")<get("area_2_")),na.rm=TRUE)]
      if(by_unit==TRUE){
        rs_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::mean(1*(get("area_1_")<get("area_2_")),na.rm=TRUE),by="ID_1_"]
        nesting_dt[,rs_nn := rs_byun[match(nesting_dt[,index],rs_byun[,ID_1_]),V1]]
        nesting_dt[index%in%full_nest_byun[V1==1,ID_1_], rs_nn := NA_real_]
      }
    }

    # Relative nesting, conditional
    if("rn_nn"%in%metrix){
      full_nest_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1*(as.numeric(sum(area_ix_)/mean(area_1_))==1&.N==1),by=ID_1_]
      nn <- full_nest_byun[data.table::data.table(o_sf)[,match(ID_1_,full_nest_byun[,ID_1_])],V1]!=1
      rn_nn <- data.table::as.data.table(o_sf)[nn&as.numeric(get("area_in_"))>tol_,base::sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"][,base::mean(get("V1"),na.rm=TRUE)]
      if(by_unit==TRUE){
        rn_byun <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"]
        nesting_dt[,rn_nn := rn_byun[match(nesting_dt[,index],rn_byun[,ID_1_]),V1]]
        nesting_dt[index%in%full_nest_byun[V1==1,ID_1_], rn_nn := NA_real_]
      }
    }

    # Proportion intact
    if("p_intact"%in%metrix){
      p_intact <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,base::mean(base::table(get("ID_1_"))==1)]
      if(by_unit==TRUE){
        p_intact_byun <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1*(.N==1),by="ID_1_"]
        nesting_dt[,p_intact := p_intact_byun[match(nesting_dt[,index],p_intact_byun[,ID_1_]),V1]]
      }
    }

    # Proportion fully nested
    if("full_nest"%in%metrix){
        full_nest_byun <- data.table::data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1*(as.numeric(sum(area_ix_)/mean(area_1_))==1&.N==1),by=ID_1_]
        full_nest <- full_nest_byun[,mean(V1)]
      if(by_unit==TRUE){
        nesting_dt[,full_nest := full_nest_byun[match(nesting_dt[,index],full_nest_byun[,ID_1_]),V1]]
      }
    }

    # Relative overlap
    if("ro"%in%metrix){
      suppressWarnings({
        suppressMessages({
          area_all_1 <- sum(sf::st_area(sf::st_buffer(poly_from,0)))
          area_under_1 <- sum(sf::st_area(sf::st_difference(sf::st_buffer(poly_from,0), sf::st_buffer(sf::st_union(sf::st_combine(sf::st_buffer(poly_to,0))),0))))
          area_all_2 <- sum(sf::st_area(sf::st_buffer(poly_to,0)))
          area_under_2 <- sum(sf::st_area(sf::st_difference(sf::st_buffer(poly_to,0), sf::st_buffer(sf::st_union(sf::st_combine(sf::st_buffer(poly_from,0))),0))))
          ro <- (1-as.numeric(area_under_2/area_all_2))-(1-as.numeric(area_under_1/area_all_1))
        })
      })
    }

    # Gibbs-Martin index
    if("gmi"%in%metrix){
      gmi <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1-sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"][,base::mean(get("V1"))]
      if(by_unit==TRUE){
        gmi_byun <- data.table::as.data.table(o_sf)[as.numeric(get("area_in_"))>tol_,1-sum(as.numeric(get("area_ix_")/get("area_1_"))^2),by="ID_1_"][,base::mean(get("V1")),by="ID_1_"]
        nesting_dt[,gmi := gmi_byun[match(nesting_dt[,index],gmi_byun[,ID_1_]),V1]]
      }
    }

    # Output
    out_list <- lapply(metrix,function(m0){
      get(m0)
    })
    names(out_list) <- metrix
    if(by_unit==TRUE){
      out_list <- append(out_list,list(nesting_dt))
      names(out_list)[length(out_list)] <- "by_unit"
    }
    return(out_list)

  }
}

#' Batch geocode addresses and coordinates with OpenStreetMap
#'
#' Finds geographic coordinates of multiple addresses and place names (forward
#' geocoding), or converts multiple longitude/latitude coordinate pairs to place
#' names and administrative units (reverse geocoding), using OpenStreetMap's
#' Nominatim API.
#'
#' @param query Addresses or place names to be geocoded. Character string.
#'   Ignored if \code{reverse=TRUE}.
#' @param delay Delay between requests, in seconds. Default is 1. Numeric.
#' @param match_num If query matches multiple locations, which match to return?
#'   Default is 1 (highest-ranking match, by relevance). Numeric.
#' @param return_all Should all matches be returned? Overrides \code{match_num}
#'   if \code{TRUE}. Default is \code{FALSE}. Logical.
#' @param details Should detailed results be returned? Default is \code{FALSE}.
#'   Logical. Ignored if \code{reverse=TRUE}.
#' @param reverse Should reverse geocoding be performed (coordinates to
#'   address)? Default is \code{FALSE}. Logical.
#' @param lon Longitudes of points to reverse geocode. Required if
#'   \code{reverse=TRUE}. Must be the same length as \code{lat}. Numeric vector.
#' @param lat Latitudes of points to reverse geocode. Required if
#'   \code{reverse=TRUE}. Must be the same length as \code{lon}. Numeric vector.
#' @param zoom Zoom level for reverse geocoding, controlling the level of detail
#'   returned (0 = country, 10 = city, 18 = building). Default is 10. Numeric.
#' @param user_agent Valid User-Agent identifying the application for
#'   OSM Nominatim. If none supplied, function will attempt to auto-detect.
#'   Character string.
#' @param verbose Print status messages and progress? Default is \code{FALSE}.
#'   Logical.
#' @return A \code{data.frame} object.
#'
#'   For forward geocoding (\code{reverse=FALSE}):
#'   \itemize{
#'   \item \code{query}. User-supplied address query. Character string.
#'   \item \code{osm_id}. OpenStreetMap ID. Character string.
#'   \item \code{address}. OpenStreetMap address. Character string.
#'   \item \code{longitude}. Horizontal coordinate. Numeric.
#'   \item \code{latitude}. Vertical coordinate. Numeric.
#'   }
#'   If \code{details=TRUE}, also contains:
#'   \itemize{
#'   \item \code{osm_type}. OpenStreetMap feature type. Character string.
#'   \item \code{importance}. Relevance of Nominatim match to query, from 0
#'     (worst) to 1 (best). Numeric.
#'   \item \code{bbox_ymin}. Minimum vertical coordinate of bounding box. Numeric.
#'   \item \code{bbox_ymax}. Maximum vertical coordinate of bounding box. Numeric.
#'   \item \code{bbox_xmin}. Minimum horizontal coordinate of bounding box. Numeric.
#'   \item \code{bbox_xmax}. Maximum horizontal coordinate of bounding box. Numeric.
#'   }
#'
#'   For reverse geocoding (\code{reverse=TRUE}):
#'   \itemize{
#'   \item \code{osm_name}. Full display name from OpenStreetMap. Character string.
#'   \item \code{osm_adm0_code}. ISO country code. Character string.
#'   \item \code{osm_adm0}. Country name. Character string.
#'   \item \code{osm_adm1}. State or first-level administrative division. Character string.
#'   \item \code{osm_adm2}. District or county. Character string.
#'   \item \code{osm_adm3}. Municipality. Character string.
#'   \item \code{osm_adm4}. Borough or village. Character string.
#'   }
#' @details Wrapper function for \code{\link[SUNGEO]{geocode_osm}}. Because the
#'   Nominatim Usage Policy stipulates an absolute maximum of 1 request per
#'   second, this function facilitates batch geocoding by adding a small delay
#'   between queries
#'   (\url{https://operations.osmfoundation.org/policies/nominatim/}).
#' @importFrom dplyr bind_rows
#' @examples
#' \donttest{
#' # Geocode multiple addresses (top matches only)
#' geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"))
#' }
#' \donttest{
#' # With progress reports
#' geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"), verbose = TRUE)
#' }
#' \donttest{
#' # Return detailed results for all matches
#' geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"),
#'   details = TRUE, return_all = TRUE)
#' }
#' \donttest{
#' # Reverse geocode multiple coordinates (city level)
#' geocode_osm_batch(
#'   reverse = TRUE,
#'   lon = c(-83.743, -84.556, -82.999),
#'   lat = c(42.278,  42.701,  39.961),
#'   zoom = 10
#' )
#' }
#' @export

geocode_osm_batch <- function(
    query      = NULL,
    delay      = 1,
    return_all = FALSE,
    match_num  = 1,
    details    = FALSE,
    reverse    = FALSE,
    lon        = NULL,
    lat        = NULL,
    zoom       = 10,
    user_agent = NULL,
    verbose    = FALSE
) {

  # Enforce minimum delay time
  if (delay < 1) {
    delay <- 1
    warning("OSM Nominatim Usage Policy requires a maximum of 1 request per second.")
  }

  # Reverse geocoder

  if (isTRUE(reverse)) {
    if (is.null(lon) || is.null(lat)) {
      stop("When reverse=TRUE, both `lon` and `lat` must be supplied.")
    }
    if (length(lon) != length(lat)) {
      stop("`lon` and `lat` must be the same length.")
    }

    osm_mat <- as.data.frame(
      dplyr::bind_rows(
        lapply(seq_along(lon), function(i) {
          geo_i <- data.frame(
            osm_name      = NA_character_,
            osm_adm0_code = NA_character_,
            osm_adm0      = NA_character_,
            osm_adm1      = NA_character_,
            osm_adm2      = NA_character_,
            osm_adm3      = NA_character_,
            osm_adm4      = NA_character_,
            stringsAsFactors = FALSE
          )
          tryCatch({
            geo_i <- geocode_osm(
              reverse    = TRUE,
              lon        = lon[i],
              lat        = lat[i],
              zoom       = zoom,
              user_agent = user_agent
            )
          }, error = function(e) {})
          Sys.sleep(delay)
          if (verbose) {
            print(paste0(i, "/", length(lon), " ", paste0(geo_i$osm_name, collapse = "; ")))
          }
          geo_i
        })
      )
    )
    return(osm_mat)
  }

  # Forward geocoder

  osm_mat <- as.data.frame(
    dplyr::bind_rows(
      lapply(seq_along(query), function(i) {
        geo_i <- data.frame(
          query      = query[i],
          osm_id     = NA_character_,
          osm_type   = NA_character_,
          importance = NA_real_,
          address    = NA_character_,
          longitude  = NA_real_,
          latitude   = NA_real_,
          bbox_ymin  = NA_real_,
          bbox_ymax  = NA_real_,
          bbox_xmin  = NA_real_,
          bbox_xmax  = NA_real_,
          stringsAsFactors = FALSE
        )
        tryCatch({
          geo_i <- geocode_osm(
            query[i],
            match_num  = match_num,
            return_all = return_all,
            details    = details,
            user_agent = user_agent
          )
        }, error = function(e) {})
        Sys.sleep(delay)
        if (verbose) {
          print(paste0(i, "/", length(query), " ", paste0(geo_i$address, collapse = "; ")))
        }
        geo_i
      })
    )
  )

  return(osm_mat)
}

#' Area and population weighted polygon-to-polygon interpolation
#'
#' Function for interpolating values from a source polygon layer to an overlapping (but spatially misaligned) destination polygon layer, using area and/or population weights.
#'
#' @param poly_from Source polygon layer. \code{sf} object.
#' @param poly_to Destination polygon layer. Must have identical CRS to \code{poly_from}. \code{sf} object.
#' @param poly_to_id Name of unique ID column for destination polygon layer. Character string.
#' @param geo_vor Voronoi polygons object (used internally by \code{point2poly_tess}). \code{sf} object.
#' @param methodz Area interpolation method(s). Could be either of "aw" (areal weighting, default) and/or "pw" (population weighting). See "details". Character string or vector of character strings.
#' @param char_methodz Interpolation method(s) for character strings. Could be either of "aw" (areal weighting, default) or "pw" (population weighting). See "details". Character string.
#' @param pop_raster Population raster to be used for population weighting, Must be supplied if \code{methodz="pw"}. Must have identical CRS to \code{poly_from}. \code{raster} or \code{SpatRaster} object.
#' @param varz Names of numeric variable(s) to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.
#' @param funz Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a numeric vector \code{x} and vector of weights \code{w}. Function or list of functions.
#' @param pycno_varz  Names of spatially extensive numeric variables for which the pycnophylactic (mass-preserving) property should be preserved. Character string or vector of character strings.
#' @param char_varz  Names of character string variables to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.
#' @param char_assign Assignment rule to be used for variables specified in \code{char_varz}. Could be either "biggest_overlap" (default) or "all_overlap". See "details". Character string or vector of character strings.
#' @param seed Seed for generation of random numbers. Default is 1. Numeric.
#' @return \code{sf} polygon object, with variables from \code{poly_from} interpolated to the geometries of \code{poly_to}.
#' @details Currently supported integration methods (\code{methodz}) include:
#' \itemize{
##'  \item Areal weighting ("aw"). Values from \code{poly_from} weighted in proportion to relative area of spatial overlap between source features and geometries of \code{poly_to}.
##'  \item Population weighting ("pw"). Values from \code{poly_from} weighted in proportion to relative population sizes in areas of spatial overlap between source features and geometries of \code{poly_to}. This routine uses a third layer (supplied in \code{pop_raster}) to calculate the weights.
##' }
#' It is possible to pass multiple arguments to \code{methodz} (e.g. \code{methodz=c("aw","pw")}), in which case the function will calculate both sets of weights, and append the resulting columns to the output.
#'
#' Interpolation procedures are handled somewhat differently for numeric and character string variables. For numeric variables supplied in \code{varz}, "aw" and/or "pw" weights are passed to the function specified in \code{funz}. If different sets of numeric variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).
#'
#' For character string (and any other) variables supplied in \code{char_varz}, "aw" and/or "pw" weights are passed to the assignment rule(s) specified in \code{char_assign}. Note that the \code{char_varz} argument may include numerical variables, but \code{varz} cannot include character string variables.
#'
#' Currently supported assignment rules for character strings (\code{char_assign}) include:
#' \itemize{
##'  \item "biggest_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned a single value from overlapping \code{poly_from} features, corresponding to the intersection with largest area and/or population weight.
##'  \item "all_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned all values from overlapping \code{poly_from} features, ranked by area and/or population weights (largest-to-smallest) of intersections.
##' }
#' It is possible to pass multiple arguments to \code{char_assign} (e.g. \code{char_assign=c("biggest_overlap","all_overlap")}), in which case the function will calculate both, and append the resulting columns to the output.
#' @import sf
#' @importFrom stats as.dist weighted.mean
#' @importFrom data.table data.table rbindlist as.data.table
#' @importFrom terra extract
#' @importFrom methods as
#' @importFrom dplyr full_join left_join
#' @importFrom purrr reduce
#' @examples
#' # Interpolation of a single variable, with area weights
#' data(clea_deu2009)
#' data(hex_05_deu)
#' out_1 <- poly2poly_ap(poly_from = clea_deu2009,
#'               poly_to = hex_05_deu,
#'               poly_to_id = "HEX_ID",
#'               varz = "to1"
#'              )
#' plot(out_1["to1_aw"])
#'
#' \donttest{
#' # Interpolation of multiple variables, with area weights
#' out_2 <- poly2poly_ap(
#'               poly_from = clea_deu2009,
#'               poly_to = hex_05_deu,
#'               poly_to_id = "HEX_ID",
#'               varz = list(
#'                 c("to1","pvs1_margin"),
#'                 c("vv1") ),
#'               pycno_varz = "vv1",
#'               funz = list(
#'                 function(x,w){stats::weighted.mean(x,w)},
#'                 function(x,w){sum(x*w)} ),
#'               char_varz = c("incumb_pty_n","win1_pty_n")
#'              )
#' plot(out_2["vv1_aw"])
#' }
#'
#'\donttest{
#' # Interpolation of a single variable, with population weights
#' data(gpw4_deu2010)
#' gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
#' out_3 <- poly2poly_ap(poly_from = clea_deu2009,
#'                          poly_to = hex_05_deu,
#'                          poly_to_id = "HEX_ID",
#'                          varz = "to1",
#'                          methodz = "pw",
#'                          pop_raster = gpw4_deu2010)
#' plot(out_3["to1_pw"])
#' }
#'
#' \donttest{
#' # Interpolation of a single variable, with area and population weights
#' out_4 <- poly2poly_ap(poly_from = clea_deu2009,
#'                          poly_to = hex_05_deu,
#'                          poly_to_id = "HEX_ID",
#'                          varz = "to1",
#'                          methodz = c("aw","pw"),
#'                          pop_raster = gpw4_deu2010)
#' plot(out_4["to1_aw"])
#' }
#' @export



poly2poly_ap <- function(
    poly_from,
    poly_to,
    poly_to_id,
    geo_vor = NULL,
    methodz="aw",
    char_methodz = "aw",
    pop_raster=NULL,
    varz=NULL,
    pycno_varz=NULL,
    char_varz=NULL,
    char_assign="biggest_overlap",
    funz=function(x,w){stats::weighted.mean(x,w,na.rm=TRUE)},
    seed = 1){

  set.seed(seed)

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  ###########################################################
  #Section A - Preparing the Gemetries for Tessalation Prcess
  ###########################################################
  # Put variables and functions into list
  #Part i -
  if(length(varz) != length(funz) && is.list(varz) && is.list(funz)){
    stop("ERROR: Length of variable list does not equal length of function list. Pleasse ensure that each set of variables correspsonds to a specific function argument.")
  }

  #Part ii -
  if(inherits(varz,"character")){varz <- list(varz)}

  #Part iii -
  if(!inherits(funz,"list")){funz <- list(funz)}
  funz <- lapply(funz, function(sub_iter){
    if(class(sub_iter)%in%"function" == FALSE && class(sub_iter)%in%'character'){
      funzInt <- get(sub_iter, mode = 'function')
      return(funzInt)
    } else {
      return(sub_iter)
    }
  })

  #
  #
  #

  # Stop if no population raster
  if("pw"%in%methodz & length(pop_raster)==0){stop("No population raster provided.")}

  # Area-weights (part 1)
  if("aw"%in%methodz|"aw"%in%char_methodz){
    # Calculate polygon areas
    poly_from$AREA_TOTAL_FROM <- as.numeric(sf::st_area(poly_from))
    poly_to$AREA_TOTAL_TO <- as.numeric(sf::st_area(poly_to))
  }

  # Population-weights (part 1)
  if("pw"%in%methodz){
    # Calculate polygon total
    poly_from$POP_TOTAL_FROM <- terra::extract(pop_raster, terra::vect(poly_from), fun=sum, na.rm=TRUE)[,2]
    poly_to$POP_TOTAL_TO     <- terra::extract(pop_raster, terra::vect(poly_to),   fun=sum, na.rm=TRUE)[,2]
  }

  #
  #
  #
  #

  ###########################################################
  #Section B -
  ###########################################################
  #Part i -
  poly_to$Return_ID <- 1:nrow(poly_to)

  #Part ii - Intersection
  suppressWarnings({
    suppressMessages({
      int_1 <- sf::st_buffer(sf::st_intersection(sf::st_buffer(poly_from,dist=0),sf::st_buffer(poly_to,dist=0)),dist=0)
    })
  })

  # Drop empty geometries
  if(any(sf::st_is_empty(int_1))){int_1 <- int_1[!sf::st_is_empty(int_1),]}


  #
  #

  #############################
  #Part iii -
  #############################
  # Area-weights (part 2)
  if("aw"%in%methodz|"aw"%in%char_methodz){
    # Calculate weights
    int_1$AREA_INT <- as.numeric(sf::st_area(int_1))
    int_1$AREA_W_EXV <- int_1$AREA_INT/int_1$AREA_TOTAL_FROM
    int_1$AREA_W_INV <- int_1$AREA_INT/int_1$AREA_TOTAL_TO
  }

  # Population-weights (part 2)
  if("pw"%in%methodz){
    # Calculate weights
    int_1$POP_INT <- terra::extract(pop_raster, terra::vect(int_1), fun=sum, na.rm=TRUE)[,2]
    int_1$POP_W_EXV <- int_1$POP_INT/int_1$POP_TOTAL_FROM
    int_1$POP_W_INV <- int_1$POP_INT/int_1$POP_TOTAL_TO
  }

  #
  #

  #########################
  #Part v - Convert to dt
  #########################
  int_1_dt <- data.table::data.table(int_1) #Convert to Data Table Object
  int_1_dt$geometry <- NULL #Remove Geometry

  #
  #
  ################################################
  #Part v - Interpolate missing population values
  ################################################
  if("pw"%in%methodz){
    ############
    #Section A -
    ############
    if(sum(is.na(int_1_dt$POP_INT))==1){
      #Part a -
      suppressWarnings({
        w <- as.matrix(sf::st_distance(sf::st_centroid(int_1)))
      })


      #Part b -
      diag(w) <- NA

      #Part c -
      int_1_dt$POP_INT[is.na(int_1_dt$POP_INT)] <- median(int_1_dt$POP_INT[t(apply(w, 1, order)[1:min(10,nrow(int_1_dt)), is.na(int_1_dt$POP_INT)])],na.rm=TRUE)

      #Part d -
      int_1_dt$POP_W_EXV[is.na(int_1_dt$POP_W_EXV)] <- min(1,int_1_dt$POP_INT[is.na(int_1_dt$POP_W_EXV)]/int_1_dt$POP_TOTAL_FROM[is.na(int_1_dt$POP_W_EXV)])
      int_1_dt$POP_W_INV[is.na(int_1_dt$POP_W_INV)] <- min(1,int_1_dt$POP_INT[is.na(int_1_dt$POP_W_INV)]/int_1_dt$POP_TOTAL_TO[is.na(int_1_dt$POP_W_INV)])

    }

    ############
    #Section b -
    ############

    if(sum(is.na(int_1_dt$POP_INT))>1){
      #Part a -
      suppressWarnings({
        w <- as.matrix(sf::st_distance(sf::st_centroid(int_1)))
      })

      #Part b -
      diag(w) <- NA

      #Part c -
      int_1_dt$POP_INT[is.na(int_1_dt$POP_INT)] <- apply(t(apply(w, 1, order)[ 1:min(10,nrow(int_1_dt)), is.na(int_1_dt$POP_INT)]),1,function(x){mean(int_1_dt$POP_INT[x],na.rm=TRUE)})

      #Part d -
      int_1_dt$POP_W_EXV[is.na(int_1_dt$POP_W_EXV)] <- min(1,int_1_dt$POP_INT[is.na(int_1_dt$POP_W_EXV)]/int_1_dt$POP_TOTAL_FROM[is.na(int_1_dt$POP_W_EXV)])
      int_1_dt$POP_W_INV[is.na(int_1_dt$POP_W_INV)] <- min(1,int_1_dt$POP_INT[is.na(int_1_dt$POP_W_INV)]/int_1_dt$POP_TOTAL_TO[is.na(int_1_dt$POP_W_INV)])
    }
  }

  #
  #

  #############################
  #Part vi - Aggregate Estimates
  #############################
  if(!is.null(geo_vor)){
    #Part a -
    geo_vor_dt <- geo_vor

    #Part b -
    sf::st_geometry(geo_vor_dt) <- NULL

    #Part B -
    geo_vor_dt <- data.table::data.table('Unique_ID' = geo_vor_dt$Unique_ID)

    #Part c -
    Aggregation_Matrix <- data.table::data.table(dplyr::full_join(geo_vor_dt, int_1_dt, by = 'Unique_ID'))
  } else {
    #Part a -
    poly_from$Unique_ID <- 1:nrow(poly_from)

    #Part b -
    geo_vor_dt <- poly_from

    #Part c -
    sf::st_geometry(geo_vor_dt) <- NULL

    #Part d -
    geo_vor_dt <- data.table::data.table('Unique_ID' = geo_vor_dt$Unique_ID)

    #Part e -
    sf::st_geometry(geo_vor_dt) <- sf::st_geometry(poly_from)

    #Part e -
    suppressWarnings({
      suppressMessages(
        Aggregation_Matrix <- data.table::data.table(sf::st_intersection(geo_vor_dt, int_1))
      )
    })

  }


  #
  #
  #
  #
  #
  #
  #

  ##########################################
  #Section C - Aggregate (numeric variables)
  ##########################################
  #Part i -
  ClassTypes_Variables <- data.table::data.table('Varz' = varz, 'Funz' = funz, 'methodz' = methodz)

  #Part ii -
  ClassTypes_Variables <- lapply(1:length(ClassTypes_Variables$Varz), function(sub_iter) {
    Varz_Int <- unlist(ClassTypes_Variables$Varz[sub_iter])

    Funz_Int <- unlist(ClassTypes_Variables$Funz[sub_iter], recursive = TRUE)

    Methodz_Int <- unlist(ClassTypes_Variables$methodz[sub_iter], recursive = TRUE)

    NAValue <- unlist(ClassTypes_Variables$NAval[sub_iter], recursive = TRUE)

    OutputMatrix <- data.table::data.table('Varz' = Varz_Int, 'Funz' = Funz_Int, 'methodz' = Methodz_Int)

    return(OutputMatrix)

  })
  ClassTypes_Variables <- data.table::rbindlist(ClassTypes_Variables,fill=TRUE)

  #Part iii -
  if(any('pw'%in%char_methodz)){
    CharacterVariables <- data.table::data.table('Varz' = char_varz, 'Funz' = NA, 'methodz' = 'pw')
  } else {
    CharacterVariables <- data.table::data.table('Varz' = char_varz, 'Funz' = NA, 'methodz' = 'aw')
  }

  #Part iv -
  if(is.null(char_varz) == FALSE){
    ClassTypes_Variables <- data.table::rbindlist(list(
      ClassTypes_Variables,
      CharacterVariables
    ),fill=TRUE)
  }

  #Part v -
  ClassTypes_Variables$Character <- FALSE
  ClassTypes_Variables$Character[ClassTypes_Variables$Varz%in%char_varz] <- TRUE

  #

  ###########################################################
  #Section C - Character string variables
  ###########################################################
  if(any(ClassTypes_Variables$Character)){
    #Part i -
    Character_Subset <- ClassTypes_Variables[ClassTypes_Variables$Character%in%TRUE,]

    #Part ii -
    Character_Aggregation_Matrix <- lapply(1:nrow(Character_Subset), function(sub_iter){
      if("biggest_overlap"%in%char_assign){
        #########################
        #Part A - Areal Weighting
        #########################
        if("aw"%in%Character_Subset$methodz[sub_iter]){
          #Part i - Locate Variables
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'AREA_W_INV', Character_Subset$Varz[sub_iter]))

          #Part ii - Subset Columns
          Internal_Matrix <- Aggregation_Matrix[,locVar, with = FALSE]

          #Part iii - Rename Value Column
          names(Internal_Matrix)[!names(Internal_Matrix)%in%c('Return_ID', 'AREA_W_INV')] <- 'var_'

          #Part iv - Collapse Data Frame
          Internal_Matrix <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){paste0(unique(x$var_[x$AREA_W_INV%in%max(x$AREA_W_INV, na.rm = TRUE)]),collapse="|")},simplify=TRUE)
          Internal_Matrix <- data.table::data.table(Return_ID=as.numeric(names(Internal_Matrix)),V1=c(Internal_Matrix))

          #Part v -
          names(Internal_Matrix)[2] <- Character_Subset$Varz[sub_iter]


        }

        ##############################
        #Part B - Populating Weighting
        ##############################
        if("pw"%in%Character_Subset$methodz[sub_iter]){
          #Part i - Locate Variables
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'POP_W_INV', Character_Subset$Varz[sub_iter]))

          #Part ii - Subset Columns
          Internal_Matrix <- Aggregation_Matrix[,locVar, with = FALSE]

          #Part iii - Rename Value Column
          names(Internal_Matrix)[names(Internal_Matrix)%in%c('Return_ID', 'POP_W_INV') == FALSE] <- 'var_'

          #Part iv - Collapse Data Frame
          Internal_Matrix <- Internal_Matrix[,list(paste(unique(var_[POP_W_INV%in%max(POP_W_INV, na.rm = TRUE)]), collapse = '|')), by = list(Return_ID)]

          #Part v -
          names(Internal_Matrix)[2] <- Character_Subset$Varz[sub_iter]


        }

      }

      ##############################################
      #All Collapse - Collapse Values that Intersect
      ##############################################
      if("all_overlap"%in%char_assign){
        #########################
        #Part A - Areal Weighting
        #########################
        if("aw"%in%Character_Subset$methodz[sub_iter]){
          #Part i - Locate Variables
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'AREA_W_INV', Character_Subset$Varz[sub_iter]))

          #Part ii - Subset Columns
          Internal_Matrix <- Aggregation_Matrix[,locVar, with = FALSE]

          #Part iii - Rename Value Column
          names(Internal_Matrix)[names(Internal_Matrix)%in%c('Return_ID', 'AREA_W_INV') == FALSE] <- 'var_'

          #Part iv - Collapse Data Frame
          Internal_Matrix <- Internal_Matrix[,list(paste(unique(var_), collapse = '|')), by = list(Return_ID)]

          #Part v -
          names(Internal_Matrix)[2] <- Character_Subset$Varz[sub_iter]


        }

        ##############################
        #Part B - Populating Weighting
        ##############################
        if("pw"%in%Character_Subset$methodz[sub_iter]){
          #Part i - Locate Variables
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'POP_W_INV', Character_Subset$Varz[sub_iter]))

          #Part ii - Subset Columns
          Internal_Matrix <- Aggregation_Matrix[,locVar, with = FALSE]

          #Part iii - Rename Value Column
          names(Internal_Matrix)[names(Internal_Matrix)%in%c('Return_ID', 'POP_W_INV') == FALSE] <- 'var_'

          #Part iv - Collapse Data Frame
          Internal_Matrix <- Internal_Matrix[,list(paste(unique(var_), collapse = '|')), by = list(Return_ID)]

          #Part v -
          names(Internal_Matrix)[2] <- Character_Subset$Varz[sub_iter]


        }

      }

      #
      #
      #

      return(Internal_Matrix)




    })

    #Part iiii -
    Empty_Sets <- sapply(Character_Aggregation_Matrix, function(x) is.null(x) || !inherits(x, "data.table") || nrow(x) == 0)
    if(TRUE%in%Empty_Sets){
      Character_Aggregation_Matrix <- Character_Aggregation_Matrix[Empty_Sets == FALSE]
    }

    #Part iv -
    Character_Aggregation_Matrix <-  purrr::reduce(Character_Aggregation_Matrix,dplyr::full_join, by = 'Return_ID')

  }


  #
  #
  #
  #
  #
  #
  #

  ###########################################################
  #Section D - Numeric Variables (intensive as default)
  ###########################################################
  #Part i -
  Numeric_Subset <- ClassTypes_Variables[ClassTypes_Variables$Character%in%FALSE,]

  #Part ii -
  Numeric_Aggregation_Matrix <- lapply(1:nrow(Numeric_Subset), function(sub_iter){
    #Part a - Locate Variables
    if("aw"%in%Numeric_Subset$methodz[sub_iter]){
      locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'AREA_W_INV', Numeric_Subset$Varz[sub_iter]))
    }

    #Part a - Locate Variables
    if("pw"%in%Numeric_Subset$methodz[sub_iter]){
      locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'POP_W_INV', Numeric_Subset$Varz[sub_iter]))
    }

    #Part b - Subset Columns
    Internal_Matrix <- data.table::as.data.table(Aggregation_Matrix[,locVar, with = FALSE])

    #Part c - Rename Value Column
    if("aw"%in%Numeric_Subset$methodz[sub_iter]){
      names(Internal_Matrix)[!names(Internal_Matrix)%in%c('Return_ID', 'AREA_W_INV')] <- 'var_'
    }

    if("pw"%in%Numeric_Subset$methodz[sub_iter]){
      names(Internal_Matrix)[!names(Internal_Matrix)%in%c('Return_ID', 'POP_W_INV')] <- 'var_'
    }

    #Part c -
    IntermediateFunction <- Numeric_Subset$Funz[sub_iter][[1]]

    #Part d -
    if("w" %in% names(formals(IntermediateFunction))){
      if("aw"%in%Numeric_Subset$methodz[sub_iter]){
        Output <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){IntermediateFunction(x$var_, w = x$AREA_W_INV)},simplify=TRUE)
        Output <- data.table::data.table(Return_ID=as.numeric(names(Output)),V1=c(Output))
      }

      if("pw"%in%Numeric_Subset$methodz[sub_iter]){
        Output <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){IntermediateFunction(x$var_, w = x$POP_W_INV)},simplify=TRUE)
        Output <- data.table::data.table(Return_ID=as.numeric(names(Output)),V1=c(Output))
      }

    } else {
      suppressWarnings({
        Output <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){IntermediateFunction(x$var_)},simplify=TRUE)
        Output <- data.table::data.table(Return_ID=as.numeric(names(Output)),V1=c(Output))
      })
    }

    #Part e -
    if("pw"%in%Numeric_Subset$methodz[sub_iter]){
      names(Output)[2] <- paste0(Numeric_Subset$Varz[sub_iter], '_pw')
    }

    if("aw"%in%Numeric_Subset$methodz[sub_iter]){
      names(Output)[2] <- paste0(Numeric_Subset$Varz[sub_iter], '_aw')
    }

    #RETURN
    return(Output)
  })

  #Part iii -
  Empty_Sets <- sapply(Numeric_Aggregation_Matrix, function(x){is.null(x) || !inherits(x, "data.table") || nrow(x) == 0})

  if(TRUE%in%Empty_Sets){
    Numeric_Aggregation_Matrix <- Numeric_Aggregation_Matrix[Empty_Sets == FALSE]
  }

  #Part iv -
  Numeric_Aggregation_Matrix <-  purrr::reduce(Numeric_Aggregation_Matrix,dplyr::full_join, by = 'Return_ID')

  #
  #
  #
  #
  #

  ###########################################################
  #Section E - Pycnophylactic variables
  ###########################################################

  if(!is.null(pycno_varz)){
    # Loop over pycno_varz
    for(p0 in 1:length(pycno_varz)){
      # Find sum of original variable
      sum_from <- data.table::as.data.table(poly_from)[,sum(get(pycno_varz[p0]),na.rm=TRUE)]

      # Find matching processed variables in target geometry
      pycno_varz_to <- grep(paste0("^",pycno_varz[p0]),Numeric_Subset[,paste0(Numeric_Subset$Varz,"_",Numeric_Subset$methodz)],value=TRUE)

      #Part i -
      Extensive_Subset <- ClassTypes_Variables[ClassTypes_Variables$Character%in%FALSE&Varz%in%pycno_varz[p0],]

      #Part ii -
      Extensive_Aggregation_Matrix <- lapply(1:nrow(Extensive_Subset), function(sub_iter){

        #Part a - Locate Variables
        if("aw"%in%Extensive_Subset$methodz[sub_iter]){
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'AREA_W_EXV', Extensive_Subset$Varz[sub_iter]))
        }

        #Part a - Locate Variables
        if("pw"%in%Extensive_Subset$methodz[sub_iter]){
          locVar <- which(names(Aggregation_Matrix)%in%c('Return_ID', 'POP_W_EXV', Extensive_Subset$Varz[sub_iter]))
        }

        #Part b - Subset Columns
        Internal_Matrix <- data.table::as.data.table(Aggregation_Matrix[,locVar, with = FALSE])

        #Part c - Rename Value Column
        if("aw"%in%Extensive_Subset$methodz[sub_iter]){
          names(Internal_Matrix)[!names(Internal_Matrix)%in%c('Return_ID', 'AREA_W_EXV')] <- 'var_'
        }

        if("pw"%in%Extensive_Subset$methodz[sub_iter]){
          names(Internal_Matrix)[!names(Internal_Matrix)%in%c('Return_ID', 'POP_W_EXV')] <- 'var_'
        }

        #Part c & d - weighted sum
        if("aw"%in%Extensive_Subset$methodz[sub_iter]){
          Output <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){sum(x$var_*x$AREA_W_EXV)},simplify=TRUE)
          Output <- data.table::data.table(Return_ID=as.numeric(names(Output)),V1=c(Output))
        }
        if("pw"%in%Extensive_Subset$methodz[sub_iter]){
          Output <- by(data=Internal_Matrix,INDICES=Internal_Matrix$Return_ID,FUN=function(x){sum(x$var_*x$POP_W_EXV)},simplify=TRUE)
          Output <- data.table::data.table(Return_ID=as.numeric(names(Output)),V1=c(Output))
        }

        #Part e -
        if("pw"%in%Extensive_Subset$methodz[sub_iter]){
          names(Output)[2] <- paste0(Extensive_Subset$Varz[sub_iter], '_pw')
        }

        if("aw"%in%Extensive_Subset$methodz[sub_iter]){
          names(Output)[2] <- paste0(Extensive_Subset$Varz[sub_iter], '_aw')
        }

        #RETURN
        return(Output)
      })

      #Part iii -
      Empty_Sets <- sapply(Extensive_Aggregation_Matrix, function(x){is.null(x) || !inherits(x, "data.table") || nrow(x) == 0})

      if(TRUE%in%Empty_Sets){
        Extensive_Aggregation_Matrix <- Extensive_Aggregation_Matrix[Empty_Sets == FALSE]
      }

      #Part iv -
      Extensive_Aggregation_Matrix <-  purrr::reduce(Extensive_Aggregation_Matrix,dplyr::full_join, by = 'Return_ID')

      # Rescale variables in target geometry
      for(p00 in 1:length(pycno_varz_to)){
        Extensive_Aggregation_Matrix[,eval(pycno_varz_to[p00]) := get(pycno_varz_to[p00])*sum_from/sum(get(pycno_varz_to[p00]),na.rm = TRUE)]
        # Copy to intensive matrix
        Numeric_Aggregation_Matrix[,eval(pycno_varz_to[p00]):=Extensive_Aggregation_Matrix[,get(pycno_varz_to[p00])]]
      }

      # # Check if numbers align
      # Numeric_Aggregation_Matrix[,sum(get(pycno_varz_to[1]))]
      # Extensive_Aggregation_Matrix[,sum(get(pycno_varz_to[1]))]
      # sum_from
    }
  }


  ###########################################################
  #Section F - Combine
  ###########################################################
  if(any(ClassTypes_Variables$Character)){
    #Part 1 -
    Output_Matrix <- purrr::reduce(list(Character_Aggregation_Matrix, Numeric_Aggregation_Matrix),dplyr::full_join, by = 'Return_ID')
  } else {
    Output_Matrix <- Numeric_Aggregation_Matrix
  }

  #Part 2 -
  polyz_ <- dplyr::left_join(poly_to, Output_Matrix, by = 'Return_ID')

  #Part 3 -
  polyz_$Return_ID <- NULL

  #Part 4 - Output
  polyz_ <- sf::st_cast(polyz_,"MULTIPOLYGON")
  return(polyz_)

}

#' Constituency level results for lower chamber legislative elections, Germany 2009.
#'
#' A simple feature collection containing the spatial geometries of electoral constituency
#' borders, and data on turnout levels, votes shares and other attributes of lower chamber
#' legislative elections.
#'
#' @format Simple feature collection with 16 features and 10 fields.
#' geometry type:  MULTIPOLYGON.
#' dimension:      XY.
#' bbox:           xmin: 5.867281 ymin: 47.27096 xmax: 15.04388 ymax: 55.05902.
#' epsg (SRID):    4326.
#' proj4string:    +proj=longlat +datum=WGS84 +no_defs.
#' \describe{
#'   \item{cst }{Constituency number. Numeric.}
#'   \item{cst_n }{Constituency name. Character.}
#'   \item{ctr }{Country number. Numeric.}
#'   \item{ctr_n }{Country name. Character.}
#'   \item{yrmo }{Year and month of election (YYYYMM). Character.}
#'   \item{to1 }{Turnout in first round. Numeric.}
#'   \item{vv1 }{Number of valid votes in first round. Numeric.}
#'   \item{pvs1_margin }{Popular vote share margin in first round. Numeric.}
#'   \item{incumb_pty_n }{Incumbent party name.}
#'   \item{win1_pty_n }{Party name of popular vote share winner in first round. Character.}
#' }
#' @source Constituency-Level Elections Archive (CLEA) \url{https://electiondataarchive.org/}
"clea_deu2009"

#' Constituency level results for lower chamber legislative elections, Germany 2009.
#'
#' A simple feature collection containing the geographic centroids of electoral
#' contituencies, and data on turnout levels, votes shares and other attributes of
#' lower chamber legislative elections.
#'
#' @format Simple feature collection with 16 features and 10 fields.
#' geometry type:  POINT.
#' dimension:      XY.
#' bbox:           xmin: 6.953882 ymin: 48.54535 xmax: 13.40315 ymax: 54.18635.
#' epsg (SRID):    4326.
#' proj4string:    +proj=longlat +datum=WGS84 +no_defs.
#' \describe{
#'   \item{cst }{Constituency number. Numeric.}
#'   \item{cst_n }{Constituency name. Character.}
#'   \item{ctr }{Country number. Numeric.}
#'   \item{ctr_n }{Country name. Character.}
#'   \item{yrmo }{Year and month of election (YYYYMM). Character.}
#'   \item{to1 }{Turnout in first round. Numeric.}
#'   \item{vv1 }{Number of valid votes in first round. Numeric.}
#'   \item{pvs1_margin }{Popular vote share margin in first round. Numeric.}
#'   \item{incumb_pty_n }{Incumbent party name.}
#'   \item{win1_pty_n }{Party name of popular vote share winner in first round. Character.}
#' }
#' @source Constituency-Level Elections Archive (CLEA) \url{https://electiondataarchive.org/}
"clea_deu2009_pt"

#' Constituency level results for lower chamber legislative elections, Germany 2009.
#'
#' A data.frame object containing the geographic centroids of electoral
#' contituencies, and data on turnout levels, votes shares and other attributes of
#' lower chamber legislative elections.
#'
#' @format data.frame with 16 observations and 12 variables.
#' \describe{
#'   \item{cst }{Constituency number. Numeric.}
#'   \item{cst_n }{Constituency name. Character.}
#'   \item{ctr }{Country number. Numeric.}
#'   \item{ctr_n }{Country name. Character.}
#'   \item{yrmo }{Year and month of election (YYYYMM). Character.}
#'   \item{to1 }{Turnout in first round. Numeric.}
#'   \item{vv1 }{Number of valid votes in first round. Numeric.}
#'   \item{pvs1_margin }{Popular vote share margin in first round. Numeric.}
#'   \item{incumb_pty_n }{Incumbent party name.}
#'   \item{win1_pty_n }{Party name of popular vote share winner in first round. Character.}
#'   \item{longitude }{Longitude of constituency centroid. Numeric.}
#'   \item{latitude }{Latitude of constituency centroid. Numeric.}
#' }
#' @source Constituency-Level Elections Archive (CLEA) \url{https://electiondataarchive.org/}
"clea_deu2009_df"

#' Population count raster for Germany, 2010.
#'
#' 2.5 arc-minute resolution raster of estimates of human population (number of persons per pixel),
#' consistent with national censuses and population registers, for the year 2010.
#'
#' @format
#' class       : SpatRaster
#' dimensions  : 186, 220, 1  (nrow, ncol, nlyr)
#' resolution  : 0.04166667, 0.04166667  (x, y)
#' extent      : 5.875, 15.04167, 47.29167, 55.04167  (xmin, xmax, ymin, ymax)
#' coord. ref. : lon/lat WGS 84 (EPSG:4326)
#' source(s)   : memory
#' name        : gpw_v4_population_count_rev11_2010_2pt5_min
#' min value   :                                        0.00
#' max value   :                                    92915.66
#' @source Gridded Population of the World (GPW) v4: Population Count, v4.11 \doi{10.7927/H4JW8BX5}.
"gpw4_deu2010"

#' Hexagonal grid for Germany.
#'
#' Regular hexagonal grid of 0.5 degree diameter cells, covering territory of Germany (2020 borders).
#'
#' @format Simple feature collection with 257 features and 3 fields.
#' geometry type:  POLYGON.
#' dimension:      XY.
#' bbox:           xmin: 5.375001 ymin: 46.76568 xmax: 15.375 ymax: 55.13726.
#' epsg (SRID):    4326.
#' proj4string:    +proj=longlat +datum=WGS84 +no_defs.
#' \describe{
#'   \item{HEX_ID }{Unique cell identifier. Character.}
#'   \item{HEX_X }{Longitude of cell centroid. Numeric.}
#'   \item{HEX_Y }{Latitude of cell centroid. Numeric.}
#' }
#' @source SUNGEO
"hex_05_deu"

#' Roads polylines for Germany, 1992
#'
#' Roads thematic layer from Digital Chart of the World. Subset: divided multi-lane highways.
#'
#' @format Simple feature collection with 1741 features and 5 fields.
#' geometry type:  MULTILINESTRING.
#' dimension:      XY.
#' bbox:           xmin: 5.750933 ymin: 47.58799 xmax: 14.75109 ymax: 54.80712
#' epsg (SRID):    4326.
#' proj4string:    +proj=longlat +datum=WGS84 +no_defs.
#' \describe{
#'   \item{MED_DESCRI }{Is the road a divided multi-lane highway with a median? Character string.}
#'   \item{RTT_DESCRI }{Primary or secondary route? Character string.}
#'   \item{F_CODE_DES }{Feature code description (road or trail). Character string.}
#'   \item{ISO }{ISO 3166-1 alpha-3 country code. Character string.}
#'   \item{ISOCOUNTRY }{Country name. Character string.}
#' }
#' @source Defense Mapping Agency (DMA), 1992. Digital Chart of the World. Defense Mapping Agency, Fairfax, Virginia. (Four CD-ROMs). Available through DIVA-GIS: \url{https://diva-gis.org/data.html} (accessed May 19, 2026).
"highways_deu1992"

#' Data availability through SUNGEO API
#'
#' Census of geospatial and processed data files available to download using \code{SUNGEO::get_data()}.
#'
#' @format List of 42 data.table objects
#' Geoset:GADM :Classes ‘data.table’ and 'data.frame': 249 obs. of  4 variables
#' Geoset:GAUL :Classes ‘data.table’ and 'data.frame': 242 obs. of  4 variables
#' Geoset:geoBoundaries :Classes ‘data.table’ and 'data.frame': 197 obs. of  4 variables
#' Geoset:GRED :Classes ‘data.table’ and 'data.frame': 74 obs. of  4 variables
#' Geoset:HEXGRID :Classes ‘data.table’ and 'data.frame': 199 obs. of  4 variables
#' Geoset:MPIDR :Classes ‘data.table’ and 'data.frame': 52 obs. of  4 variables
#' Geoset:NHGIS :Classes ‘data.table’ and 'data.frame': 1 obs. of  4 variables
#' Geoset:PRIOGRID :Classes ‘data.table’ and 'data.frame': 199 obs. of  4 variables
#' Geoset:SHGIS :Classes ‘data.table’ and 'data.frame': 68 obs. of  4 variables
#' \describe{
#'   \item{country_iso3 }{Codes for available countries (ISO 3166-1 alpha-3). Character string.}
#'   \item{country_name }{Names of available countries. Character string.}
#'   \item{geoset_years }{Years of available historical boundary files. Character string.}
#'   \item{space_units }{Available spatial units of analysis. Character string.}
#' }
#' Elections:LowerHouse:CLEA :Classes ‘data.table’ and 'data.frame': 168 obs. of  6 variables
#' Demographics:Ethnicity:EPR :Classes ‘data.table’ and 'data.frame': 180 obs. of  6 variables
#' Demographics:Ethnicity:GREG :Classes ‘data.table’ and 'data.frame': 234 obs. of  6 variables
#' Demographics:Population:GHS :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
#' Events:PoliticalViolence:ABADarfur :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ACLED :Classes ‘data.table’ and 'data.frame': 100 obs. of  6 variables
#' Events:PoliticalViolence:BeissingerProtest :Classes ‘data.table’ and 'data.frame': 15 obs. of  6 variables
#' Events:PoliticalViolence:BeissingerRiot :Classes ‘data.table’ and 'data.frame': 15 obs. of  6 variables
#' Events:PoliticalViolence:BeissingerUkraine :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:COCACW :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCAfghanistanWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCIraqSIGACT :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCIraqWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCMexicoDrugRelatedMurders :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCMexicoHomicide :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCPakistanBFRS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:ESOCPakistanWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:GED :Classes ‘data.table’ and 'data.frame': 121 obs. of  6 variables
#' Events:PoliticalViolence:Lankina :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:NIRI :Classes ‘data.table’ and 'data.frame': 12 obs. of  6 variables
#' Events:PoliticalViolence:NVMS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:PITF :Classes ‘data.table’ and 'data.frame': 133 obs. of  6 variables
#' Events:PoliticalViolence:SCAD :Classes ‘data.table’ and 'data.frame': 60 obs. of  6 variables
#' Events:PoliticalViolence:yzCaucasus2000 :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:yzChechnya :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:yzLibya :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Events:PoliticalViolence:yzUkraine2014 :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
#' Infrastructure:Roads:gRoads :Classes ‘data.table’ and 'data.frame': 240 obs. of  6 variables
#' Infrastructure:NightLights:DMSP :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
#' PublicHealth:Covid19:JHUCSSEC19 :Classes ‘data.table’ and 'data.frame': 207 obs. of  6 variables
#' Terrain:Elevation:ETOPO1 :Classes ‘data.table’ and 'data.frame': 256 obs. of  6 variables
#' Terrain:LandCover:GLCC :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
#' Weather:AirTemperatureAndPrecipitation:NOAA :Classes ‘data.table’ and 'data.frame': 209 obs. of  6 variables
#' \describe{
#'   \item{country_iso3 }{Codes for available countries (ISO 3166-1 alpha-3). Character string.}
#'   \item{country_name }{Names of available countries. Character string.}
#'   \item{year_range }{Range of available years for data topic. Character string.}
#'   \item{time_units }{Available time units. Character string.}
#'   \item{space_units }{Available spatial units. Character string.}
#'   \item{geosets }{Names of available geographic boundary data sources. Character string.}
#' }
#' @source Sub-National Geospatial Data Archive System: Geoprocessing Toolkit (updated March 17, 2023).
"available_data"


#' Country code dictionary
#'
#' Reference table of country names and ISO-3166 codes, adapted from \code{countrycode} package.
#'
#' @format data.table object, with	8626 obs. of  3 variables:
#' \describe{
#'   \item{country_name }{Country names. Character string.}
#'   \item{country_name_alt }{Alternative spellings of country names, ASCII characters only. Character string.}
#'   \item{country_iso3 }{Country codes (ISO 3166-1 alpha-3). Character string.}
#' }
#' @source Vincent Arel-Bundock. Package countrycode: Convert Country Names and Country Code, version 1.40. CRAN (October 12, 2022).
"cc_dict"

#' Convert data.frame object into simple features object
#'
#' Function takes in x-, y-coordinates, and a data.frame of variables (optional) and returns an SFC object
#'
#' @param x_coord Numeric vector with longitude or easting projected coordinates. When \code{input_data} or \code{file} is supplied, can be either column name or numeric vector of the same length as \code{nrow(input_data)}.
#' @param y_coord Numeric vector with latitude or northing projected coordinates. Must be equal to the vector length of \code{x_coord}. When \code{input_data} or \code{file} is supplied, can be either column name or numeric vector of the same length as \code{nrow(input_data)}.
#' @param input_data Optional data frame object, containing \code{x_coord} and \code{y_coord}. \code{nrow(input_data)} must be equal to the vector length of \code{x_coord}. NOTE: Rows corresponding to non-usable coordinates are removed from the final output.
#' @param file Optional path to csv file. Overrides \code{input_data}.
#' @param n_max Maximum number of rows to read in \code{file}. Default is \code{Inf}.
#' @param start Number of rows to skip in \code{file}. Default is 0 (start on first row).
#' @param projection_input Projection string associated with \code{x_coord} and \code{y_coord}. Default is \code{'+proj=longlat'}.
#' @param zero.policy If \code{TRUE}, removes rows where corresponding coordinates equals (0,0). Default is \code{FALSE}.
#' @param show_removed If \code{TRUE}, returns a vector of indices corresponding to non-usable coordinates. Default is \code{FALSE}.
#' @return If \code{show_removed==FALSE}, returns an \code{sf} object, with rows corresponding to non-usable coordinates removed. If \code{show_removed==TRUE}, returns a list, with an \code{sf} object (\code{Spatial_Coordinates}), and a vector of indices corresponding to non-usable coordinates removed (\code{Removed_Rows}).
#' @importFrom data.table fread
#' @importFrom sf st_as_sf st_crs st_geometry
#' @examples
#' # Coordinates supplied as vectors
#' data(clea_deu2009_df)
#' out_1 <- df2sf(x_coord=clea_deu2009_df$longitude,y_coord = clea_deu2009_df$latitude)
#' class(out_1)
#' plot(out_1$geometry)
#'
#' # Coordinates supplied as column mames
#' out_2 <- df2sf(x_coord="longitude",y_coord ="latitude", input_data = clea_deu2009_df)
#' plot(out_2["geometry"])
#'
#' # Load from external file
#' \donttest{
#' tmp <- tempfile()
#' write.csv(clea_deu2009_df,file=tmp)
#' out_3 <- df2sf(x_coord="longitude",y_coord ="latitude", file=tmp)
#' plot(out_3["geometry"])
#' }
#' @export
#'

df2sf <- function(x_coord, y_coord, input_data = NULL, file = NULL, n_max = Inf, start = 0, projection_input = 'EPSG:4326', zero.policy = FALSE, show_removed = FALSE){

  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  #Part 1 -
  if(is.null(file) == FALSE){
    input_data <- as.data.frame(data.table::fread(file, nrows = n_max, skip = start))

  }

  #Part 2 -
  if(!is.null(input_data)&(length(x_coord)>1|length(y_coord)>1)){
    message('Warning: When input_data is supplied, x_coord and y_coord should be either column names or numeric vectors of the same length as nrow(input_data)')
  }

  if(is.null(input_data) == FALSE & (length(x_coord)==1&length(y_coord)==1)){
    locationCoords_x <- which(names(input_data)%in%x_coord)

    locationCoords_y <- which(names(input_data)%in%y_coord)

    Coordinate_Matrix <- cbind(input_data[locationCoords_x], input_data[locationCoords_y])

  } else {
    Coordinate_Matrix <- cbind(x_coord, y_coord)

  }

  #Part 3 - Identify NA Results
  if(zero.policy == TRUE){
    Location_Missing_Rows <- which(
      is.na(Coordinate_Matrix[,1]) |
        is.na(Coordinate_Matrix[,2]) |

        is.nan(Coordinate_Matrix[,1]) |
        is.nan(Coordinate_Matrix[,2]) |

        Coordinate_Matrix[,1] == 0 |
        Coordinate_Matrix[,2] == 0)
  } else {
    Location_Missing_Rows <- which(
      is.na(Coordinate_Matrix[,1]) |
        is.na(Coordinate_Matrix[,2]) |

        is.nan(Coordinate_Matrix[,1]) |
        is.nan(Coordinate_Matrix[,2])
    )
  }

  #Part 4 -
  if(length(Location_Missing_Rows) > 0){
    Coordinate_Matrix <- Coordinate_Matrix[-Location_Missing_Rows,]
  }

  #Part 5 -
  Coordinate_Matrix <- as.data.frame(Coordinate_Matrix)

  #Part 6 -
  ST_SFC <- sf::st_as_sf(Coordinate_Matrix, coords = c(1,2))

  #Part 7 -
  sf::st_crs(ST_SFC) <- projection_input

  #Part 8 -
  if(is.null(input_data) == FALSE){
    if(length(Location_Missing_Rows) > 0){
      input_data <- input_data[-Location_Missing_Rows,]

      sf::st_geometry(input_data) <- ST_SFC$geometry
    } else {
      sf::st_geometry(input_data) <- ST_SFC$geometry
    }
  }

  #RETURN
  if(show_removed){
    if(is.null(input_data)){
      return(list('Spatial_Coordinates' = ST_SFC, 'Removed_Rows' = Location_Missing_Rows))
    } else {
      return(list('Spatial_Coordinates' = input_data, 'Removed_Rows' = Location_Missing_Rows))
    }
  } else {
    if(is.null(input_data)){
      return(ST_SFC)
    } else {
      return(input_data)
    }
  }

  rm(Coordinate_Matrix, Location_Missing_Rows, ST_MULTIPOINT, ST_SFC)
  gc()
}

#' Point-to-polygon interpolation, simple overlay method
#'
#' Function for assigning values from a source point layer to a destination polygon layer, using simple point-in-polygon overlays
#'
#' @param pointz Source points layer. \code{sf} object.
#' @param polyz Destination polygon layer. Must have identical CRS to \code{pointz}. \code{sf} object.
#' @param varz Names of variable(s) to be assigned from source polygon layer to destination polygons. Character string or vector of character strings.
#' @param char_varz Names of character string variable(s) in \code{varz}. Character string or vector of character strings.
#' @param funz Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a vector \code{x}. Function or list of functions.
#' @param na_val Value to be assigned to missing values. Defaul is \code{NA}. Logical or list.
#' @param drop_na_cols Drop columns with completely missing values. Defaul is \code{FALSE}. Logical.
#' @return Returns a \code{sf} polygon object, with variables from \code{pointz} assigned to the geometries of \code{polyz}.
#' @details Assignment procedures are the same for numeric and character string variables. All variables supplied in \code{varz} are passed directly to the function specified in \code{funz}. If different sets of variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).
#' @import sf
#' @importFrom data.table data.table rbindlist as.data.table
#' @importFrom methods as
#' @importFrom dplyr full_join left_join
#' @importFrom purrr reduce
#' @examples
#' # Assignment of a single variable (sums)
#' data(hex_05_deu)
#' data(clea_deu2009_pt)
#' out_1 <- point2poly_simp(pointz=clea_deu2009_pt,
#'                          polyz=hex_05_deu,
#'                          varz="vv1")
#' plot(out_1["vv1"])
#'
#' # Replace NA's with 0's
#' out_2 <- point2poly_simp(pointz = clea_deu2009_pt,
#'                          polyz = hex_05_deu,
#'                          varz = "vv1",
#'                          na_val = 0)
#' plot(out_2["vv1"])
#'
#' \donttest{
#' # Multiple variables, with different assignment functions
#' out_3 <- point2poly_simp(pointz = clea_deu2009_pt,
#'                          polyz = hex_05_deu,
#'                          varz = list(
#'                            c("to1","pvs1_margin"),
#'                            c("vv1"),
#'                            c("incumb_pty_n","win1_pty_n")),
#'                          funz = list(
#'                            function(x){mean(x,na.rm=TRUE)},
#'                            function(x){sum(x,na.rm=TRUE)},
#'                            function(x){paste0(unique(na.omit(x)),collapse=" | ") }),
#'                          na_val = list(NA_real_,0,NA_character_))
#' plot(out_3["pvs1_margin"])
#' }
#'
#' @export
#'

point2poly_simp <- function(pointz,
                            polyz,
                            varz,
                            char_varz=NULL,
                            funz=list(function(x){sum(x,na.rm=TRUE)}),
                            na_val=NA,
                            drop_na_cols=FALSE){


  # Turn off s2 processing
  suppressMessages({
    sf::sf_use_s2(FALSE)
  })

  ################################################
  #Part 1 -  Put variables and functions into list
  ################################################
  #Part i -
  if(length(varz) != length(funz) && is.list(varz) && is.list(funz)){
    stop("ERROR: Length of variable list does not equal length of function list. Pleasse ensure that each set of variables correspsonds to a specific function argument.")
  }

  #Part ii -
  if(inherits(varz,"character")){varz <- list(varz)}

  #Part iii -
  if(!inherits(funz,"list")){funz <- list(funz)}
  funz <- lapply(funz, function(sub_iter){
    if({class(sub_iter)%in%"function"} == FALSE && {class(sub_iter)%in%"character"}){
      funzInt <- get(sub_iter, mode = "function")
      return(funzInt)
    } else {
      return(sub_iter)
    }
  })

  #Part iv -
  if(inherits(na_val,"logical")){na_val <- list(na_val)}

  ################################################
  #Part 2:
  ################################################
  if(sf::st_crs(pointz) != sf::st_crs(polyz)){
    #Part i -
    polyz <- sf::st_transform(polyz, crs = sf::st_crs(pointz))
  }

  ############################
  #Part 3 - Empty points layer
  ############################
  #Part i -
  pointz_dt <- data.table::as.data.table(pointz) #Convert to data table object

  #Part ii -
  pointz_dt_NACols <- sapply(pointz_dt, function(x){  #Generate a Vector for Identifying all NA columns
    all(is.na(x))
  })

  #Part iii - drop NA columns (optional)
  if(drop_na_cols==TRUE){
    pointz_dt <- pointz_dt[,pointz_dt_NACols == FALSE, with = FALSE] #Subset columns
  }
  ###################################
  #Part 4 -  Match points to polygons
  ###################################
  #Part i -
  polyz$polygon_id_ <- 1:nrow(polyz)

  #Part ii -
  suppressWarnings({
    suppressMessages(
      Point_Overlay <- sf::st_within(pointz,polyz)
    )
  })

  #Part iii -
  ErrorCheck <- sapply(Point_Overlay, function(x) length(x))

  ##########################################
  #Part 5 - Add polygon index to point layer
  ##########################################
  #Part i -
  pointz_dt$polygon_id_ <- NA
  pointz_dt$polygon_id_[ErrorCheck == 1] <- polyz$polygon_id_[unlist(Point_Overlay[ErrorCheck == 1])]

  #Part ii -
  pointz$polygon_id_ <- NA
  pointz$polygon_id_[ErrorCheck == 1] <- polyz$polygon_id_[unlist(Point_Overlay[ErrorCheck == 1])]

  #######################################
  #Part 6 - Aggregate (numeric variables)
  #######################################
  #Part i -
  ClassTypes_Variables <- data.table::data.table('Varz' = varz, 'Funz' = funz, 'NAval' = na_val)

  #Part ii -
  ClassTypes_Variables <- lapply(1:length(ClassTypes_Variables$Varz), function(sub_iter) {
    Varz_Int <- unlist(ClassTypes_Variables$Varz[sub_iter])

    Funz_Int <- unlist(ClassTypes_Variables$Funz[sub_iter], recursive = TRUE)

    NAValue <- unlist(ClassTypes_Variables$NAval[sub_iter], recursive = TRUE)

    OutputMatrix <- data.table::data.table('Varz' = Varz_Int, 'Funz' = Funz_Int, 'NAval' = NAValue)

    return(OutputMatrix)

  })
  ClassTypes_Variables <- data.table::rbindlist(ClassTypes_Variables)

  #Part iii -
  # sub_iter <- 1
  Aggregation_Matrix <- lapply(1:nrow(ClassTypes_Variables), function(sub_iter){
    #Part a -
    Select_Columns <- pointz_dt[,names(pointz_dt)%in%c(ClassTypes_Variables$Varz[sub_iter], 'polygon_id_'), with = FALSE]

    #Part b -
    names(Select_Columns) <- c('var', 'polygon_id_')

    #Part c -
    IntermediateFunction <- ClassTypes_Variables$Funz[sub_iter][[1]]

    #Part d -
    suppressWarnings({
      Output <- by(data=Select_Columns,INDICES=Select_Columns$polygon_id_,FUN=function(x){IntermediateFunction(x$var)},simplify=TRUE)
      Output <- data.table::data.table(polygon_id_=as.numeric(names(Output)),V1=c(Output))
    })

    #Part e -
    names(Output)[2] <- ClassTypes_Variables$Varz[sub_iter]

    #RETURN
    return(Output)
  })

  #Part iv -
  Empty_Sets <- sapply(Aggregation_Matrix, function(x) is.null(x) || !inherits(x, "data.table") || nrow(x) == 0)
  if(TRUE %in% Empty_Sets){
    Aggregation_Matrix <- Aggregation_Matrix[Empty_Sets == FALSE]
  }

  #Part v -
  Combined_Matrix <- purrr::reduce(Aggregation_Matrix,dplyr::full_join, by = 'polygon_id_')

  #Part vi - Remove NA Rows (Failed Matches)
  if(any(is.na(Combined_Matrix$polygon_id_))){
      Combined_Matrix <- Combined_Matrix[is.na(Combined_Matrix$polygon_id_) == FALSE,]
  }

  #######################################
  #Part 7 -
  #######################################
  #Part i -
  Coordinates <- sf::st_geometry(polyz)
  sf::st_geometry(polyz) <- NULL

  #Part ii -
  polyz_ <- as.data.frame(dplyr::left_join(polyz, Combined_Matrix, by = 'polygon_id_') )

  #Part iii -
  for(iter in 1:length(ClassTypes_Variables$Varz)){
    VectorSubset <- polyz_[,names(polyz_)%in%ClassTypes_Variables$Varz[iter]]

    VectorSubset[is.na(VectorSubset)] <- ClassTypes_Variables$NAval[iter]

    polyz_[,names(polyz_)%in%ClassTypes_Variables$Varz[iter]] <- VectorSubset
  }

  #Part iv -
  sf::st_geometry(polyz_) <- Coordinates

  # Ensure classes are same as in source file (+ coerce char_varz to character)
  classez <- sapply(as.data.frame(pointz,stringsAsFactors=FALSE)[,unique(unlist(varz))],class)
  if(length(char_varz)>0){classez[names(classez)%in%char_varz] <- "character"}
  polyz_ng <- sf::st_drop_geometry(polyz_)
  for(cl0 in 1:length(classez)){polyz_ng[,names(classez[cl0])] <- methods::as(polyz_ng[,names(classez[cl0])],classez[cl0])}
  polyz_ <- sf::st_set_geometry(polyz_ng,polyz_$geometry); rm(polyz_ng)

  #Output
  return(polyz_)
}

#' Smart numerical rounding function
#'
#' Function to round numerical values with minimal information loss (e.g. to avoid "0.000" values in tables).
#'
#' @param x Vector of values to be rounded. Numeric.
#' @param rnd Requested number of decimal places. Default is 0. Non-negative integer.
#' @param return_char Return rounded values as character string? Default is TRUE. Logical.
#' @return If \code{return_char=TRUE}, returns a character string of same length as \code{x}. If \code{return_char=FALSE}, returns a numerical vector of same length as \code{x}.
#' @details Rounds the values in its first argument to the specified number of decimal places (default 0). If brute-force rounding produces zero values (e.g. "0.00"), the number of decimal places is expanded to include the first significant digit.
#' @examples
#' # Round a vector of numbers, character string output (best for tables)
#' out_1 <- smart_round(c(.0013,2.3,-1,pi),rnd=2)
#' out_1
#'
#' # Round a vector of numbers, numerical output
#' out_2 <- smart_round(c(.0013,2.3,-1,pi),rnd=2,return_char=FALSE)
#' out_2
#' @export

smart_round <- function(x,rnd=0,return_char=TRUE){
  x_ <- x
  x_[abs(round(x,rnd))<(1*10^(0-rnd))] <- signif(x[abs(round(x,rnd))<(1*10^(0-rnd))],1)
  x_[abs(round(x,rnd))>=(1*10^(0-rnd))] <- round(x[abs(round(x,rnd))>=(1*10^(0-rnd))],rnd)
  x_[abs(round(x_,rnd+1))>=(1*10^(0-(rnd+1)))] <- round(x_[abs(round(x_,rnd+1))>=(1*10^(0-(rnd+1)))],rnd+1)

  # Change class (if applicable)
  if(return_char==TRUE){
    x_ <- as.character(x_)
  }
  return(x_)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SUNGEO-package.R
\docType{package}
\name{SUNGEO-package}
\alias{SUNGEO-package}
\title{SUNGEO: Sub-National Geospatial Data Archive: Geoprocessing Toolkit}
\description{
Integrates spatially-misaligned GIS datasets across incompatible geographic units. Part of the Sub-National Geospatial Data Archive System. For the underlying methods, see Zhukov, Byers, Davidson, and Kollman (2024) "Integrating Data Across Misaligned Spatial Units," Political Analysis, Volume 32, Number 1, pp. 17-33 \doi{10.1017/pan.2023.5}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://www.sungeo.org/}
  \item \url{https://github.com/zhukovyuri/SUNGEO}
}

}
\author{
\strong{Maintainer}: Yuri M. Zhukov \email{ymz2@georgetown.edu}

Authors:
\itemize{
  \item Yuri M. Zhukov \email{ymz2@georgetown.edu}
  \item Jason Byers
  \item Marty Davidson
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fix_geom.R
\name{fix_geom}
\alias{fix_geom}
\title{Fix polygon geometries}
\usage{
fix_geom(x, n_it = 10)
}
\arguments{
\item{x}{Polygon layer to be checked and fixed. \code{sf} object.}

\item{n_it}{Number of iterations. Default is 10. Numeric..}
}
\value{
Returns a \code{sf} polygon object, with self-intersections and other geometry problems fixed.
}
\description{
Function to check validity and fix broken geometries in simple features polygon objects
}
\examples{
# Fix geometries for a single dataset
data(clea_deu2009)
out_1 <- fix_geom(clea_deu2009)
all(sf::st_is_valid(out_1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_list.R
\name{merge_list}
\alias{merge_list}
\title{Merge list of tables on common variable(s)}
\usage{
merge_list(lst)
}
\arguments{
\item{lst}{List of tables to be merged. List object.}
}
\value{
\code{data.table} object
}
\description{
Function that finds a set of common columns in a list of tables, and merges the tables on these columns.
}
\examples{
# Merge list of three tables with different common variables
A <- data.table::data.table(month=month.name,year=rep(1991:1992,each=12),A=rnorm(24))
B <- data.table::data.table(year=c(1991,1992),B=rbeta(2,1,1))
C <- data.table::data.table(month=month.name,C=runif(12))

out_1 <- merge_list(list(A,B,C))
out_1

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/update_bbox.R
\name{update_bbox}
\alias{update_bbox}
\title{Update bounding box of sf object}
\usage{
update_bbox(sfobj)
}
\arguments{
\item{sfobj}{Layer to be updated. \code{sf} object.}
}
\value{
\code{sf} object, with corrected bounds.
}
\description{
Function to update the coordinates of the bounding box of sf vector data objects (e.g. after cropping or subsetting).
}
\examples{
# Update bbox for subset of sf object
data(clea_deu2009)
out_1 <- update_bbox(clea_deu2009[clea_deu2009$cst_n\%in\%c("Berlin"),])
out_1

# Bounding box of full dataset
data.table::as.data.table(clea_deu2009)[,sf::st_bbox(geometry)]

# Bounding box of subset (incorrect)
data.table::as.data.table(clea_deu2009)[cst_n\%in\%c("Berlin"),sf::st_bbox(geometry)]

# Corrected bounding box
data.table::as.data.table(out_1)[,sf::st_bbox(geometry)]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{clea_deu2009_df}
\alias{clea_deu2009_df}
\title{Constituency level results for lower chamber legislative elections, Germany 2009.}
\format{
data.frame with 16 observations and 12 variables.
\describe{
  \item{cst }{Constituency number. Numeric.}
  \item{cst_n }{Constituency name. Character.}
  \item{ctr }{Country number. Numeric.}
  \item{ctr_n }{Country name. Character.}
  \item{yrmo }{Year and month of election (YYYYMM). Character.}
  \item{to1 }{Turnout in first round. Numeric.}
  \item{vv1 }{Number of valid votes in first round. Numeric.}
  \item{pvs1_margin }{Popular vote share margin in first round. Numeric.}
  \item{incumb_pty_n }{Incumbent party name.}
  \item{win1_pty_n }{Party name of popular vote share winner in first round. Character.}
  \item{longitude }{Longitude of constituency centroid. Numeric.}
  \item{latitude }{Latitude of constituency centroid. Numeric.}
}
}
\source{
Constituency-Level Elections Archive (CLEA) \url{https://electiondataarchive.org/}
}
\usage{
data(clea_deu2009_df)
}
\description{
A data.frame object containing the geographic centroids of electoral
contituencies, and data on turnout levels, votes shares and other attributes of
lower chamber legislative elections.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{hex_05_deu}
\alias{hex_05_deu}
\title{Hexagonal grid for Germany.}
\format{
Simple feature collection with 257 features and 3 fields.
geometry type:  POLYGON.
dimension:      XY.
bbox:           xmin: 5.375001 ymin: 46.76568 xmax: 15.375 ymax: 55.13726.
epsg (SRID):    4326.
proj4string:    +proj=longlat +datum=WGS84 +no_defs.
\describe{
  \item{HEX_ID }{Unique cell identifier. Character.}
  \item{HEX_X }{Longitude of cell centroid. Numeric.}
  \item{HEX_Y }{Latitude of cell centroid. Numeric.}
}
}
\source{
SUNGEO
}
\usage{
data(hex_05_deu)
}
\description{
Regular hexagonal grid of 0.5 degree diameter cells, covering territory of Germany (2020 borders).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/poly2poly_ap.R
\name{poly2poly_ap}
\alias{poly2poly_ap}
\title{Area and population weighted polygon-to-polygon interpolation}
\usage{
poly2poly_ap(
  poly_from,
  poly_to,
  poly_to_id,
  geo_vor = NULL,
  methodz = "aw",
  char_methodz = "aw",
  pop_raster = NULL,
  varz = NULL,
  pycno_varz = NULL,
  char_varz = NULL,
  char_assign = "biggest_overlap",
  funz = function(x, w) {
     stats::weighted.mean(x, w, na.rm = TRUE)
 },
  seed = 1
)
}
\arguments{
\item{poly_from}{Source polygon layer. \code{sf} object.}

\item{poly_to}{Destination polygon layer. Must have identical CRS to \code{poly_from}. \code{sf} object.}

\item{poly_to_id}{Name of unique ID column for destination polygon layer. Character string.}

\item{geo_vor}{Voronoi polygons object (used internally by \code{point2poly_tess}). \code{sf} object.}

\item{methodz}{Area interpolation method(s). Could be either of "aw" (areal weighting, default) and/or "pw" (population weighting). See "details". Character string or vector of character strings.}

\item{char_methodz}{Interpolation method(s) for character strings. Could be either of "aw" (areal weighting, default) or "pw" (population weighting). See "details". Character string.}

\item{pop_raster}{Population raster to be used for population weighting, Must be supplied if \code{methodz="pw"}. Must have identical CRS to \code{poly_from}. \code{raster} or \code{SpatRaster} object.}

\item{varz}{Names of numeric variable(s) to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.}

\item{pycno_varz}{Names of spatially extensive numeric variables for which the pycnophylactic (mass-preserving) property should be preserved. Character string or vector of character strings.}

\item{char_varz}{Names of character string variables to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.}

\item{char_assign}{Assignment rule to be used for variables specified in \code{char_varz}. Could be either "biggest_overlap" (default) or "all_overlap". See "details". Character string or vector of character strings.}

\item{funz}{Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a numeric vector \code{x} and vector of weights \code{w}. Function or list of functions.}

\item{seed}{Seed for generation of random numbers. Default is 1. Numeric.}
}
\value{
\code{sf} polygon object, with variables from \code{poly_from} interpolated to the geometries of \code{poly_to}.
}
\description{
Function for interpolating values from a source polygon layer to an overlapping (but spatially misaligned) destination polygon layer, using area and/or population weights.
}
\details{
Currently supported integration methods (\code{methodz}) include:
\itemize{
 \item Areal weighting ("aw"). Values from \code{poly_from} weighted in proportion to relative area of spatial overlap between source features and geometries of \code{poly_to}.
 \item Population weighting ("pw"). Values from \code{poly_from} weighted in proportion to relative population sizes in areas of spatial overlap between source features and geometries of \code{poly_to}. This routine uses a third layer (supplied in \code{pop_raster}) to calculate the weights.
}
It is possible to pass multiple arguments to \code{methodz} (e.g. \code{methodz=c("aw","pw")}), in which case the function will calculate both sets of weights, and append the resulting columns to the output.

Interpolation procedures are handled somewhat differently for numeric and character string variables. For numeric variables supplied in \code{varz}, "aw" and/or "pw" weights are passed to the function specified in \code{funz}. If different sets of numeric variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).

For character string (and any other) variables supplied in \code{char_varz}, "aw" and/or "pw" weights are passed to the assignment rule(s) specified in \code{char_assign}. Note that the \code{char_varz} argument may include numerical variables, but \code{varz} cannot include character string variables.

Currently supported assignment rules for character strings (\code{char_assign}) include:
\itemize{
 \item "biggest_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned a single value from overlapping \code{poly_from} features, corresponding to the intersection with largest area and/or population weight.
 \item "all_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned all values from overlapping \code{poly_from} features, ranked by area and/or population weights (largest-to-smallest) of intersections.
}
It is possible to pass multiple arguments to \code{char_assign} (e.g. \code{char_assign=c("biggest_overlap","all_overlap")}), in which case the function will calculate both, and append the resulting columns to the output.
}
\examples{
# Interpolation of a single variable, with area weights
data(clea_deu2009)
data(hex_05_deu)
out_1 <- poly2poly_ap(poly_from = clea_deu2009,
              poly_to = hex_05_deu,
              poly_to_id = "HEX_ID",
              varz = "to1"
             )
plot(out_1["to1_aw"])

\donttest{
# Interpolation of multiple variables, with area weights
out_2 <- poly2poly_ap(
              poly_from = clea_deu2009,
              poly_to = hex_05_deu,
              poly_to_id = "HEX_ID",
              varz = list(
                c("to1","pvs1_margin"),
                c("vv1") ),
              pycno_varz = "vv1",
              funz = list(
                function(x,w){stats::weighted.mean(x,w)},
                function(x,w){sum(x*w)} ),
              char_varz = c("incumb_pty_n","win1_pty_n")
             )
plot(out_2["vv1_aw"])
}

\donttest{
# Interpolation of a single variable, with population weights
data(gpw4_deu2010)
gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
out_3 <- poly2poly_ap(poly_from = clea_deu2009,
                         poly_to = hex_05_deu,
                         poly_to_id = "HEX_ID",
                         varz = "to1",
                         methodz = "pw",
                         pop_raster = gpw4_deu2010)
plot(out_3["to1_pw"])
}

\donttest{
# Interpolation of a single variable, with area and population weights
out_4 <- poly2poly_ap(poly_from = clea_deu2009,
                         poly_to = hex_05_deu,
                         poly_to_id = "HEX_ID",
                         varz = "to1",
                         methodz = c("aw","pw"),
                         pop_raster = gpw4_deu2010)
plot(out_4["to1_aw"])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geocode_osm.R
\name{geocode_osm}
\alias{geocode_osm}
\title{Geocode addresses and coordinates with OpenStreetMap}
\usage{
geocode_osm(
  query,
  match_num = 1,
  return_all = FALSE,
  details = FALSE,
  reverse = FALSE,
  lon = NULL,
  lat = NULL,
  zoom = 10,
  user_agent = NULL
)
}
\arguments{
\item{query}{Address or place name to be geocoded. Character string. Ignored
if \code{reverse=TRUE}.}

\item{match_num}{If query matches multiple locations, which match to return?
Default is 1 (highest-ranking match, by relevance). Numeric.}

\item{return_all}{Should all matches be returned? Overrides \code{match_num}
if \code{TRUE}. Default is \code{FALSE}. Logical.}

\item{details}{Should detailed results be returned? Default is \code{FALSE}.
Logical. Ignored if \code{reverse=TRUE}.}

\item{reverse}{Should reverse geocoding be performed (coordinates to address)?
Default is \code{FALSE}. Logical.}

\item{lon}{Longitude of the point to reverse geocode. Required if
\code{reverse=TRUE}. Numeric.}

\item{lat}{Latitude of the point to reverse geocode. Required if
\code{reverse=TRUE}. Numeric.}

\item{zoom}{Zoom level for reverse geocoding, controlling the level of detail
returned (0 = country, 10 = city, 18 = building). Default is 10. Numeric.}

\item{user_agent}{Valid User-Agent identifying the application for
OSM Nominatim. If none supplied, function will attempt to auto-detect.
Character string.}
}
\value{
A \code{data.frame} object.

  For forward geocoding (\code{reverse=FALSE}):
  \itemize{
  \item \code{query}. User-supplied address query. Character string.
  \item \code{osm_id}. OpenStreetMap ID. Character string.
  \item \code{address}. OpenStreetMap address. Character string.
  \item \code{longitude}. Horizontal coordinate. Numeric.
  \item \code{latitude}. Vertical coordinate. Numeric.
  }
  If \code{details=TRUE}, also contains:
  \itemize{
  \item \code{osm_type}. OpenStreetMap feature type. Character string.
  \item \code{importance}. Relevance of Nominatim match to query, from 0
    (worst) to 1 (best). Numeric.
  \item \code{bbox_ymin}. Minimum vertical coordinate of bounding box. Numeric.
  \item \code{bbox_ymax}. Maximum vertical coordinate of bounding box. Numeric.
  \item \code{bbox_xmin}. Minimum horizontal coordinate of bounding box. Numeric.
  \item \code{bbox_xmax}. Maximum horizontal coordinate of bounding box. Numeric.
  }

  For reverse geocoding (\code{reverse=TRUE}):
  \itemize{
  \item \code{osm_name}. Full display name from OpenStreetMap. Character string.
  \item \code{osm_adm0_code}. ISO country code. Character string.
  \item \code{osm_adm0}. Country name. Character string.
  \item \code{osm_adm1}. State or first-level administrative division. Character string.
  \item \code{osm_adm2}. District or county. Character string.
  \item \code{osm_adm3}. Municipality. Character string.
  \item \code{osm_adm4}. Borough, village or street. Character string.
  }
}
\description{
Finds geographic coordinates of addresses and place names (forward geocoding),
or converts longitude/latitude coordinates to place names and administrative
units (reverse geocoding), using OpenStreetMap's Nominatim API.
}
\details{
Note that the Nominatim Usage Policy stipulates an absolute maximum
  of 1 request per second
  (\url{https://operations.osmfoundation.org/policies/nominatim/}).
  For batch geocoding of multiple addresses, please use
  \code{\link[SUNGEO]{geocode_osm_batch}}.
}
\examples{
# Geocode an address (top match only)
geocode_osm("Michigan Stadium")

# Return detailed results for top match
geocode_osm("Michigan Stadium", details = TRUE)

# Return detailed results for all matches
geocode_osm("Michigan Stadium", details = TRUE, return_all = TRUE)

# Reverse geocode a coordinate pair
geocode_osm(reverse = TRUE, lon = -83.74868, lat = 42.26587, zoom = 18)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{highways_deu1992}
\alias{highways_deu1992}
\title{Roads polylines for Germany, 1992}
\format{
Simple feature collection with 1741 features and 5 fields.
geometry type:  MULTILINESTRING.
dimension:      XY.
bbox:           xmin: 5.750933 ymin: 47.58799 xmax: 14.75109 ymax: 54.80712
epsg (SRID):    4326.
proj4string:    +proj=longlat +datum=WGS84 +no_defs.
\describe{
  \item{MED_DESCRI }{Is the road a divided multi-lane highway with a median? Character string.}
  \item{RTT_DESCRI }{Primary or secondary route? Character string.}
  \item{F_CODE_DES }{Feature code description (road or trail). Character string.}
  \item{ISO }{ISO 3166-1 alpha-3 country code. Character string.}
  \item{ISOCOUNTRY }{Country name. Character string.}
}
}
\source{
Defense Mapping Agency (DMA), 1992. Digital Chart of the World. Defense Mapping Agency, Fairfax, Virginia. (Four CD-ROMs). Available through DIVA-GIS: \url{https://diva-gis.org/data.html} (accessed May 19, 2026).
}
\usage{
data(highways_deu1992)
}
\description{
Roads thematic layer from Digital Chart of the World. Subset: divided multi-lane highways.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/point2poly_tess.R
\name{point2poly_tess}
\alias{point2poly_tess}
\title{Point-to-polygon interpolation, tessellation method}
\usage{
point2poly_tess(
  pointz,
  polyz,
  poly_id,
  char_methodz = "aw",
  methodz = "aw",
  pop_raster = NULL,
  varz = NULL,
  pycno_varz = NULL,
  char_varz = NULL,
  char_assign = "biggest_overlap",
  funz = function(x, w) {
     stats::weighted.mean(x, w, na.rm = TRUE)
 },
  return_tess = FALSE,
  seed = 1
)
}
\arguments{
\item{pointz}{Source points layer. \code{sf} object.}

\item{polyz}{Destination polygon layer. Must have identical CRS to \code{pointz}. \code{sf} object.}

\item{poly_id}{Name of unique ID column for destination polygon layer. Character string.}

\item{char_methodz}{Interpolation method(s) for character strings. Could be either of "aw" (areal weighting, default) or "pw" (population weighting). See "details". Character string.}

\item{methodz}{Interpolation method(s) for numeric covariates. Could be either of "aw" (areal weighting, default) and/or "pw" (population weighting). See "details". Character string or vector of character strings.}

\item{pop_raster}{Population raster to be used for population weighting, Must be supplied if \code{methodz="pw"}. Must have identical CRS to \code{poly_from}. \code{raster} or \code{SpatRaster} object.}

\item{varz}{Names of numeric variable(s) to be interpolated from source polygon layer to destination polygons. Character string or list of character strings.}

\item{pycno_varz}{Names of spatially extensive numeric variables for which the pycnophylactic (mass-preserving) property should be preserved. Character string or vector of character strings.}

\item{char_varz}{Names of character string variables to be interpolated from source polygon layer to destination polygons. Character string or vector of character strings.}

\item{char_assign}{Assignment rule to be used for variables specified in \code{char_varz}. Could be either "biggest_overlap" (default) or "all_overlap". See "details". Character string or vector of character strings.}

\item{funz}{Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a numeric vector \code{x} and vector of weights \code{w}. Function or list of functions.}

\item{return_tess}{Return Voronoi polygons, in addition to destinaton polygon layer? Default is \code{FALSE}. Logical.}

\item{seed}{Seed for generation of random numbers. Default is 1. Numeric.}
}
\value{
If \code{return_tess=FALSE}, returns a \code{sf} polygon object, with variables from \code{pointz} interpolated to the geometries of \code{polyz}.

If \code{return_tess=TRUE}, returns a list, containing
\itemize{
 \item "result". The destination polygon layer. \code{sf} object.
 \item "tess". The (intermediate) Voronoi tessellation polygon layer. \code{sf} object.
 }
}
\description{
Function for interpolating values from a source point layer to a destination polygon layer, using Voronoi tessellation and area/population weights.
}
\details{
This function interpolates point data to polygons with a two-step process. In the first step (tessellation), each point is assigned a Voronoi cell, drawn such that (a) the distance from its borders to the focal point is less than or equal to the distance to any other point, and (b) no gaps between cells remain. The second step (interpolation) performs a polygon-in-polygon interpolation, using the Voronoi cells as source polygons.

Currently supported integration methods in the second step (\code{methodz}) include:
\itemize{
 \item Areal weighting ("aw"). Values from \code{poly_from} weighted in proportion to relative area of spatial overlap between source features and geometries of \code{poly_to}.
 \item Population weighting ("pw"). Values from \code{poly_from} weighted in proportion to relative population sizes in areas of spatial overlap between source features and geometries of \code{poly_to}. This routine uses a third layer (supplied in \code{pop_raster}) to calculate the weights.
}
When a list of variables are supplied and one methods argument specified, then the chosen method will be applied to all variables.

When a list of of variables are supplied and multiple methods arguments specified, then weighting methods will be applied in a pairwise order. For example, specifying \code{varz = list(c("to1","pvs1_margin"), c("vv1"))} and \code{methodz = c('aw', 'pw')} will apply areal weighting to the the first set of variables (to1 and pvs1_margin) and population weighing to the second set (vv1).

Interpolation procedures are handled somewhat differently for numeric and character string variables. For numeric variables supplied in \code{varz}, "aw" and/or "pw" weights are passed to the function specified in \code{funz}. If different sets of numeric variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).

For character string (and any other) variables supplied in \code{char_varz}, "aw" and/or "pw" weights are passed to the assignment rule(s) specified in \code{char_assign}. Note that the \code{char_varz} argument may include numerical variables, but \code{varz} cannot include character string variables.

Currently supported assignment rules for character strings (\code{char_assign}) include:
\itemize{
 \item "biggest_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned a single value from overlapping \code{poly_from} features, corresponding to the intersection with largest area and/or population weight.
 \item "all_overlap". For each variable in \code{char_varz}, the features in \code{poly_to} are assigned all values from overlapping \code{poly_from} features, ranked by area and/or population weights (largest-to-smallest) of intersections.
}
It is possible to pass multiple arguments to \code{char_assign} (e.g. \code{char_assign=c("biggest_overlap","all_overlap")}), in which case the function will calculate both, and append the resulting columns to the output.
}
\examples{
# Interpolation of a single variable, with area weights
data(hex_05_deu)
data(clea_deu2009_pt)
out_1 <- point2poly_tess(pointz = clea_deu2009_pt,
                             polyz = hex_05_deu,
                             poly_id = "HEX_ID",
                             varz = "to1")
plot(out_1["to1_aw"])

# Extract and inspect tessellation polygons
out_2 <- point2poly_tess(pointz = clea_deu2009_pt,
                             polyz = hex_05_deu,
                             poly_id = "HEX_ID",
                             varz = "to1",
                             return_tess = TRUE)
plot(out_2$tess["to1"])
plot(out_2$result["to1_aw"])

\donttest{
# Interpolation of multiple variables, with area and population weights
data(gpw4_deu2010)
gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
out_3 <- point2poly_tess(pointz = clea_deu2009_pt,
                         polyz = hex_05_deu,
                         poly_id = "HEX_ID",
                         methodz = c("aw","pw"),
                         varz = list(
                           c("to1","pvs1_margin"),
                           c("vv1")
                         ),
                         pycno_varz = "vv1",
                         funz = list(
                           function(x,w){stats::weighted.mean(x,w)},
                           function(x,w){sum(x*w)}
                           ),
                         char_varz = c("incumb_pty_n","win1_pty_n"),
                         pop_raster = gpw4_deu2010)
plot(out_3["vv1_pw"])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_data.R
\name{get_data}
\alias{get_data}
\title{Download data from SUNGEO server}
\usage{
get_data(
  country_names = NULL,
  country_iso3 = NULL,
  geoset = "GADM",
  geoset_yr = 2018,
  space_unit = "adm1",
  time_unit = "year",
  topics = NULL,
  year_min = 1990,
  year_max = 2017,
  print_url = TRUE,
  print_time = TRUE,
  error_stop = FALSE,
  by_topic = TRUE,
  skip_missing = TRUE,
  cache_param = FALSE,
  short_message = TRUE
)
}
\arguments{
\item{country_names}{Country name(s). Character string (single country) or vector of character strings (multiple countries).}

\item{country_iso3}{Country code (ISO 3166-1 alpha-3). Character string (single country) or vector of character strings (multiple countries).}

\item{geoset}{Name of geographic boundary set. Can be one of \code{"GADM"} (Database of Global Administrative Areas), \code{"GAUL"} (Global Administrative Unit Layers), \code{"geoBoundaries"}, \code{"GRED"} (GeoReferenced Electoral Districts Datasets), \code{"HEXGRID"} (SUNGEO Hexagonal Grid), \code{"MPIDR"} (Max Planck Institute for Demographic Research Population History GIS Collection), \code{"NHGIS"} (National Historical Geographic Information System), \code{"PRIOGRID"} (PRIO-GRID 2.0), \code{"SHGIS"} (SUNGEO Historical GIS). Default is \code{"GADM"}. Character string.}

\item{geoset_yr}{Year of geographic boundaries. See \code{get_info()['geosets']} for availability. Default is \code{2018}. Integer.}

\item{space_unit}{Geographic level of analysis. Can be one of \code{"adm0"} (country), \code{"adm1"} (province), \code{"adm2"} (district), \code{"cst"} (GRED electoral constituency), \code{"hex05"} (SUNGEO Hexagonal Grid cell), \code{"prio"} (PRIO-GRID cell). See \code{get_info()['geosets']} for availability by geoset, country and topic. Default is \code{"adm1"}. Character string.}

\item{time_unit}{Temporal level of analysis. Can be one of \code{"year"}, \code{"month"}, \code{"week"}. See \code{get_info()['topics']} for availability by topic. Default is \code{"year"}. Character string.}

\item{topics}{Data topics. See \code{get_info()['summary']} for full list. Character string (single topic) or vector of character strings (multiple topics).}

\item{year_min}{Time range of requested data: start year. See \code{get_info()['topics']} for availability by topic. Default is \code{1990}. Integer.}

\item{year_max}{Time range of requested data: end year. See \code{get_info()['topics']} for availability by topic. Default is \code{2017}. Integer.}

\item{print_url}{Print url string of requested data to console? Default is \code{TRUE}. Logical.}

\item{print_time}{Print processing time for API query to console? Default is \code{TRUE}. Logical.}

\item{error_stop}{Error handling. If \code{TRUE}, function terminates request if an error is encountered. If \code{FALSE}, error is skipped and error message is recorded in a new \code{message} column. Default is \code{FALSE}. Logical.}

\item{by_topic}{Break query down by topic and country? If \code{TRUE}, a separate request is sent to the API for each country and topic, and the results are combined on the client side. This ensures that data that are available for some, but not all countries are returned, rather than resulting in a failed request. If \code{FALSE}, a single request is sent to the API for all countries and topics, and the results are combined on the server side. Only data that are available for all countries are returned. Default is \code{TRUE}. Logical.}

\item{skip_missing}{Skip missing data topics? If \code{TRUE}, missing data topics are skipped, columns are populated with NAs, and corresponding error message is recorded in a new \code{message} column. If \code{FALSE}, returns NULL results for missing topics. Default is \code{TRUE}. Logical.}

\item{cache_param}{Store cached query on server? This can speed up processing for repeated queries. Default is \code{FALSE}. Logical.}

\item{short_message}{Shorten error messages? If \code{TRUE}, a short, informative error message is recorded in the \code{message} column. If \code{FALSE}, full error message is recorded. Default is \code{TRUE}. Logical.}
}
\value{
data.table object, with requested data from SUNGEO API.
}
\description{
Function to download data files through the SUNGEO API. Function produces a data.table object, corresponding to the user's choice of countries, topics, sources, and spatial and temporal units.
}
\examples{
\donttest{
# Single country, single topic
out_1 <- get_data(country_name="Afghanistan",topics="Demographics:Population:GHS")
out_1
}

\donttest{
out_2 <- get_data(
	country_name=c("Afghanistan","Moldova"),
	topics=c("Demographics:Ethnicity:EPR","Demographics:Population:GHS"))
out_2
}

\donttest{
# Other boundary sets, spatial and time units
out_3 <- get_data(
	country_name="Albania",
	topics="Weather:AirTemperatureAndPrecipitation:NOAA",
	geoset="GAUL",geoset_yr=1990,space_unit="adm2",time_unit="month",
	year_min=1990,year_max=1991)
out_3
}
}
\seealso{
\code{\link{get_info}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_info.R
\name{get_info}
\alias{get_info}
\title{Information on available SUNGEO data files}
\usage{
get_info(country_names = NULL, country_iso3s = NULL, topics = NULL)
}
\arguments{
\item{country_names}{Country name(s). Character string (single country) or vector of character strings (multiple countries).}

\item{country_iso3s}{Country code (ISO 3166-1 alpha-3). Character string (single country) or vector of character strings (multiple countries).}

\item{topics}{Data topics. See \code{get_info()} for full list. Character string (single topic) or vector of character strings (multiple topics).}
}
\value{
list object, with three slots: 'summary', 'topics', and 'geoset'.
}
\description{
This function reports the availability of data files on the SUNGEO server, searchable by country and topic.
}
\examples{
# Get list of all available data
out_1 <- get_info()
out_1["summary"]
out_1["topics"]
out_1["geosets"]

# Get list of available data for a single country
out_2 <- get_info(country_names="Afghanistan")
out_2

# Get list of available data for a single topic
out_3 <- get_info(topics="Elections:LowerHouse:CLEA")
out_3

# Get list of available data for a multiple countries and topics
out_4 <- get_info(
                 country_names=c("Afghanistan","Zambia"),
                 topics=c("Elections:LowerHouse:CLEA","Events:PoliticalViolence:GED"))
out_4

}
\seealso{
\code{\link{get_data}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{available_data}
\alias{available_data}
\title{Data availability through SUNGEO API}
\format{
List of 42 data.table objects
Geoset:GADM :Classes ‘data.table’ and 'data.frame': 249 obs. of  4 variables
Geoset:GAUL :Classes ‘data.table’ and 'data.frame': 242 obs. of  4 variables
Geoset:geoBoundaries :Classes ‘data.table’ and 'data.frame': 197 obs. of  4 variables
Geoset:GRED :Classes ‘data.table’ and 'data.frame': 74 obs. of  4 variables
Geoset:HEXGRID :Classes ‘data.table’ and 'data.frame': 199 obs. of  4 variables
Geoset:MPIDR :Classes ‘data.table’ and 'data.frame': 52 obs. of  4 variables
Geoset:NHGIS :Classes ‘data.table’ and 'data.frame': 1 obs. of  4 variables
Geoset:PRIOGRID :Classes ‘data.table’ and 'data.frame': 199 obs. of  4 variables
Geoset:SHGIS :Classes ‘data.table’ and 'data.frame': 68 obs. of  4 variables
\describe{
  \item{country_iso3 }{Codes for available countries (ISO 3166-1 alpha-3). Character string.}
  \item{country_name }{Names of available countries. Character string.}
  \item{geoset_years }{Years of available historical boundary files. Character string.}
  \item{space_units }{Available spatial units of analysis. Character string.}
}
Elections:LowerHouse:CLEA :Classes ‘data.table’ and 'data.frame': 168 obs. of  6 variables
Demographics:Ethnicity:EPR :Classes ‘data.table’ and 'data.frame': 180 obs. of  6 variables
Demographics:Ethnicity:GREG :Classes ‘data.table’ and 'data.frame': 234 obs. of  6 variables
Demographics:Population:GHS :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
Events:PoliticalViolence:ABADarfur :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ACLED :Classes ‘data.table’ and 'data.frame': 100 obs. of  6 variables
Events:PoliticalViolence:BeissingerProtest :Classes ‘data.table’ and 'data.frame': 15 obs. of  6 variables
Events:PoliticalViolence:BeissingerRiot :Classes ‘data.table’ and 'data.frame': 15 obs. of  6 variables
Events:PoliticalViolence:BeissingerUkraine :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:COCACW :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCAfghanistanWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCIraqSIGACT :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCIraqWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCMexicoDrugRelatedMurders :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCMexicoHomicide :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCPakistanBFRS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:ESOCPakistanWITS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:GED :Classes ‘data.table’ and 'data.frame': 121 obs. of  6 variables
Events:PoliticalViolence:Lankina :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:NIRI :Classes ‘data.table’ and 'data.frame': 12 obs. of  6 variables
Events:PoliticalViolence:NVMS :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:PITF :Classes ‘data.table’ and 'data.frame': 133 obs. of  6 variables
Events:PoliticalViolence:SCAD :Classes ‘data.table’ and 'data.frame': 60 obs. of  6 variables
Events:PoliticalViolence:yzCaucasus2000 :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:yzChechnya :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:yzLibya :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Events:PoliticalViolence:yzUkraine2014 :Classes ‘data.table’ and 'data.frame': 1 obs. of  6 variables
Infrastructure:Roads:gRoads :Classes ‘data.table’ and 'data.frame': 240 obs. of  6 variables
Infrastructure:NightLights:DMSP :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
PublicHealth:Covid19:JHUCSSEC19 :Classes ‘data.table’ and 'data.frame': 207 obs. of  6 variables
Terrain:Elevation:ETOPO1 :Classes ‘data.table’ and 'data.frame': 256 obs. of  6 variables
Terrain:LandCover:GLCC :Classes ‘data.table’ and 'data.frame': 257 obs. of  6 variables
Weather:AirTemperatureAndPrecipitation:NOAA :Classes ‘data.table’ and 'data.frame': 209 obs. of  6 variables
\describe{
  \item{country_iso3 }{Codes for available countries (ISO 3166-1 alpha-3). Character string.}
  \item{country_name }{Names of available countries. Character string.}
  \item{year_range }{Range of available years for data topic. Character string.}
  \item{time_units }{Available time units. Character string.}
  \item{space_units }{Available spatial units. Character string.}
  \item{geosets }{Names of available geographic boundary data sources. Character string.}
}
}
\source{
Sub-National Geospatial Data Archive System: Geoprocessing Toolkit (updated March 17, 2023).
}
\usage{
data(available_data)
}
\description{
Census of geospatial and processed data files available to download using \code{SUNGEO::get_data()}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{cc_dict}
\alias{cc_dict}
\title{Country code dictionary}
\format{
data.table object, with	8626 obs. of  3 variables:
\describe{
  \item{country_name }{Country names. Character string.}
  \item{country_name_alt }{Alternative spellings of country names, ASCII characters only. Character string.}
  \item{country_iso3 }{Country codes (ISO 3166-1 alpha-3). Character string.}
}
}
\source{
Vincent Arel-Bundock. Package countrycode: Convert Country Names and Country Code, version 1.40. CRAN (October 12, 2022).
}
\usage{
data(cc_dict)
}
\description{
Reference table of country names and ISO-3166 codes, adapted from \code{countrycode} package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utm_select.R
\name{utm_select}
\alias{utm_select}
\title{Automatically convert geographic (degree) to planar coordinates (meters)}
\usage{
utm_select(x, max_zones = 5, return_list = FALSE)
}
\arguments{
\item{x}{Layer to be reprojected. \code{sf}, \code{sp}, \code{SpatRaster} or \code{RasterLayer} object.}

\item{max_zones}{Maximum number of UTM zones for single layer. Default is 5. Numeric.}

\item{return_list}{Return list object instead of reprojected layer (see Details). Default is \code{FALSE}. Logical.}
}
\value{
Re-projected layer. \code{sf} or \code{RasterLayer} object, depending on input.

If \code{return_list=TRUE}, returns a list object containing
\itemize{
 \item "x_out". The re-projected layer. \code{sf} or \code{RasterLayer} object, depending on input.
 \item "proj4_best".proj4string of the projection. Character string.
 }
}
\description{
Function to automatically convert simple feature, spatial and raster objects with geographic coordinates (longitude, latitude / WGS 1984, EPSG:4326) to planar UTM coordinates. If the study region spans multiple UTM zones, defaults to Albers Equal Area.
}
\details{
Optimal map projection for the object \code{x} is defined by matching its horizontal extent with that of the 60 UTM zones. If object spans multiple UTM zones, uses either the median zone (if number of zones is equal to or less than \code{max_zones}) or Albers Equal Area projection with median longitude as projection center (if number of zones is greater than \code{max_zones}).
}
\examples{
# Find a planar projection for an unprojected (WSG 1984) hexagonal grid of Germany
data(hex_05_deu)
sf::st_crs(hex_05_deu)
out_1 <- utm_select(hex_05_deu)
sf::st_crs(out_1)

# Find a planar projection for a raster
data(gpw4_deu2010)
gpw4_deu2010 <- terra::rast(gpw4_deu2010) # unwrap PackedSpatRaster
sf::st_crs(gpw4_deu2010)
out_2 <- utm_select(gpw4_deu2010)
sf::st_crs(out_2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sf2raster.R
\name{sf2raster}
\alias{sf2raster}
\title{Convert simple features object into regularly spaced raster}
\usage{
sf2raster(
  polyz_from = NULL,
  pointz_from = NULL,
  input_variable = NULL,
  reverse = FALSE,
  poly_to = NULL,
  return_output = NULL,
  return_field = NULL,
  aggregate_function = list(function(x) mean(x, na.rm = TRUE)),
  reverse_function = list(function(x) mean(x, na.rm = TRUE)),
  grid_dim = c(1000, 1000),
  cartogram = FALSE,
  carto_var = NULL,
  message_out = TRUE,
  return_list = FALSE
)
}
\arguments{
\item{polyz_from}{Source polygon layer. \code{sf} object.}

\item{pointz_from}{Source point layer. \code{sf} object.}

\item{input_variable}{Name of input variable from source layer. Character string.}

\item{reverse}{Reverse translation from raster layer to \code{sf} polygon object (polygon features only). Default is \code{FALSE}.}

\item{poly_to}{Destination polygon layer for reverse conversion. Must be specified if \code{reverse=TRUE}. \code{sf} object.}

\item{return_output}{Return output for reverse conversion. Must be specified if \code{reverse=TRUE}.}

\item{return_field}{Return field for reverse conversion. Must be specified if \code{reverse=TRUE}.}

\item{aggregate_function}{Aggregation function to be applied to variables specified in \code{input_variable}. Must take as an input a numeric vector \code{x}. Function or list of functions. Default is \code{mean}.}

\item{reverse_function}{Aggregation function for reverse conversion. Must be specified if \code{reverse=TRUE}. Function or list of functions. Default is \code{mean}.}

\item{grid_dim}{Dimensions of raster grid. Numerical vector of length 2 (number of rows, number of columns). Default is \code{c(1000,1000)}.}

\item{cartogram}{Cartogram transformation. Logical. Default is \code{FALSE}.}

\item{carto_var}{Input variable for cartogram transformation. Must be specified if \code{cartogram=TRUE}. Character string.}

\item{message_out}{Print informational messages. Logical. Default is \code{TRUE}.}

\item{return_list}{Return full set of results, including input polygons, centroids and field raster. Default is \code{FALSE}. Logical.}
}
\value{
If \code{return_list=FALSE} (default) and \code{reverse=FALSE} (default), returns \code{RasterLayer} object, with cell values corresponding to \code{input_variable}.

If \code{return_list=TRUE} and input layer is polygon, returns a list containing
\itemize{
 \item "return_output". Output raster, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
 \item "return_centroid". Raster of centroids, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
 \item "poly_to". Source polygons, with columns corresponding to \code{input_variable} and auto-generated numerical ID \code{Field}. \code{sf} object.
 \item "return_field". Output raster, with values corresponding to auto-generated numerical ID \code{Field}. \code{RasterLayer} object.}

If \code{return_list=TRUE} and  input layer is points, returns a list containing
\itemize{
 \item "return_output". Output raster, with values corresponding to \code{input_variable}. \code{RasterLayer} object.
 \item "return_point". Source points, with column corresponding to \code{input_variable}.
 }
If \code{reverse=TRUE}, returns an \code{sf} polygon layer, with columns corresponding to \code{input_variable} and auto-generated numerical ID \code{Field}.
}
\description{
This function takes in an \code{sf} spatial object (polygon or point) and returns a regularly spaced RasterLayer. Reverse translation option allows users to create an \code{sf} polygon object from the regularly spaced RasterLayer. This function can also conver the \code{sf} object into a cartogram with a user-specified variable name.
}
\examples{
# Rasterization of polygon layer.
data(clea_deu2009)
out_1 <- sf2raster(polyz_from = utm_select(clea_deu2009),
                   input_variable = "to1")
terra::plot(out_1)

\donttest{
# Rasterization of point layer
data(clea_deu2009_pt)
out_2 <- sf2raster(pointz_from = utm_select(clea_deu2009_pt),
                   input_variable = "to1",
                   grid_dim = c(25,25))
terra::plot(out_2)
}
\donttest{
# Cartogram (vote turnout scaled by number of valid votes)
out_3 <- sf2raster(polyz_from = utm_select(clea_deu2009),
                   input_variable = "to1",
                   cartogram = TRUE,
                   carto_var = "vv1")
terra::plot(out_3)
}
\donttest{
# Polygonization of cartogram raster
out_4a <- sf2raster(polyz_from = utm_select(clea_deu2009),
                    input_variable = "to1",
                    cartogram = TRUE,
                    carto_var = "vv1",
                    return_list = TRUE)
out_4 <- sf2raster(reverse = TRUE,
                   poly_to = out_4a$poly_to,
                   return_output = out_4a$return_output,
                   return_field = out_4a$return_field)
terra::plot(out_4)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df2sf.R
\name{df2sf}
\alias{df2sf}
\title{Convert data.frame object into simple features object}
\usage{
df2sf(
  x_coord,
  y_coord,
  input_data = NULL,
  file = NULL,
  n_max = Inf,
  start = 0,
  projection_input = "EPSG:4326",
  zero.policy = FALSE,
  show_removed = FALSE
)
}
\arguments{
\item{x_coord}{Numeric vector with longitude or easting projected coordinates. When \code{input_data} or \code{file} is supplied, can be either column name or numeric vector of the same length as \code{nrow(input_data)}.}

\item{y_coord}{Numeric vector with latitude or northing projected coordinates. Must be equal to the vector length of \code{x_coord}. When \code{input_data} or \code{file} is supplied, can be either column name or numeric vector of the same length as \code{nrow(input_data)}.}

\item{input_data}{Optional data frame object, containing \code{x_coord} and \code{y_coord}. \code{nrow(input_data)} must be equal to the vector length of \code{x_coord}. NOTE: Rows corresponding to non-usable coordinates are removed from the final output.}

\item{file}{Optional path to csv file. Overrides \code{input_data}.}

\item{n_max}{Maximum number of rows to read in \code{file}. Default is \code{Inf}.}

\item{start}{Number of rows to skip in \code{file}. Default is 0 (start on first row).}

\item{projection_input}{Projection string associated with \code{x_coord} and \code{y_coord}. Default is \code{'+proj=longlat'}.}

\item{zero.policy}{If \code{TRUE}, removes rows where corresponding coordinates equals (0,0). Default is \code{FALSE}.}

\item{show_removed}{If \code{TRUE}, returns a vector of indices corresponding to non-usable coordinates. Default is \code{FALSE}.}
}
\value{
If \code{show_removed==FALSE}, returns an \code{sf} object, with rows corresponding to non-usable coordinates removed. If \code{show_removed==TRUE}, returns a list, with an \code{sf} object (\code{Spatial_Coordinates}), and a vector of indices corresponding to non-usable coordinates removed (\code{Removed_Rows}).
}
\description{
Function takes in x-, y-coordinates, and a data.frame of variables (optional) and returns an SFC object
}
\examples{
# Coordinates supplied as vectors
data(clea_deu2009_df)
out_1 <- df2sf(x_coord=clea_deu2009_df$longitude,y_coord = clea_deu2009_df$latitude)
class(out_1)
plot(out_1$geometry)

# Coordinates supplied as column mames
out_2 <- df2sf(x_coord="longitude",y_coord ="latitude", input_data = clea_deu2009_df)
plot(out_2["geometry"])

# Load from external file
\donttest{
tmp <- tempfile()
write.csv(clea_deu2009_df,file=tmp)
out_3 <- df2sf(x_coord="longitude",y_coord ="latitude", file=tmp)
plot(out_3["geometry"])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smart_round.R
\name{smart_round}
\alias{smart_round}
\title{Smart numerical rounding function}
\usage{
smart_round(x, rnd = 0, return_char = TRUE)
}
\arguments{
\item{x}{Vector of values to be rounded. Numeric.}

\item{rnd}{Requested number of decimal places. Default is 0. Non-negative integer.}

\item{return_char}{Return rounded values as character string? Default is TRUE. Logical.}
}
\value{
If \code{return_char=TRUE}, returns a character string of same length as \code{x}. If \code{return_char=FALSE}, returns a numerical vector of same length as \code{x}.
}
\description{
Function to round numerical values with minimal information loss (e.g. to avoid "0.000" values in tables).
}
\details{
Rounds the values in its first argument to the specified number of decimal places (default 0). If brute-force rounding produces zero values (e.g. "0.00"), the number of decimal places is expanded to include the first significant digit.
}
\examples{
# Round a vector of numbers, character string output (best for tables)
out_1 <- smart_round(c(.0013,2.3,-1,pi),rnd=2)
out_1

# Round a vector of numbers, numerical output
out_2 <- smart_round(c(.0013,2.3,-1,pi),rnd=2,return_char=FALSE)
out_2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot_spot.R
\name{hot_spot}
\alias{hot_spot}
\title{Automatically calculate Local G hot spot intensity}
\usage{
hot_spot(
  insert,
  variable = NULL,
  style = "W",
  k = 9,
  remove_missing = TRUE,
  NA_Value = 0,
  include_Moran = FALSE
)
}
\arguments{
\item{insert}{Spatial point or spatial polygon object. Acceptable formats include \code{sf}, \code{SpatialPolygonsDataFrame}, \code{SpatialPointsDataFrame}.}

\item{variable}{Column name or numeric vector containing the variable from which the local G statistic will be calculated. Must possess a natural scale that orders small and large observations (i.e. number, percentage, ratio and not model residuals).}

\item{style}{Style can take values \code{'W'}, \code{'B'}, \code{'C'}, \code{'U'}, \code{'mimax'}, \code{'S'} (see  \code{\link[spdep]{nb2listw}}). Character string.}

\item{k}{Number of neighbors. Default is 9. Numeric.}

\item{remove_missing}{Whether to calculate statistic without missing values. If \code{FALSE}, substitute value must be supplied to \code{NA_Value}.}

\item{NA_Value}{Substitute for missing values. Default value is 0. Numeric.}

\item{include_Moran}{Calculate local Moran's I statistics. Default is \code{FALSE}. Logical.}
}
\value{
If \code{input} is \code{sf}, \code{SpatialPolygonsDataFrame} or \code{SpatialPointsDataFrame} object, returns \code{sf} object with same geometries and columns as \code{input}, appended with additional column containing Local G estimates (\code{LocalG}). If \code{input} is \code{RasterLayer} object, returns \code{RasterBrick} object containing original values (\code{Original}) and Local G estimates (\code{LocalG}).
}
\description{
Function automatically calculates the Local G hot spot intensity measure for spatial points or spatial polygons. Uses RANN for efficient nearest neighbor calculation (spatial points only); users can specify the number of neighbors (k). Users can specify the neighborhood style (see spdep::nb2listw) with default being standardized weight matrix (W).
}
\examples{
# Calculate Local G for sf point layer
data(clea_deu2009_pt)
out_1 <- hot_spot(insert=clea_deu2009_pt, variable = clea_deu2009_pt$to1)
class(out_1)
plot(out_1["LocalG"])

# Calculate Local G for sf polygon layer (variable as numeric vector)
data(clea_deu2009)
out_2 <- hot_spot(insert=clea_deu2009, variable = clea_deu2009$to1)
summary(out_2$LocalG)
plot(out_2["LocalG"])

\donttest{
# Calculate Local G for sf polygon layer (variable as column name)
out_3 <- hot_spot(insert=clea_deu2009, variable = "to1")
summary(out_3$LocalG)
plot(out_3["LocalG"])
}

\donttest{
# Calculate Local G for a SpatialPolygonsDataFrame object (variable as column name)
out_4 <- hot_spot(insert=as(clea_deu2009,"Spatial"), variable = "to1")
summary(out_4$LocalG)
plot(out_4["LocalG"])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geocode_osm_batch.R
\name{geocode_osm_batch}
\alias{geocode_osm_batch}
\title{Batch geocode addresses and coordinates with OpenStreetMap}
\usage{
geocode_osm_batch(
  query = NULL,
  delay = 1,
  return_all = FALSE,
  match_num = 1,
  details = FALSE,
  reverse = FALSE,
  lon = NULL,
  lat = NULL,
  zoom = 10,
  user_agent = NULL,
  verbose = FALSE
)
}
\arguments{
\item{query}{Addresses or place names to be geocoded. Character string.
Ignored if \code{reverse=TRUE}.}

\item{delay}{Delay between requests, in seconds. Default is 1. Numeric.}

\item{return_all}{Should all matches be returned? Overrides \code{match_num}
if \code{TRUE}. Default is \code{FALSE}. Logical.}

\item{match_num}{If query matches multiple locations, which match to return?
Default is 1 (highest-ranking match, by relevance). Numeric.}

\item{details}{Should detailed results be returned? Default is \code{FALSE}.
Logical. Ignored if \code{reverse=TRUE}.}

\item{reverse}{Should reverse geocoding be performed (coordinates to
address)? Default is \code{FALSE}. Logical.}

\item{lon}{Longitudes of points to reverse geocode. Required if
\code{reverse=TRUE}. Must be the same length as \code{lat}. Numeric vector.}

\item{lat}{Latitudes of points to reverse geocode. Required if
\code{reverse=TRUE}. Must be the same length as \code{lon}. Numeric vector.}

\item{zoom}{Zoom level for reverse geocoding, controlling the level of detail
returned (0 = country, 10 = city, 18 = building). Default is 10. Numeric.}

\item{user_agent}{Valid User-Agent identifying the application for
OSM Nominatim. If none supplied, function will attempt to auto-detect.
Character string.}

\item{verbose}{Print status messages and progress? Default is \code{FALSE}.
Logical.}
}
\value{
A \code{data.frame} object.

  For forward geocoding (\code{reverse=FALSE}):
  \itemize{
  \item \code{query}. User-supplied address query. Character string.
  \item \code{osm_id}. OpenStreetMap ID. Character string.
  \item \code{address}. OpenStreetMap address. Character string.
  \item \code{longitude}. Horizontal coordinate. Numeric.
  \item \code{latitude}. Vertical coordinate. Numeric.
  }
  If \code{details=TRUE}, also contains:
  \itemize{
  \item \code{osm_type}. OpenStreetMap feature type. Character string.
  \item \code{importance}. Relevance of Nominatim match to query, from 0
    (worst) to 1 (best). Numeric.
  \item \code{bbox_ymin}. Minimum vertical coordinate of bounding box. Numeric.
  \item \code{bbox_ymax}. Maximum vertical coordinate of bounding box. Numeric.
  \item \code{bbox_xmin}. Minimum horizontal coordinate of bounding box. Numeric.
  \item \code{bbox_xmax}. Maximum horizontal coordinate of bounding box. Numeric.
  }

  For reverse geocoding (\code{reverse=TRUE}):
  \itemize{
  \item \code{osm_name}. Full display name from OpenStreetMap. Character string.
  \item \code{osm_adm0_code}. ISO country code. Character string.
  \item \code{osm_adm0}. Country name. Character string.
  \item \code{osm_adm1}. State or first-level administrative division. Character string.
  \item \code{osm_adm2}. District or county. Character string.
  \item \code{osm_adm3}. Municipality. Character string.
  \item \code{osm_adm4}. Borough or village. Character string.
  }
}
\description{
Finds geographic coordinates of multiple addresses and place names (forward
geocoding), or converts multiple longitude/latitude coordinate pairs to place
names and administrative units (reverse geocoding), using OpenStreetMap's
Nominatim API.
}
\details{
Wrapper function for \code{\link[SUNGEO]{geocode_osm}}. Because the
  Nominatim Usage Policy stipulates an absolute maximum of 1 request per
  second, this function facilitates batch geocoding by adding a small delay
  between queries
  (\url{https://operations.osmfoundation.org/policies/nominatim/}).
}
\examples{
\donttest{
# Geocode multiple addresses (top matches only)
geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"))
}
\donttest{
# With progress reports
geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"), verbose = TRUE)
}
\donttest{
# Return detailed results for all matches
geocode_osm_batch(c("Ann Arbor", "East Lansing", "Columbus"),
  details = TRUE, return_all = TRUE)
}
\donttest{
# Reverse geocode multiple coordinates (city level)
geocode_osm_batch(
  reverse = TRUE,
  lon = c(-83.743, -84.556, -82.999),
  lat = c(42.278,  42.701,  39.961),
  zoom = 10
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{gpw4_deu2010}
\alias{gpw4_deu2010}
\title{Population count raster for Germany, 2010.}
\format{
class       : SpatRaster
dimensions  : 186, 220, 1  (nrow, ncol, nlyr)
resolution  : 0.04166667, 0.04166667  (x, y)
extent      : 5.875, 15.04167, 47.29167, 55.04167  (xmin, xmax, ymin, ymax)
coord. ref. : lon/lat WGS 84 (EPSG:4326)
source(s)   : memory
name        : gpw_v4_population_count_rev11_2010_2pt5_min
min value   :                                        0.00
max value   :                                    92915.66
}
\source{
Gridded Population of the World (GPW) v4: Population Count, v4.11 \doi{10.7927/H4JW8BX5}.
}
\usage{
data(gpw4_deu2010)
}
\description{
2.5 arc-minute resolution raster of estimates of human population (number of persons per pixel),
consistent with national censuses and population registers, for the year 2010.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nesting.R
\name{nesting}
\alias{nesting}
\title{Relative scale and nesting coefficients}
\usage{
nesting(
  poly_from = NULL,
  poly_to = NULL,
  metrix = "all",
  tol_ = 0.001,
  by_unit = FALSE
)
}
\arguments{
\item{poly_from}{Source polygon layer. \code{sf} object (polygon or multipolygon).}

\item{poly_to}{Destination polygon layer. Must have identical CRS to \code{poly_from}. \code{sf} object (polygon or multipolygon).}

\item{metrix}{Requested scaling and nesting metrics. See "details". Default is "all". Character string or vector of character strings.}

\item{tol_}{Minimum area of polygon intersection, in square meters. Default is 0.001. Numeric.}

\item{by_unit}{Include a by-unit decomposition of requested nesting metrics (if available)? Default is FALSE. Logical.}
}
\value{
Named list, with numeric values for each requested metric in \code{metrix}. If \code{by_unit==TRUE}, last element of list is a data.table, with nesting metrics disaggregated by source unit, where the first column is a row index for the source polygon layer.
}
\description{
Function to calculate relative scale and nesting metrics for changes of support from a source polygon layer to an overlapping (but spatially misaligned) destination polygon layer.
}
\details{
Currently supported metrics (\code{metrix}) include:
\itemize{
 \item Relative scale ("rs"). Measures whether a change-of-support (CoS) task is one of aggregation or disaggregation, by calculating the share of source units that are smaller than destination units. Its range is from 0 to 1, where values of 1 indicate pure aggregation (all source units are smaller than destination units) and values of 0 indicate no aggregation (all source units are at least as large as destination units). Values between 0 and 1 indicate a hybrid (i.e. some source units are smaller, others are larger than target units).
 \item Relative nesting ("rn"). Measures how closely source and destination boundaries align, by calculating the share of source units that cannot be split across multiple destination units. Its range is from 0 to 1, where values of 0 indicate no nesting (every source unit can be split across multiple destination units) and values of 1 indicate full nesting (no source unit can be split across multiple destination units).
 \item Relative scale, symmetric ("rs_sym"). Alternative measure of "rs", which ranges from -1 to 1. It calculates a difference between two proportions: the share of source units that is smaller than destination units (i.e. "rs" from standpoint of source units), and the share that is larger (i.e. "rs" from standpoint of destination units). Values of -1 indicate pure disaggregation (all source units are larger than destination units), 1 indicates pure aggregation (all source units are smaller than destination units). Values of 0 indicate that all source units are the same size as target units.
 \item Relative nesting, symmetric ("rn_sym"). Alternative measure of "rn", which ranges from -1 to 1. It calculates a difference between two components: the nesting of source units within destination units (i.e. "rn" from standpoint of source units), and the nesting of destination units within source units (i.e. "rn" from standpoint of destination units. Values of 1 indicate that source units are perfectly nested within destination units; -1 indicates that destination units are perfectly nested within source units.
 \item Relative scale, alternative ("rs_alt"). Alternative measure of "rs", rescaled as a proportion of destination unit area. This measure can take any value on the real line, with positive values indicating aggregation and negative values indicating disaggregation.
 \item Relative nesting, alternative ("rn_alt"). Alternative measure of "rn", which places more weight on areas of maximum overlap. The main difference between this measure and "rn" is its use of the maximum intersection area for each source polygon instead of averaging over the quadratic term. Two sets of polygons are considered nested if one set is completely contained within another, with as few splits as possible. If none or only a sliver of a source polygon area falls outside a single destination polygon, those polygons are "more nested" than a case where half of a source polygon falls in destination polygon A and half falls into another polygon B.
 \item Relative scale, conditional ("rs_nn"). Alternative measure of "rs", calculated for the subset of source units that are not fully nested within destination units.
 \item Relative nesting, conditional ("rn_nn"). Alternative measure of "rn", calculated for the subset of source units that are not fully nested within destination units.
 \item Proportion intact ("p_intact"). A nesting metric that requires no area calculations at all. This measure ranges from 0 to 1, where 1 indicates full nesting (i.e. every source unit is intact/no splits), and 0 indicates no nesting (i.e. no source unit is intact/all are split).
 \item Proportion fully nested ("full_nest"). A stricter version of "p_intact". This measure ranges from 0 to 1, where 1 indicates full nesting (i.e. every source unit is intact/no splits AND falls completely inside the destination layer), and 0 indicates no nesting (i.e. no source unit is both intact and falls inside destination layer).
 \item Relative overlap ("ro"). Assesses extent of spatial overlap between source and destination polygons. This measure is scaled between -1 and 1. Values of 0 indicate perfect overlap (there is no part of source units that fall outside of destination units, and vice versa). Values between 0 and 1 indicate a "source underlap" (some parts of source polygons fall outside of destination polygons; more precisely, a larger part of source polygon area falls outside destination polygons than the other way around). Values between -1 and 0 indicate a "destination underlap" (some parts of destination polygons fall outside of source polygons; a larger part of destination polygon area falls outside source polygons than the other way around). Values of -1 and 1 indicate no overlap (all source units fall outside destination units, and vice versa). This is a theoretical limit only; the function returns an error if there is no overlap.
 \item Gibbs-Martin index of diversification ("gmi"). Inverse of "rn", where values of 1 indicate that every source unit is evenly split across multiple destination units, and 0 indicates that no source unit is split across any destination units.
}
It is possible to pass multiple arguments to \code{metrix} (e.g. \code{metrix=c("rn","rs")}). The default (\code{metrix="all"}) returns all of the above metrics.

The function automatically reprojects source and destination geometries to Lambert Equal Area prior to calculation, with map units in meters.

Values of \code{tol_} can be adjusted to increase or decrease the sensitivity of these metrics to small border misalignments. The default value discards polygon intersections smaller than 0.001 square meters in area.
}
\examples{
# Calculate all scale and nesting metrics for two sets of polygons
data(clea_deu2009)
data(hex_05_deu)
nest_1 <- nesting(
              poly_from = clea_deu2009,
              poly_to = hex_05_deu
              )
nest_1

# Calculate just Relative Nesting, in the opposite direction
nest_2 <- nesting(
              poly_from = hex_05_deu,
              poly_to = clea_deu2009,
              metrix = "rn"
              )
nest_2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/point2poly_simp.R
\name{point2poly_simp}
\alias{point2poly_simp}
\title{Point-to-polygon interpolation, simple overlay method}
\usage{
point2poly_simp(
  pointz,
  polyz,
  varz,
  char_varz = NULL,
  funz = list(function(x) {
     sum(x, na.rm = TRUE)
 }),
  na_val = NA,
  drop_na_cols = FALSE
)
}
\arguments{
\item{pointz}{Source points layer. \code{sf} object.}

\item{polyz}{Destination polygon layer. Must have identical CRS to \code{pointz}. \code{sf} object.}

\item{varz}{Names of variable(s) to be assigned from source polygon layer to destination polygons. Character string or vector of character strings.}

\item{char_varz}{Names of character string variable(s) in \code{varz}. Character string or vector of character strings.}

\item{funz}{Aggregation function to be applied to variables specified in \code{varz}. Must take as an input a vector \code{x}. Function or list of functions.}

\item{na_val}{Value to be assigned to missing values. Defaul is \code{NA}. Logical or list.}

\item{drop_na_cols}{Drop columns with completely missing values. Defaul is \code{FALSE}. Logical.}
}
\value{
Returns a \code{sf} polygon object, with variables from \code{pointz} assigned to the geometries of \code{polyz}.
}
\description{
Function for assigning values from a source point layer to a destination polygon layer, using simple point-in-polygon overlays
}
\details{
Assignment procedures are the same for numeric and character string variables. All variables supplied in \code{varz} are passed directly to the function specified in \code{funz}. If different sets of variables are to be aggregated with different functions, both \code{varz} and \code{funz} should be specified as lists (see examples below).
}
\examples{
# Assignment of a single variable (sums)
data(hex_05_deu)
data(clea_deu2009_pt)
out_1 <- point2poly_simp(pointz=clea_deu2009_pt,
                         polyz=hex_05_deu,
                         varz="vv1")
plot(out_1["vv1"])

# Replace NA's with 0's
out_2 <- point2poly_simp(pointz = clea_deu2009_pt,
                         polyz = hex_05_deu,
                         varz = "vv1",
                         na_val = 0)
plot(out_2["vv1"])

\donttest{
# Multiple variables, with different assignment functions
out_3 <- point2poly_simp(pointz = clea_deu2009_pt,
                         polyz = hex_05_deu,
                         varz = list(
                           c("to1","pvs1_margin"),
                           c("vv1"),
                           c("incumb_pty_n","win1_pty_n")),
                         funz = list(
                           function(x){mean(x,na.rm=TRUE)},
                           function(x){sum(x,na.rm=TRUE)},
                           function(x){paste0(unique(na.omit(x)),collapse=" | ") }),
                         na_val = list(NA_real_,0,NA_character_))
plot(out_3["pvs1_margin"])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_ticker.R
\name{make_ticker}
\alias{make_ticker}
\title{Make date ticker}
\usage{
make_ticker(
  date_min = 19000101,
  date_max = as.integer(gsub("-", "", as.Date(Sys.Date())))
)
}
\arguments{
\item{date_min}{Start date, in YYYYMMDD format. Default is \code{19000101}. Integer.}

\item{date_max}{End date, in YYYYMMDD format. Default is today. Integer.}
}
\value{
\code{data.table} object, with seven columns:
\itemize{
 \item \code{DATE}. Date in YYYYMMDD format. Integer.
 \item \code{DATE_ALT}. Date in \code{Date} (YYYY-MM-DD) format. Date.
 \item \code{TID}. Date ID, in consecutive integer format. Integer.
 \item \code{YRWK}. Week in YYYYWW format. Integer.
 \item \code{WID}. Weed ID, in consecutive integer format. Integer.
 \item \code{YRMO}. Month in YYYYMM format. Integer.
 \item \code{MID}. Month ID, in consecutive integer format. Integer.
 \item \code{YEAR}. Year in YYYY format. Integer.
}
}
\description{
Function to create a table of consecutive dates, in SUNGEO-compliant format.
}
\examples{
# All dates from January 1, 1900 to today
out_1 <- make_ticker()
out_1

# All dates from January 1, 1200 to today
out_2 <- make_ticker(date_min=12000101)
out_2

# All dates from January 1, 1500 to December 31, 1899
out_3 <- make_ticker(date_min=15000101, date_max=18991231)
out_3

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line2poly.R
\name{line2poly}
\alias{line2poly}
\title{Line-in-polygon analysis}
\usage{
line2poly(
  linez,
  polyz,
  poly_id,
  measurez = c("length", "density", "distance"),
  outvar_name = "line",
  unitz = "km",
  reproject = TRUE,
  na_val = NA,
  verbose = TRUE
)
}
\arguments{
\item{linez}{Source polyline layer. \code{sf} object.}

\item{polyz}{Destination polygon layer. Must have identical CRS to \code{linez}. \code{sf} object.}

\item{poly_id}{Name of unique ID column for destination polygon layer. Character string.}

\item{measurez}{Desired measurements. Could be any of "length" (sum of line lengths by polygon), "density" (sum of line lengths divided by area of polygon) and/or "distance" (distance from each polygon to nearest line feature). Default is to report all three. Character string or vector of character strings.}

\item{outvar_name}{Name (root) to be given to output variable. Default is \code{"line"}. Character string.}

\item{unitz}{Units of measurement (linear). Defaul is \code{"km"}. Character string.}

\item{reproject}{Temporarily reproject layers to planar projection for geometric operations? Defaul is \code{TRUE}. Logical.}

\item{na_val}{Value to be assigned to missing values (line lengths and densities only). Defaul is \code{NA}. Logical or list.}

\item{verbose}{Print status messages and progress? Default is \code{TRUE}. Logical.}
}
\value{
An \code{sf} polygon object, with summary statisics of \code{linez} features aggregated to the geometries of \code{polyz}.

If \code{measurez = "lengths"}, contains fields with suffixes
\itemize{
 \item "\code{_length}". Sum of line lengths within each polygon, in km or other units supplied in \code{unitz}.
 }
If \code{measurez = "density"}, contains fields with suffixes
\itemize{
 \item "\code{_length}". Sum of line lengths within each polygon, in km or other units supplied in \code{unitz}.
 \item "\code{_area}". Area of each polygon, in km^2 or the square of linear units supplied in \code{unitz}.
 \item "\code{_density}". Sum of line lengths divided by area of each polygon, in km/km^2 or other units supplied in \code{unitz}.
 }
If \code{measurez = "distance"}, contains fields with suffixes
\itemize{
 \item "\code{_distance}". Distance from each polygon to nearest line feature, in km or other units supplied in \code{unitz}.
 }
If \code{measurez = c("length","density","distance")} (default), contains all of the above.
}
\description{
Function for basic geometry calculations on polyline features, within an overlapping destination polygon layer.
}
\examples{
# Road lengths, densities and distance from polygon to nearest highway
data(hex_05_deu)
data(highways_deu1992)
out_1 <- line2poly(linez = highways_deu1992,
                   polyz = hex_05_deu,
                   poly_id = "HEX_ID")
plot(out_1["line_length"])
plot(out_1["line_density"])
plot(out_1["line_distance"])

\donttest{
# Replace missing road lengths and densities with 0's, rename variables
out_2 <- line2poly(linez = highways_deu1992,
                   polyz = hex_05_deu,
                   poly_id = "HEX_ID",
                   outvar_name = "road",
                   na_val = 0)
plot(out_2["road_length"])
plot(out_2["road_density"])
plot(out_2["road_distance"])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SUNGEO.R
\docType{package}
\name{SUNGEO}
\alias{SUNGEO}
\title{\code{SUNGEO}}
\description{
Sub-National Geospatial Data Archive System: Geoprocessing Toolkit
}
\details{
See the README on
[GitHub](https://github.com/zhukovyuri/SUNGEO#readme)
}
